
/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : BookmarksList.cpp
 * DATE   : 11.11.2006
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "BookmarksList.h"
#include "EventsController.h"
#include "Bookmarks.h"
#include <QHBoxLayout>
#include <QGroupBox>
#include <QListWidget>
#include <QFileInfo>
#include <QPushButton>
#include <QtDebug>

/*------- local constants:
-------------------------------------------------------------------*/
const char* const BookmarksList::CAPTION     = QT_TR_NOOP( "All bookmarks list" );
const char* const BookmarksList::FILES       = QT_TR_NOOP( "Files" );
const char* const BookmarksList::BOOKMARKS   = QT_TR_NOOP( "Bookmarks" );
const char* const BookmarksList::CLEAR_LABEL = QT_TR_NOOP( "&Clear all" );
const char* const BookmarksList::GO_LABEL    = QT_TR_NOOP( "&Goto" );
const char* const BookmarksList::EXIT_LABEL  = QT_TR_NOOP( "&Exit" );


//*******************************************************************
// BookmarksList                                         CONSTRUCTOR
//*******************************************************************
BookmarksList::BookmarksList( QWidget* const in_parent )
: QDialog         ( in_parent )
, d_mainlay       ( new QHBoxLayout( this ) )
, d_files_gbox    ( new QGroupBox( tr( FILES ), this ) )
, d_files_lay     ( new QVBoxLayout( d_files_gbox ) )
, d_files_lstw    ( new QListWidget( d_files_gbox ) )
, d_bookmarks_gbox( new QGroupBox( tr( BOOKMARKS ), this ) )
, d_bookmarks_lay ( new QVBoxLayout( d_bookmarks_gbox ) )
, d_bookmarks_lstw( new QListWidget( d_bookmarks_gbox ) )
, d_btn_lay       ( new QVBoxLayout )
, d_clear_btn     ( new QPushButton( tr( CLEAR_LABEL ), this ) )
, d_go_btn        ( new QPushButton( tr( GO_LABEL ), this ) )
, d_exit_btn      ( new QPushButton( tr( EXIT_LABEL ), this ) )
, d_selected_file ( "" )
, d_selected_line ( -1 )
{
	d_files_lstw->setAlternatingRowColors( true );
	d_bookmarks_lstw->setAlternatingRowColors( true );

	setWindowTitle( tr( CAPTION ) );
	
	d_files_lay->addWidget( d_files_lstw );
	d_mainlay->addWidget( d_files_gbox );

	d_bookmarks_lay->addWidget( d_bookmarks_lstw );
	d_mainlay->addWidget( d_bookmarks_gbox );
	
	d_btn_lay->addStretch();
	d_btn_lay->addWidget( d_clear_btn );
	d_btn_lay->addWidget( d_go_btn );
	d_btn_lay->addWidget( d_exit_btn );
	d_mainlay->addLayout( d_btn_lay );
	
	connect( d_exit_btn      , SIGNAL( clicked() ), this, SLOT( reject() ) );
	connect( d_clear_btn     , SIGNAL( clicked() ), this, SLOT( clear_all() ) );
	connect( d_go_btn        , SIGNAL( clicked() ), this, SLOT( goto_bookmark() ) );
	connect( d_files_lstw    , SIGNAL( currentItemChanged( QListWidgetItem*,QListWidgetItem* ) ), this, SLOT( file_changed( QListWidgetItem* ) ) );
	connect( d_bookmarks_lstw, SIGNAL( itemDoubleClicked( QListWidgetItem* ) ), this, SLOT( bookmark_selected( QListWidgetItem* ) ) );
}
// end of BookmarksList

//*******************************************************************
// showEvent                                       PRIVATE inherited
//*******************************************************************
void BookmarksList::showEvent( QShowEvent* const in_event )
{
	const QStringList files = Bookmarks::instance()->data().keys();

	if( !files.empty() ) {
		qint32 maxlen = 0;
		QFileInfo fi;
		QString fname;
		QStringList::const_iterator it = files.begin();
		const QStringList::const_iterator end = files.end();
		while( it != end ) {
			fi.setFile( *it );
			fname = fi.fileName();
			if( fname.length() > maxlen ) maxlen = fname.length();
			QListWidgetItem* const item = new QListWidgetItem( fname, d_files_lstw );
			if( item ) item->setData( PATH_ROLE, *it );
			++it;
		}
		d_files_lstw->sortItems();
		const qint32 dx = ( maxlen + 2 ) * d_files_lstw->fontMetrics().width( "X" );
		d_files_lstw->setMaximumWidth( dx );
		d_bookmarks_lstw->setMaximumWidth( dx );
		d_files_lstw->setCurrentRow( 0 );
		adjustSize();
	}
	QDialog::showEvent( in_event );
}
// end of showEvent

//*******************************************************************
// file_changed                                         PRIVATE slot
//*******************************************************************
void BookmarksList::file_changed( QListWidgetItem* const in_item )
{
	if( in_item ) {
		d_selected_file = in_item->data( PATH_ROLE ).toString();
		d_bookmarks_lstw->clear();
		QHash< qint32, Points::PointInfo > data;
		Bookmarks::instance()->file_data( d_selected_file, data );
		if( !data.empty() ) {
			QList< qint32 > pts = data.keys();
			qSort( pts );

			QString text;
			QList< qint32 >::const_iterator it = pts.begin();
			const QList< qint32 >::const_iterator end = pts.end();
			while( it != end ) {
				text = '(';
				text += QString::number( *it + 1 );
				text += ") ";
				text += data[ *it ].text();
				QListWidgetItem* const item = new QListWidgetItem( text, d_bookmarks_lstw );
				if( item ) item->setData( NUMBER_ROLE, *it );
				++it;
			}
		}
	}
}
// end of file_changed

//*******************************************************************
// clear_all                                            PRIVATE slot
//*******************************************************************
void BookmarksList::clear_all()
{
	Bookmarks::instance()->clear();
	EventsController::instance()->send_event( Event::BOOKMARK_CLEAR_ALL_REQUEST );
	reject();
}
// end of bookmark_selected

//*******************************************************************
// bookmark_selected                                    PRIVATE slot
//*******************************************************************
void BookmarksList::bookmark_selected( QListWidgetItem* const in_item )
{
	if( in_item ) {
		d_selected_line = ( d_selected_file.isEmpty() ) ? -1 : in_item->data( NUMBER_ROLE ).toInt();
		accept();
	}
}
// end of bookmark_selected

//*******************************************************************
// goto_bookmark                                        PRIVATE slot
//*******************************************************************
void BookmarksList::goto_bookmark()
{
	bookmark_selected( d_bookmarks_lstw->currentItem() );
}
// end of goto_bookmark
