/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : BrowserBuild.cpp
 * DATE   : 24.09.2006
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "BrowserBuild.h"
#include "Config.h"
#include "EventsController.h"
#include <QTextCharFormat>
#include <QTextCursor>
#include <QTextBlock>
#include <QtDebug>

/*------- local constants:
-------------------------------------------------------------------*/
const char* const BrowserBuild::FINISH  = QT_TR_NOOP( "*** Finish ***" );
const char* const BrowserBuild::ELAPSED = QT_TR_NOOP( "Time elapsed: min:%1, sec:%2, msec:%3" );

//*******************************************************************
// BrowserBuild                                          CONSTRUCTOR
//*******************************************************************
BrowserBuild::BrowserBuild( QWidget* const in_parent )
: QTextBrowser( in_parent )
, org_palette_( palette() )
, org_font_   ( font() )
, syntax_     ( new SyntaxHighlighter( document() ) )
, time_       ( QTime() )
{
	setAutoFillBackground( true );
	set_properties();

	EventsController::instance()->append( this, Event::PROCESS_STARTED );
	EventsController::instance()->append( this, Event::PROCESS_BUILD_DATA );
	EventsController::instance()->append( this, Event::PROCESS_BUILD_DATA_NEWLINE );
	EventsController::instance()->append( this, Event::PROCESS_BROKEN );
	EventsController::instance()->append( this, Event::BUILD_TAB_PROP_CHANGED );
}
// end of BrowserBuild

//*******************************************************************
// BrowserBuild::~BrowserBuild                            DESTRUCTOR
//*******************************************************************
BrowserBuild::~BrowserBuild()
{
	EventsController::instance()->remove( this );
}
// end of ~BrowserBuild

//*******************************************************************
// customEvent                                     PRIVATE inherited
//*******************************************************************
void BrowserBuild::customEvent( QEvent* const in_event )
{
	Event* const event = dynamic_cast< Event* >( in_event );
	const qint32 type = static_cast<qint32>( event->type() );
         
	switch( type ) {
		case Event::PROCESS_STARTED:
			time_.start();
			clear();
			break;
		case Event::PROCESS_BUILD_DATA:
			add_line( event );
			break;
		case Event::PROCESS_BUILD_DATA_NEWLINE:
			append( " " );
			break;
		case Event::PROCESS_BROKEN:
			process_broken();
			break;
		case Event::BUILD_TAB_PROP_CHANGED:
			set_properties();
			break;
	}
}
// end of customEvent

//*******************************************************************
// set_properties                                            PRIVATE
//*******************************************************************
void BrowserBuild::set_properties()
{
	if( Config::instance()->build_tab_use_defaults() ) {
		setPalette( org_palette_ );
		setFont( org_font_ );
	}
	else {
		QPalette p( org_palette_ );
		p.setColor( viewport()->foregroundRole(), QColor::fromRgba( Config::instance()->build_tab_font_color() ) );
		p.setColor( viewport()->backgroundRole(), QColor::fromRgba( Config::instance()->build_tab_base_bkg() ) );
		viewport()->setPalette( p );
	
		// Ustawienie nowego fontu.
		QFont fnt;
		fnt.fromString( Config::instance()->build_tab_font() );
		setFont( fnt );

		// Ponowne napisanie tekstu.		
		syntax_->rehighlight();
	}
}
// end set_properties

//*******************************************************************
// add_line                                                  PRIVATE
//*******************************************************************
void BrowserBuild::add_line( Event* const in_event )
{
	append( in_event->d_data.at( 0 ).toString() );
}
// end of add_line

//*******************************************************************
// process_broken                                            PRIVATE
//*******************************************************************
void BrowserBuild::process_broken()
{
	const qint32 msec = time_.elapsed();
	const qint32 min  = msec / 60000;
	const qint32 sec  = msec % 60000;

	append( " " );
	append( tr( FINISH ) );
	append( tr( ELAPSED ).arg( min ).arg( sec / 1000 ).arg( sec % 1000 ) );
	append( " " );

	const QTextBlock end = document()->end();

	// Znalezienie ostatniej linii(bloku) tekstu.
	QTextBlock block = document()->begin();
	while( block != end ) {
		block = block.next();
	}
	// Ustawienie kursora na koncu tekstu.
	if( block != end ) {
		QTextCursor cursor = textCursor();
		cursor.setPosition( block.position() );
		setTextCursor( cursor );
		ensureCursorVisible();
	}
}
// end of process_broken

//*******************************************************************
// highlightBlock                                  PRIVATE inherited
//*******************************************************************
void BrowserBuild::SyntaxHighlighter::highlightBlock( const QString& in_block )
{
	QTextCharFormat default_format;
	default_format.setFont( document()->defaultFont() );
	default_format.setForeground( QColor::fromRgba( Config::instance()->build_tab_font_color() ) );
	setFormat( 0, in_block.length(), default_format );
}
// end of highlightBlock
