/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : BrowserDbg.h
 * DATE   : 13.11.2006
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "BrowserDbg.h"
#include "EventsController.h"
#include "BrowserDbgScope.h"
#include "BrowserDbgData.h"
#include "DbgCallStack.h"
#include "DbgRegisters.h"
#include "DbgDisasm.h"
#include "DbgMemory.h"
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QSplitter>
#include <QPushButton>
#include <QtDebug>

/*------- local constants:
-------------------------------------------------------------------*/
const char* const BrowserDbg::CALL_STACK = QT_TR_NOOP( "call stack" );
const char* const BrowserDbg::REGISTERS  = QT_TR_NOOP( "registers" );
const char* const BrowserDbg::MEMORY     = QT_TR_NOOP( "memory" );
const char* const BrowserDbg::DISASM     = QT_TR_NOOP( "disasm" );
const qint32      BrowserDbg::BTN_MARGIN = 10;

//*******************************************************************
// BrowserDbg                                            CONSTRUCTOR
//*******************************************************************
BrowserDbg::BrowserDbg( QWidget* const in_parent )
: QWidget             ( in_parent )
, d_label_width(      fontMetrics().width( tr( CALL_STACK ) ) )
, d_splitter          ( new QSplitter( Qt::Horizontal ) )
, d_call_stack_btn    ( create_button( tr( CALL_STACK ) ) )
, d_regs_btn          ( create_button( tr( REGISTERS ) ) )
, d_memory_btn        ( create_button( tr( MEMORY ) ) )
, d_disasm_btn        ( create_button( tr( DISASM ) ) )
, d_call_stack_dialog ( 0 )
, d_registers_dialog  ( 0 )
, d_disasm_dialog     ( 0 )
, d_memory_dialog     ( 0 )
{
   QHBoxLayout* const mlay = new QHBoxLayout( this );
   mlay->setMargin( 0 );

   mlay->addWidget( d_splitter );
   d_splitter->setHandleWidth( 2 );
   d_splitter->addWidget( new BrowserDbgScope );
   d_splitter->addWidget( new BrowserDbgData );
   
   QVBoxLayout* const blay = new QVBoxLayout;
   blay->setMargin( 0 );
   blay->setSpacing( 2 );

   blay->addWidget( d_call_stack_btn );
   blay->addWidget( d_regs_btn );
   blay->addWidget( d_disasm_btn );
   blay->addWidget( d_memory_btn );
   blay->addStretch();
   
   mlay->addLayout( blay );

   EventsController::instance()->append( this, Event::DBG_FINISHED );

   connect( d_call_stack_btn , SIGNAL( toggled(bool) ), this, SLOT( call_stack(bool) ) );
   connect( d_regs_btn       , SIGNAL( toggled(bool) ), this, SLOT( registers (bool) ) );
   connect( d_memory_btn     , SIGNAL( toggled(bool) ), this, SLOT( memory    (bool) ) );
   connect( d_disasm_btn     , SIGNAL( toggled(bool) ), this, SLOT( disasm    (bool) ) );   
}
// end of BrowserDbg

//*******************************************************************
// ~BrowserDbg                                            DESTRUCTOR
//*******************************************************************
BrowserDbg::~BrowserDbg()
{
   EventsController::instance()->remove( this );
}
// end of ~BrowserDbg

//*******************************************************************
// customEvent                                     PRIVATE inherited
//*******************************************************************
void BrowserDbg::customEvent( QEvent* const in_event )
{
   Event* const event = dynamic_cast< Event* >( in_event );
   const qint32 type = static_cast<qint32>( event->type() );
         
   switch( type ) {
      case Event::DBG_FINISHED:
         finish();
         break;
   }
}
// end of customEvent

//*******************************************************************
// create_button                                             PRIVATE
//*******************************************************************
QPushButton* BrowserDbg::create_button( const QString& in_text )
{
   QPushButton* const btn = new QPushButton( in_text );
   if( btn ) {
      btn->setCheckable( true );
      btn->setFixedWidth( d_label_width + BTN_MARGIN );
   }
   return btn;
}
// end of create_button

//*******************************************************************
// call_stack                                                PRIVATE
//*******************************************************************
void BrowserDbg::call_stack( const bool in_stat )
{
   if( in_stat ) {
      d_call_stack_dialog = new DbgCallStack();
      connect( d_call_stack_dialog, SIGNAL( finished(int) ), this, SLOT( call_stack_closed() ) );
      d_call_stack_dialog->show();
   }
   else {
      if( d_call_stack_dialog ) {
         disconnect( d_call_stack_dialog, SIGNAL( finished(int) ), this, SLOT( call_stack_closed() ) );
         d_call_stack_dialog->close();
         delete d_call_stack_dialog;
         d_call_stack_dialog = 0;
      }
   }
}
// end of call_stack

//*******************************************************************
// call_stack_closed                                    PRIVATE slot
//*******************************************************************
void BrowserDbg::call_stack_closed()
{
   d_call_stack_dialog = 0;
   d_call_stack_btn->setChecked( false );
}
// end of call_stack_closed

//*******************************************************************
// registers                                                 PRIVATE
//*******************************************************************
void BrowserDbg::registers( const bool in_stat )
{
   if( in_stat ) {
      d_registers_dialog = new DbgRegisters();
      connect( d_registers_dialog, SIGNAL( finished(int) ), this, SLOT( registers_closed() ) );
      d_registers_dialog->show();
   }
   else {
      if( d_registers_dialog ) {
         disconnect( d_registers_dialog, SIGNAL( finished(int) ), this, SLOT( registers_closed() ) );
         d_registers_dialog->close();
         delete d_registers_dialog;
         d_registers_dialog = 0;
      }
   }
}
// end of registers

//*******************************************************************
// registers_closed                                     PRIVATE slot
//*******************************************************************
void BrowserDbg::registers_closed()
{
   d_registers_dialog = 0;
   d_regs_btn->setChecked( false );
}
// end of registers_closed

//*******************************************************************
// memory                                                    PRIVATE
//*******************************************************************
void BrowserDbg::memory( const bool in_stat )
{
   if( in_stat ) {
      d_memory_dialog = new DbgMemory();
      connect( d_memory_dialog, SIGNAL( finished(int) ), this, SLOT( memory_closed() ) );
      d_memory_dialog->show();
   }
   else {
      if( d_memory_dialog ) {
         disconnect( d_memory_dialog, SIGNAL( finished(int) ), this, SLOT( memory_closed() ) );
         d_memory_dialog->close();
         delete d_memory_dialog;
         d_memory_dialog = 0;
      }
   }
}
// end of memory

//*******************************************************************
// memory_closed                                        PRIVATE slot
//*******************************************************************
void BrowserDbg::memory_closed()
{
   d_memory_dialog = 0;
   d_memory_btn->setChecked( false );
}
// end of memory_closed

//*******************************************************************
// disasm                                                    PRIVATE
//*******************************************************************
void BrowserDbg::disasm( const bool in_stat )
{
   if( in_stat ) {
      d_disasm_dialog = new DbgDisasm();
      connect( d_disasm_dialog, SIGNAL( finished(int) ), this, SLOT( disasm_closed() ) );
      d_disasm_dialog->show();
   }
   else {
      if( d_disasm_dialog ) {
         disconnect( d_disasm_dialog, SIGNAL( finished(int) ), this, SLOT( disasm_closed() ) );
         d_disasm_dialog->close();
         delete d_disasm_dialog;
         d_disasm_dialog = 0;
      }
   }
}
// end of disasm

//*******************************************************************
// disasm_closed                                        PRIVATE slot
//*******************************************************************
void BrowserDbg::disasm_closed()
{
   d_disasm_dialog = 0;
   d_disasm_btn->setChecked( false );
}
// end of registers_closed

//*******************************************************************
// finish                                                    PRIVATE
//*******************************************************************
void BrowserDbg::finish()
{
   d_call_stack_btn->setChecked( false );
   d_regs_btn->setChecked( false );
   d_memory_btn->setChecked( false );
   d_disasm_btn->setChecked( false );
}
// end of finish
