/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : BrowserSearch.cpp
 * DATE   : 24.09.2006
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "BrowserSearch.h"
#include "Config.h"
#include "EventsController.h"
#include <QFileInfo>
#include <QMouseEvent>
#include <QtDebug>


//*******************************************************************
// BrowserSearch                                         CONSTRUCTOR
//********************************************************************
BrowserSearch::BrowserSearch( QWidget* const in_parent )
: QListWidget  ( in_parent )
, d_org_palette( palette() )
, d_org_font   ( font() )
{
	set_properties();

	EventsController::instance()->append( this, Event::FIND_IN_PROJECT_FILES_REQUEST );
	EventsController::instance()->append( this, Event::SEARCH_BROWSER_ITEM );
	EventsController::instance()->append( this, Event::SEARCH_TAB_PROP_CHANGED );
	
	connect( this, SIGNAL( itemDoubleClicked( QListWidgetItem* ) ), this, SLOT( selection( QListWidgetItem* ) ) );
	connect( this, SIGNAL( itemEntered      ( QListWidgetItem* ) ), this, SLOT( selection( QListWidgetItem* ) ) );
}
// end of BrowserSearch

//*******************************************************************
// ~BrowserSearch                                         DESTRUCTOR
//********************************************************************
BrowserSearch::~BrowserSearch()
{
	EventsController::instance()->remove( this );
}
// end of ~BrowserSearch

//*******************************************************************
// customEvent                                     PRIVATE inherited
//*******************************************************************
void BrowserSearch::customEvent( QEvent* const in_event )
{
	Event* const event = dynamic_cast< Event* >( in_event );
	const qint32 type = static_cast<int>( event->type() );
	   
	switch( type ) {
		case Event::FIND_IN_PROJECT_FILES_REQUEST:
			clear();
			break;
		case Event::SEARCH_BROWSER_ITEM:
			add_item( event );
			break;
		case Event::SEARCH_TAB_PROP_CHANGED:
			set_properties();
			break;
	}
}
// end of customEvent

//*******************************************************************
// set_properties                                            PRIVATE
//*******************************************************************
void BrowserSearch::set_properties()
{
	setAlternatingRowColors( Config::instance()->search_tab_use_altbkg() );
	
	if( Config::instance()->search_tab_use_defaults() ) {
		setPalette( d_org_palette );
		setFont( d_org_font );
	}
	else {
		QPalette p( d_org_palette );
		p.setColor( QPalette::Text, QColor::fromRgba( Config::instance()->search_tab_font_color() ) );
		p.setColor( QPalette::Base, QColor::fromRgba( Config::instance()->search_tab_base_bkg() ) );
		p.setColor( QPalette::AlternateBase, QColor::fromRgba( Config::instance()->search_tab_altbase_bkg() ) );
		setPalette( p );
	
		// Ustawienie nowego fontu.
		QFont fnt;
		fnt.fromString( Config::instance()->search_tab_font() );
		setFont( fnt );
	}
}
// end set_properties

//*******************************************************************
// selection                                            PRIVATE slot
//-------------------------------------------------------------------
// Uzytkownik wybral pozycje.
// Oznacza to, ze chce aby otworzyc plik i ustawic kursor w linii
// zawierajacej znaleziona fraze.
// Jako parametry eventa wysylamy nastepujace informacje:
// ( 0 ) - sciezke do pliku, ktory ma byc otwarty,
// ( 1 ) - numer linii zawierajacej fraze, kursor ma byc ustawiony
//         w tej wlasnie linii.
//*******************************************************************
void BrowserSearch::selection( QListWidgetItem* const in_item )
{
	if( in_item ) {
		EventsController::instance()->send_event( Event::SELECTED_FILE_AND_ROW, in_item->data( PATH_ROLE ), in_item->data( LINE_NUMBER_ROLE ) );
	}
}
// end of selection

//*******************************************************************
// add_item                                                  PRIVATE
//-------------------------------------------------------------------
// W evencie otrzymujem trzy parametry. Sa to:
// ( 0 ) - sciezka do pliku
// ( 1 ) - numer linii zawierajacej fraze,
// ( 2 ) - tekst calej linii, w ktorym zawiera sie fraza.
//*******************************************************************
void BrowserSearch::add_item( Event* const in_event )
{
	const QFileInfo fi( in_event->d_data.at( 0 ).toString() );

	QString info = fi.fileName();
	info += " (";
	info += QString::number( in_event->d_data.at( 1 ).toInt() );
	info += ") ";
	info += in_event->d_data.at( 2 ).toString().trimmed();

	QListWidgetItem* const item = new QListWidgetItem( info, this );
	if( item ) {
		item->setData( LINE_NUMBER_ROLE, in_event->d_data.at( 1 ) );
		item->setData( PATH_ROLE, fi.absoluteFilePath() );
		addItem( item );
	}
}
// end of add_item
