/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : ClasserParser.cpp
 * DATE   : 8.10.2006
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "QBtParser.h"
#include "ClassParser.h"
#include "EventsController.h"
#include <QFile>
#include <QRegExp>
#include <QTextStream>
#include <QtDebug>
#include <algorithm>

//*******************************************************************
// ClassParser                                           CONSTRUCTOR
//*******************************************************************
ClassParser::ClassParser() : QThread()
, d_buffer    ( "" )
, d_fpath     ( "" )
, d_thread_idx( -1 )
, d_fread     ( false )
{
   connect( this, SIGNAL( finished() ), this, SLOT( finished_slot() ));
}
// end of ClassParser


//*******************************************************************
// set_data                                                   PUBLIC
//*******************************************************************
// Wersja w ktorej nalezy plik odczytac z dysku - parse.
void ClassParser::set_data( const QString& in_fpath, const qint32 in_thread_idx )
{
   d_fpath = in_fpath;
   d_thread_idx  = in_thread_idx;
   d_fread = true;
   d_buffer.clear();
}
// Wersja w ktorej tekst do parsowania zostal przyslany - reparse.
void ClassParser::set_data( const QString& in_fpath, const qint32 in_thread_idx, const QString& in_text )
{
   d_fpath = in_fpath;
   d_thread_idx = in_thread_idx;
   d_fread = false;
   d_buffer = in_text;
}
// end of set_data

//*******************************************************************
// finished_slot                                        PRIVATE slot
//-------------------------------------------------------------------
// Po zakonczeniu dzialania funckji 'run' QThread wywola sygnal,
// ktory tu obsluzymy.
//*******************************************************************
void ClassParser::finished_slot()
{
   const qint32 thread_idx = d_thread_idx;

   d_thread_idx = -1;
   d_buffer.clear();
   d_fpath.clear();

   emit finished( thread_idx );   // sygnal do rodzica, ze skonczylismy prace
}
// end of finished_slot

//*******************************************************************
// run                                                       PRIVATE
//*******************************************************************
void ClassParser::run()
{
   d_class_items.clear();
   d_class.clear();
   d_scope = ParsedItem::NO_SCOPE;

   if( !d_fpath.isEmpty() ) {
      bool ok = true;
      if( d_fread ) {
         ok = read_file( d_fpath );
      }
      d_len = d_buffer.length();
      if( d_len > 0 ) {
         d_idx = 0;
         parse();
      }
   }
   d_class_items.squeeze();
}
// end of run

//*******************************************************************
// parse                                                     PRIVATE
//*******************************************************************
qint32 ClassParser::parse()
{
   qint32 pos = -1;
   QChar  chr;

   while( index_is_valid() ) {
      chr = d_buffer[ d_idx ];
      if( '#' == chr ) {
         ++d_idx;
         skip_preprocesor();
         continue;
      }
      if( is_cpp_comment() ) {
         d_idx += 2;
         skip_cpp_comment();
         continue;
      }
      if( is_begin_of_c_comment() ) {
         d_idx += 2;
         skip_c_comment();
         continue;
      }
      if( chr.isLetter() || ( '~' == chr ) ) {
         if( -1 == pos ) {
            pos = d_idx++;
            continue;
         }
      }
      if( pos != -1 ) {
              if( ':' == chr ) parse_scope ( pos );
         else if( chr == '{' ) parse_block ( pos );
         else if( chr == ';' ) parse_item  ( pos );
         else if( chr == '(' ) parse_func  ( pos );
         else if( chr == '/' ) omit_comment( pos );
         else if( chr == '\'' ) { ++d_idx; skip_char () ; }
         else if( chr == '\"' ) { ++d_idx; skip_string(); }
         else                  ++d_idx;
         continue;
      }
      else if( chr == '}' ) {
         end_block( pos );
      }
      ++d_idx;
   }

   return 0;
}
// end of parse

//*******************************************************************
// omit_comment                                              PRIVATE
//*******************************************************************
void ClassParser::omit_comment( qint32& out_pos )
{
   if( ( d_idx < ( d_len - 1 ) ) && is_acceptable() ) {
      if( '*' == d_buffer[ d_idx + 1 ] ) {
         d_idx += 2;
         skip_c_comment();
         out_pos = -1;
      }
      else if( '/' == d_buffer[ d_idx + 1 ] ) {
         d_idx += 2;
         skip_cpp_comment();
         out_pos = -1;
      }
      else {
         ++d_idx;
      }
   }
   else {
      ++d_idx;
   }
}
// end of omit_comment

//*******************************************************************
// parse_block                                               PRIVATE
//-------------------------------------------------------------------
// Natrafiono na klamrowy nawias otwierajacy '{'.
//*******************************************************************
void ClassParser::parse_block( qint32& in_pos )
{
   const QStringList tokens( d_buffer.mid( in_pos, d_idx - in_pos ).simplified().split( QRegExp( "\\b" ), QString::SkipEmptyParts ) );
   QString text( "" );   
   bool is_class = false;

   const QStringList::const_iterator end = tokens.end();
   QStringList::const_iterator it = tokens.begin();
   while( it != end ) {
      text = (*it).trimmed();
      if( !text.isEmpty() ) {
         if( !is_class ) {
            is_class = ( "class" == text );
         }
         else {
            /* czy nalezy zapamietac informacje o klasie
               w osobnym obiekcie? Chyba nie.
            ParsedItem item;
            item.name( text );
            item.parent( ( d_class.isEmpty() ) ? "" : d_class.top() );
            item.type( ParsedItem::CLASS );
            item.file( d_fpath );
            d_class_items[ text ] << item;
            */
            
            d_class << text;
            d_scope = ParsedItem::PRIVATE;
            ++d_idx;
            break;
         }
      }
      ++it;
   }   

   // [ typedef ] struct, union, enum
   if( !is_class ) {
      ++d_idx;
      search_end_of( '{', '}' );

      // za blokiem nie bedacym cialem funckji musi znajdowac
      // sie jeszcze srednik ';'.
      first_of( ";" );
   }

   in_pos = -1;
}
// end of parse_block

//*******************************************************************
// end_block                                                 PRIVATE
//-------------------------------------------------------------------
// Natrafiono na klamrowy nawias zamykajacy '}'.
// Dotyczy to tylko i wylacznie zamkniecia aktualnej klasy.
//*******************************************************************
void ClassParser::end_block( qint32& in_pos )
{
   if( !d_class.isEmpty() ) d_class.pop();
   d_scope = ParsedItem::NO_SCOPE;

   in_pos = -1;
   ++d_idx;
   first_of( ";" );
}
// end of end_block

//*******************************************************************
// parse_func                                                PRIVATE
//-------------------------------------------------------------------
// Natrafiono na nawias otwierajacy '(', czyli jest tylko jedna
// mozliwosc - mamy poczatek funkcji.
//*******************************************************************
void ClassParser::parse_func( qint32& in_pos )
{
   QRegExp expr;
   const QString text( d_buffer.mid( in_pos, d_idx - in_pos ) );

   QString class_name = ( d_class.empty() ) ? "" : d_class.top();

   ParsedItem item;
   item.file( d_fpath );
   item.type( ParsedItem::METHOD );
   item.scope( d_scope );
   item.pos( in_pos );
   
   // 1. WYZNACZENIE NAZWY FUNKCJI.
   // Jesli mamy doczynienia z defincja funkcji wyznaczamy
   // takze nazwe klasy.
   expr.setPattern( "([~]{0,1}\\w+\\s*[=]{0,2})\\s*$" );
   if( expr.indexIn( text ) != -1 ) {
      const qint32 n = expr.numCaptures();
      if( n > 0 ) {
         item.name( expr.cap( n ).trimmed() );
         item.parent( class_name );
      }
   }

   // 2. PARAMETRY I CIALO FUNCKJI.
   ++d_idx;
   if( search_end_of( '(', ')' ) ) {   // szukamy parametrow.
      // Pelna nazwa funkcji.
      item.text( d_buffer.mid( in_pos, d_idx - in_pos ).simplified().trimmed() );
         
      // Sprawdzenie czy funkcja zawiera definicja.
      const qint32 idx = first_of( ";{" );
      if( idx != -1 ) {
         if( ';' == d_buffer[ idx ] ) {
            // Tylko deklaracja.
            item.what( ParsedItem::DECL );
            d_idx = idx + 1;
         }
         else {
            // Jesli jestesmy w srodku klasy to mamy deklaracje z implementacja (inline),
            // w przeciwnym przypadku mamy tylko implementacje.
            item.what( ( ParsedItem::NO_SCOPE == d_scope ) ? ParsedItem::IMPL : ParsedItem::DECL_IMPL );
            d_idx = idx + 1;
            search_end_of( '{', '}' );
         }
      }
   }

   d_class_items[ class_name ] << item;
   in_pos = -1;
}
// end of parse_func

//*******************************************************************
// parse_item                                                PRIVATE
//*******************************************************************
void ClassParser::parse_item( qint32& in_pos )
{
   const QString class_name = ( d_class.isEmpty() ) ? "" : d_class.top();
   
   QString text( d_buffer.mid( in_pos, d_idx - in_pos ).simplified().trimmed() );
   if( !text.isEmpty() ) {
      if( text.indexOf( '=' ) != -1 ) {
         // Przypisanie wartosci do statycznej stalej.
         // Narazie mnie nie interesuje.
      }
      else if( text.startsWith( "typedef" ) ) {
         // Mozna obsluzyc typedef
         // Narazie mnie nie interesuje
      }
      else if( text.startsWith( "class" ) )  {
         // deklaracja wyprzedzajaca
         // qDebug() << text;
      }
      else {
         if( !class_name.isEmpty() ) {
            ParsedItem item;
         
            item.file( d_fpath );
            item.type( ParsedItem::MEMBER );
            item.scope( d_scope );
            item.parent( class_name );
            item.text( text );
            item.pos( in_pos );

            // Testujemy przypadek czy przez przypadek
            // nie mamy zmiennej tablicowej.
            const qint32 idx = text.indexOf( '[' );
            if( idx != -1 ) {
               text.resize( idx );
            }
            
            QRegExp expr( "(\\w+)\\s*$" );
            if( expr.indexIn( text ) != -1 ) {
               item.name( expr.cap( expr.numCaptures() ) );
            }
            d_class_items[ class_name ] << item;
         }
      }
   }

   ++d_idx;
   in_pos = -1;
}
// end of parse_item

//*******************************************************************
// parse_scope                                               PRIVATE
//-------------------------------------------------------------------
// Natrafiono na dwukropek.
// Interesuje nas tylko private:, protected:, public:, itd.
//*******************************************************************
void ClassParser :: parse_scope( qint32& in_pos )
{
   const QString text( d_buffer.mid( in_pos, d_idx - in_pos ) );
   qint32 idx         = -1;
   
   if( ( idx = text.indexOf( QRegExp( "\\bprivate\\b" ) ) ) != -1 ) {
      d_scope = ParsedItem::PRIVATE;
   }
   else if( ( idx = text.indexOf( QRegExp( "\\bprotected\\b" ) ) ) != -1 ) {
      d_scope = ParsedItem::PROTECTED;
   }
   else if( ( idx = text.indexOf( QRegExp( "\\bpublic\\b" ) ) ) != -1 ) {
      d_scope = ParsedItem::PUBLIC;
   }
   else if( ( idx = text.indexOf( QRegExp( "\\bsignals\\b" ) ) ) != -1 ) {
      d_scope = ParsedItem::PROTECTED | ParsedItem::SIGNAL;
   }

   if( idx != -1 ) {
      if( text.indexOf( QRegExp( "\\bslots\\b" ) ) != -1 ) {
         d_scope |= ParsedItem::SLOT;
      }
      in_pos = -1;
   }
   else {
      if( ( d_idx < ( d_len - 1 ) ) && ( ':' == d_buffer[ d_idx + 1 ] ) ) {
         // tak naprawde mamy doczynienia z sekwencja '::'
         ++d_idx;
      }
      else if( text.indexOf( QRegExp( "\\bclass\\b" ) ) != -1 ) {
         // Paczatek klasy
         // nic nie robimy - to bedzie obsluzone w 'parse_block'
         // po przechwyceniu nawiasu '{'.
      }
      else {
         // jakas etykieta - olewamy ja, kontynujemy dalsza prace za nia.
         in_pos = -1;
      }
   }
   ++d_idx;
}
// end of parse_scope

//*******************************************************************
// go_after_char                                             PRIVATE
//-------------------------------------------------------------------
// Wyszukanie w tekscie przyslanego znaku.
// Jesli znak zostal znaleziony 'd_idx' jest ZA tym znakiem.
//*******************************************************************
void ClassParser::go_after_char( const QChar& in_char )
{
   while( ( d_idx >= 0 ) && ( d_idx < d_len ) ) {
      if( ( in_char == d_buffer[ d_idx ] ) && is_acceptable() ) {
         ++d_idx;
         break;
      }
      ++d_idx;
   }
}
// end of skip_string

//*******************************************************************
// skip_char                                                 PRIVATE
//*******************************************************************
inline void ClassParser::skip_char()
{
   go_after_char( '\'' );
}
// end of skip_char

//*******************************************************************
// skip_string                                               PRIVATE
//*******************************************************************
inline void ClassParser::skip_string()
{
   go_after_char( '\"' );
}
// end of skip_string

//*******************************************************************
// skip_cpp_comment                                          PRIVATE
//*******************************************************************
inline void ClassParser::skip_cpp_comment()
{
   go_after_char( '\n' );
}
// end of skip_cpp_comment

//*******************************************************************
// search_end_of                                             PRIVATE
//-------------------------------------------------------------------
// Szuka znaku konczacego blok.
// O jaki blok chodzi okreslaja przyslane znaki okreslajace
// znaki rozpoczynajace i konczace znaki.
// UWAGA:
//    'd_idx' okreslajacy aktualny indeks w tekscie powinien byc
//    juz WEWNATRZ bloku, tzn. ZA znakiem rozpoczynajacym blok,
//    o ktory nam chodzi.
// RETURN:
//     true - jesli znaleziono koniec bloku,
//            'd_idx' jest ZA znakiem konczacym blok.
//    false - nie znaleziono konca bloku,
//*******************************************************************
bool ClassParser::search_end_of( const QChar& in_smark, const QChar& in_emark )
{
   bool   retval = false;
   qint32 level  = 1;
   
   while( index_is_valid() ) {
      // ------- single quote (znaki)
      if( ( '\'' == d_buffer[ d_idx ] ) && is_acceptable() ) {
         ++d_idx;
         skip_char();
         continue;
      }
      // ------- double quote (lancuchy tekstowe)
      if( ( '\"' == d_buffer[ d_idx ] )  && is_acceptable() ) {
         ++d_idx;
         skip_string();
         continue;
      }
      // ------- c-komentarza
      if( is_begin_of_c_comment() ) {
         d_idx += 2;
         skip_c_comment();
         continue;
      }
      // ------- cpp-komentarz
      if( is_cpp_comment() ) {
         d_idx += 2;
         skip_cpp_comment();
         continue;
      }

      if( ( in_smark == d_buffer[ d_idx ] ) && is_acceptable() ) {
         ++level;
         ++d_idx;
         continue;
      }
      else if( ( in_emark == d_buffer[ d_idx ] ) && is_acceptable() ) {
         if( 0 == --level ) {
            ++d_idx;
            retval = true;
            break;
         }
      }
      ++d_idx;
   }
   return retval;
}
// end of search_end_of

//*******************************************************************
// first_of                                                  PRIVATE
//-------------------------------------------------------------------
// Znajduje jeden ze znakow zawartych w przyslanym zbiorze.
// Zwraca DOKLADNA pozycje znaku i 'd_idx' jest ZA nim.
//*******************************************************************
qint32 ClassParser::first_of( const QString& in_set )
{
   qint32 retval = -1;
   
   while( index_is_valid() ) {
      // ------- single quote (znaki)
      if( ( '\'' == d_buffer[ d_idx ] ) && is_acceptable() ) {
         ++d_idx;
         skip_char();
         continue;
      }
      // ------- double quote (lancuchy tekstowe)
      if( ( '\"' == d_buffer[ d_idx ] )  && is_acceptable() ) {
         ++d_idx;
         skip_string();
         continue;
      }
      // ------- c-komentarza
      if( is_begin_of_c_comment() ) {
         d_idx += 2;
         skip_c_comment();
         continue;
      }
      // ------- cpp-komentarz
      if( is_cpp_comment() ) {
         d_idx += 2;
         skip_cpp_comment();
         continue;
      }
      if( in_set.indexOf( d_buffer[ d_idx] ) != -1 ) {
         retval = d_idx++;
         break;
      }
      ++d_idx;
   }
   return retval;
}
// end of first_of

//*******************************************************************
// skip_c_comment                                            PRIVATE
//-------------------------------------------------------------------
// Wyszukiwanie konca komentarza typu C (/* ... */).
// UWAGA: 'd_idx' MUSI byc juz WEWNATRZ komentarza.
//*******************************************************************
void ClassParser::skip_c_comment()
{
   while( index_is_valid() ) {
      // sprawdzenie czy koniec c-komentarza.
      if( is_end_of_c_comment() ) {
         d_idx += 2;
         break;
      }
      // omijamy znaki
      if( ( '\'' == d_buffer[ d_idx ] ) && is_acceptable() ) {
         ++d_idx;
         skip_char();
         continue;
      }
      // omijamy lancuchy tekstowe
      if( ( '\"' == d_buffer[ d_idx ] )  && is_acceptable() ) {
         ++d_idx;
         skip_string();
         continue;
      }
      ++d_idx;
   }
}
// end of skip_c_comment

//*******************************************************************
// skip_preprocesor                                          PRIVATE
//-------------------------------------------------------------------
// Wyszukiwanie konca linii dyrektywy preprocesora (rozpoczetej znakiem #)..
// Koniec dyrektywy normalnie jest na koncu aktualnej linii.
// Jedynym wyjatkiem jest sytuacja, kiedy w tej samej linii zaczyna
// sie c-komentarz. Wowczas musimy isc az za ten komentarz.
// UWAGA:
//    'd_idx' MUSI byc juz ZA znakiem '#'.
// RETURN:
//    Po znalezieniu konca dyrektywy 'd_idx' jest ZA ostatnim
//    znakiem dyrektywy (lub c-komentarza).
//*******************************************************************
void ClassParser::skip_preprocesor()
{
   while( index_is_valid() ) {
      // ------- koniec linii - koniec dyrektywy
      if( ( '\n' == d_buffer[ d_idx ] ) && is_acceptable() ) {
         ++d_idx;
         break;   // koniec linii oznacza koniec dyrektywy
      }
      // ------- omijamy znaki
      if( ( '\'' == d_buffer[ d_idx ] ) && is_acceptable() ) {
         ++d_idx;
         skip_char();
         continue;
      }
      // ------- omijamy lancuchy tekstowe
      if( ( '\"' == d_buffer[ d_idx ] )  && is_acceptable() ) {
         ++d_idx;
         skip_string();
         continue;
      }
      // ------- c-komentarz
      if( is_begin_of_c_comment() ) {
         d_idx += 2;
         skip_c_comment();
         break;   // koniec c-komentarza oznacza koniec dyrektywy
      }
      ++d_idx;
   }
}
// end of skip_preprocesor

//*******************************************************************
// index_is_valid                                            PRIVATE
//-------------------------------------------------------------------
// Sprawdzenie czy aktualny indeks 'd_idx' jest WEWNATRZ tekstu.
//*******************************************************************
inline bool ClassParser::index_is_valid() const
{
   return ( ( d_idx >= 0 ) && ( d_idx < d_len ) );
}
// end of index_is_valid

//*******************************************************************
// is_acceptable                                             PRIVATE
//-------------------------------------------------------------------
// Sprawdzenie czy znak rzeczywiscie oznacza to co reprezentuje.
// Znaczenie znaku moze modyfikowac znak '\' znajdujacy sie przed nim.
// Ale trzeba sprawdzic czy '\' rzeczywiscie oznacza modyfikacje
// znaczenia. Jesli przed nim znajduje sie kolejny '\' to '\' nie jest
// modyfikatorem lecz prawdziwym znakiem '\'.
//*******************************************************************
inline bool ClassParser::is_acceptable() const
{
   return is_acceptable( d_idx );
}
inline bool ClassParser::is_acceptable( const qint32 in_idx ) const
{
   bool retval = true;
   
   if( ( in_idx > 0 ) && ( '\\' == d_buffer[ in_idx - 1 ] ) ) {
      retval = ( ( in_idx >  1 ) && ( '\\' == d_buffer[ in_idx - 2 ] ));
   }

   return retval;
}
// end of is_acceptable



//*******************************************************************
// is_cpp_comment                                            PRIVATE
//-------------------------------------------------------------------
// Sprawdzenie czy znak na pozycji 'd_idx' jest poczatkiem
// cpp-komentarza (// ... ).
//*******************************************************************
inline bool ClassParser::is_cpp_comment() const
{
   const bool retval =
      ( '/' == d_buffer[ d_idx ]     ) &&
      ( d_idx < ( d_len - 1 )        ) &&
      ( '/' == d_buffer[ d_idx + 1 ] ) &&
      ( is_acceptable()              );
   return retval;  
}
// end of is_cpp_comment

//*******************************************************************
// is_begin_of_c_comment                                     PRIVATE
//-------------------------------------------------------------------
// Sprawdzenie czy znak na pozycji 'd_idx' jest POCZATKIEM
// c-komentarza (/* ... */).
// PATRZ: is_end_of_c_comment.
//*******************************************************************
inline bool ClassParser::is_begin_of_c_comment() const
{
   const bool retval =
      ( '/' == d_buffer[ d_idx ]     ) &&
      ( d_idx < ( d_len - 1 )        ) &&
      ( '*' == d_buffer[ d_idx + 1 ] ) &&
      ( is_acceptable()              );
   return retval;  
}
// end of is_begin_of_c_comment

//*******************************************************************
// is_end_of_c_comment                                       PRIVATE
//-------------------------------------------------------------------
// Sprawdzenie czy znak na pozycji 'd_idx' jest KONCEM
// c-komentarza (/* ... */).
// PATRZ: is_begin_of_c_comment.
//*******************************************************************
inline bool ClassParser::is_end_of_c_comment() const
{
   const bool retval =
      ( '*' == d_buffer[ d_idx ]     ) &&
      ( d_idx < ( d_len - 1 )        ) &&
      ( '/' == d_buffer[ d_idx + 1 ] ) &&
      ( is_acceptable()              );
   return retval;  
}
// end of is_end_of_c_comment

//*******************************************************************
// read_file                                                 PRIVATE
//-------------------------------------------------------------------
// Odczytuje plik z dysku.
// Caly plik jest zapamietany w zmiennej 'd_buffer' jako jedna linia
// tekstu (brak struktury podzialu na linie).
//*******************************************************************
bool ClassParser::read_file( const QString& in_fpath )
{
   bool retval = false;
   d_buffer.clear();

   QFile file( in_fpath );
   if( file.open( QIODevice::ReadOnly ) ) {
      QTextStream in( &file );
      d_buffer = in.readAll();
      file.close();
      retval = true;
   }
   return retval;
}
// end of read_file

//*******************************************************************
// print                                                     PRIVATE
//*******************************************************************
void ClassParser::print()
{
   QVector<ParsedItem::Type> types;
   types
         << ParsedItem::GLOBAL
         << ParsedItem::CLASS
         << ParsedItem::STRUCT
         << ParsedItem::UNION
         << ParsedItem::ENUM
         << ParsedItem::METHOD
         << ParsedItem::MEMBER
         ;

   ClassItems::iterator it = d_class_items.begin();
   while( it != d_class_items.end() ) {
      const QString class_name = it.key();
/*
      qDebug() << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++";
      qDebug() << "Class name: " << class_name;
      //if( class_name.isEmpty() )
      foreach( ParsedItem::Type typ, types ) {
         if( d_class_items[ class_name ].contains( typ ) ) {
            QVector<ParsedItem>& vr = d_class_items[ class_name][ typ ];
            QVector<ParsedItem>::const_iterator vit = vr.begin();
            while( vit != vr.end() ) {
               (*vit).print();
               ++vit;
            }
         }
      }
*/      
      ++it;
   }
}
// end of print


//###################################################################
//#                                                                 #
//#                            ParsedItem                           #
//#                                                                 #
//###################################################################


//*******************************************************************
// ParsedItem                                            CONSTRUCTOR
//*******************************************************************
ParsedItem::ParsedItem()
{
   reset();
}
// end of ParsedItem

//*******************************************************************
// ParsedItem                                       COPY CONSTRUCTOR
//*******************************************************************
ParsedItem::ParsedItem( const ParsedItem& in_other )
{
   d_file   = in_other.d_file;
   d_name   = in_other.d_name;
   d_type   = in_other.d_type;
   d_what   = in_other.d_what;
   d_text   = in_other.d_text;
   d_scope  = in_other.d_scope;
   d_parent = in_other.d_parent;
   d_pos    = in_other.d_pos;
}
// end of ParsedItem

//*******************************************************************
// ~ParsedItem                                            DESTRUCTOR
//*******************************************************************
ParsedItem::~ParsedItem()
{
   reset();
}
// end of ~ParsedItem

//*******************************************************************
// operator=                                                  PUBLIC
//*******************************************************************
ParsedItem& ParsedItem::operator=( const ParsedItem& in_other )
{
   if( this != &in_other ) {
      ParsedItem temp( in_other );
      std::swap( *this, temp );
   }
   return *this;
}
// end of operator=

//*******************************************************************
// operator==                                                 PUBLIC
//*******************************************************************
bool ParsedItem::operator==( const ParsedItem& in_other )
{
   bool retval = true;
   
   if( this != &in_other ) {
      retval =
         ( d_file   == in_other.d_file   ) &&
         ( d_name   == in_other.d_name   ) &&
         ( d_type   == in_other.d_type   ) &&
         ( d_what   == in_other.d_what   ) &&
         ( d_text   == in_other.d_text   ) &&
         ( d_scope  == in_other.d_scope  ) &&
         ( d_parent == in_other.d_parent ) &&
         ( d_pos    == in_other.d_pos    );
   }
   return retval;
}
// end of operator==

//*******************************************************************
// reset                                                     PRIVATE
//*******************************************************************
void ParsedItem::reset()
{
   d_file.clear();
   d_name.clear();
   d_type = GLOBAL;
   d_what = UNKNOWN;
   d_text.clear();
   d_scope = NO_SCOPE;
   d_parent.clear();
   d_pos = -1;
}
// end of reset

//*******************************************************************
// print                                                      PUBLIC
//*******************************************************************
void ParsedItem::print() const
{
   qDebug() << "\t-->File  : " << d_file;
   qDebug() << "\t\tParent: " << d_parent;
   qDebug() << "\t\tScope : " << scope_as_text( d_scope );
   qDebug() << "\t\tName  ; " << d_name;
   qDebug() << "\t\tText  : " << d_text.simplified();
   qDebug() << "";
}

//*******************************************************************
// scope_as_text                                              PUBLIC
//*******************************************************************
QString ParsedItem::scope_as_text( const quint32 in_scope ) const
{
   QString retval;

        if( in_scope & PUBLIC    ) retval = "public";
   else if( in_scope & PROTECTED ) retval = "protected";
   else if( in_scope & PRIVATE   ) retval = "private";
   
        if( in_scope & SIGNAL ) retval += " signals";
   else if( in_scope & SLOT   ) retval += " slots"  ;
   
   return retval;
}
// end of scope_as_text
