/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : Classer.cpp
 * DATE   : 8.10.2006
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "Classer.h"
#include "Shared.h"
#include "ClassParser.h"
#include "EventsController.h"
#include <QApplication>
#include <QMutexLocker>
#include <QtDebug>

/*------- local variables:
-------------------------------------------------------------------*/
Classer* Classer::d_instance = 0;
QMutex   Classer::d_work_mtx;

//*******************************************************************
// Classer                                               CONSTRUCTOR
//*******************************************************************
Classer::Classer() : QObject()
, d_idx( -1 )
{
	setParent( qApp );
	
	d_files.clear();
	for( qint32 i = 0; i < THREAD_MAX_COUNT; ++i ) {
		connect( &d_thread[ i ], SIGNAL( finished( qint32 ) ), this, SLOT( finished_handler( qint32 )));
		d_used[ i ] = 0;
	}
}
// end of Classer

//*******************************************************************
// ~Classer                                               DESTRUCTOR
//*******************************************************************
Classer::~Classer()
{
	for( qint32 i = 0; i < THREAD_MAX_COUNT; ++i ) {
		d_thread[ i ].terminate();
		d_thread[ i ].wait();
		d_used[ i ] = 0;
	}
}
// end of ~Classer

//*******************************************************************
// instance                                                   PUBLIC
//*******************************************************************
Classer* Classer::instance()
{
	if( 0 == d_instance ) {
		d_instance = new Classer();
	}
	return d_instance;
}
// end of instance

//*******************************************************************
// end_of_process                                            PRIVATE
//*******************************************************************
void Classer::end_of_proccess()
{
	if( d_reparse ) {
		/* Przypadek gdy dokonywalismy ponownego parsowania pliku.
		 * Wektor 'd_classes' zawiera tylko i wylacznie dopiero co
		 * parsowane klasy. Aby nie przeciazac procesora uaktualnianiem
		 * listy klas sprawdzamy czy w ogole cos sie zmienilo.
		 * Jezeli aktualna klasa nie jest zgodna z klasa zapamietana
		 * w 'd_uptodate_classes' oznacza ze cos sie zmienilo.
		 * Aktualny stan klasy zapamietujemy w 'd_uptodate_classes'
		 * i ustawiamy flage oznaczajaca, ze sa zmiany.
		 */
		d_del_classes.clear();
		d_new_classes.clear();
		d_mod_classes.clear();

		if( d_classes != d_uptodate_classes ) {
			// SPRAWDZAMY CZY USUNIETO KLASY LUB CZY SIE ZMIENILY.
			const ClassParser::ClassItems::const_iterator end_uptodate = d_uptodate_classes.end();
			ClassParser::ClassItems::const_iterator it = d_uptodate_classes.begin();
			while( it != end_uptodate ) {
				const QString class_name = it.key();
				if( d_classes.contains( class_name ) ) {
					if( d_classes[ class_name ] != d_uptodate_classes[ class_name ] ) {
						d_mod_classes << class_name;
					}
				}
				else {
					d_del_classes << class_name;
				}
				++it;
			}
			// SPRAWDZAMY CZY POJAWILY SIE NOWE KLASY.
			const ClassParser::ClassItems::const_iterator end_classes = d_classes.end();
			it = d_classes.begin();
			while( it != end_classes ) {
				const QString class_name = it.key();
				if( !d_uptodate_classes.contains( class_name ) ) {
					d_new_classes << class_name;
				}
				++it;
			}
		}
		squeeze_classes( d_uptodate_classes = d_classes );
		d_classes.clear();
		
		if( !d_new_classes.empty() || !d_del_classes.empty() || !d_mod_classes.empty() ) {
			EventsController::instance()->send_event( Event::PROJECT_REPARSED );
		}
	}
	else {
		squeeze_classes( d_uptodate_classes = d_classes );
		d_classes.clear();
		EventsController::instance()->send_event( Event::PROJECT_PARSED );
	}
}
// end of end_of_process

//*******************************************************************
// parse                                                      PUBLIC
//*******************************************************************
void Classer::parse( const QStringList& in_files )
{
	d_reparse = false;
	d_files = in_files;
	d_classes.clear();

	if( !d_files.empty() ) {
		d_idx = 0;
		while( ( d_idx < d_files.size() ) && ( d_idx < THREAD_MAX_COUNT ) ) {
			parse( d_files.at( d_idx++ ) );
		}
	}
}
// end of parse

//*******************************************************************
// reparse                                                    PUBLIC
//*******************************************************************
void Classer::reparse( const QString& in_fpath, const QString& in_text )
{
	d_reparse = true;
	d_files.clear();
	d_classes = d_uptodate_classes;
	remove_data( d_classes, in_fpath );

	d_files.push_back( in_fpath );
	d_idx = 0;
	
	d_work_mtx.lock();
		const qint32 idx = get_thread();
		d_used[ idx ] = 1;
		d_thread[ idx ].set_data( d_files.at( d_idx++ ), idx, in_text );
		d_thread[ idx ].start();
	d_work_mtx.unlock();
}
// end of reparse

//*******************************************************************
// parse_next                                                PRIVATE
//-------------------------------------------------------------------
// Funkcja ta jest wywolywana po zakonczeniu ktoregos z watkow.
// Powoduje zainicjowanie kolejnego watku na rzecz analizy kolejnego
// pliku.
//*******************************************************************
void Classer::parse_next()
{
	if( !d_files.empty() ) {
		if( d_idx < d_files.size() ) {
			parse( d_files.at( d_idx++ ) );
		}
		else {
			d_files.clear();
			d_idx = -1;
			if( all_finished() ) end_of_proccess();
		}
	}
	else {
		if( all_finished() ) end_of_proccess();
	}
}
// end of parse_next


//*******************************************************************
// parse                                                      PUBLIC
//*******************************************************************
void Classer::parse( const QString& in_fpath )
{
	d_work_mtx.lock();
		qint32 idx = get_thread();
		d_used[ idx ] = 1;
		d_thread[ idx ].set_data( in_fpath, idx );
		d_thread[ idx ].start();
	d_work_mtx.unlock();
}
// end of parse

//*******************************************************************
// finished                                             PRIVATE slot
//*******************************************************************
void Classer::finished_handler( const qint32 in_idx )
{
	d_work_mtx.lock();
		// PO ZAKONCZENIU PRACY WATKU POBIERAMY WYNIKI JEGO PRACY.
		const ClassParser::ClassItems items = d_thread[ in_idx ].data();
		add_data( items );

		// ZAKONCZENIE WATKU.
		d_thread[ in_idx ].terminate();
		d_thread[ in_idx ].wait();
		d_used[ in_idx ] = 0;
	d_work_mtx.unlock();

	parse_next();
}
// end of finidshed

//*******************************************************************
// remove_data                                               PRIVATE
//-------------------------------------------------------------------
// Usuniecie ze zbioru informacji o klasach, ktore pochodza ze
// wskazanego pliku. Uzywane przy reparsingu.
//*******************************************************************
void Classer::remove_data( ClassParser::ClassItems& inout_classes, const QString& in_fpath )
{
	ClassParser::ClassItems new_classes;
	QVector< ParsedItem > new_items;
	
	ClassParser::ClassItems::iterator class_it = inout_classes.begin();
	while( class_it != inout_classes.end() ) {
		const QString class_name = class_it.key();
		const QVector< ParsedItem >& items = inout_classes[ class_name ];
		QVector< ParsedItem >::const_iterator it = items.begin();
		while( it != items.end() ) {
			if( in_fpath != it->file() ) {
				new_items << *it;
			}
			++it;
		}
		if( !new_items.empty() ) {
			new_classes[ class_name ] = new_items;
		}
		new_items.clear();
		++class_it;
	}
	
	inout_classes = new_classes;	
}
// end of remove_data

//*******************************************************************
// add_data                                                  PRIVATE
//-------------------------------------------------------------------
// Przepisanie wskazanych danych (z parsowania) do naszego
// zbiorczego bufora danych 'd_classes'.
//*******************************************************************
void Classer::add_data( const ClassParser::ClassItems& in_src )
{
	if( !in_src.empty() ) {
		const ClassParser::ClassItems::const_iterator end = in_src.end();
		ClassParser::ClassItems::const_iterator it = in_src.begin();
		while( it != end ) {
			const QString class_name = it.key();
			d_classes[ class_name ] << in_src[ class_name ];
			++it;
		}
	}
}
// end of add_data

//*******************************************************************
// squeeze_classes                                           PRIVATE
//*******************************************************************
void Classer::squeeze_classes( ClassParser::ClassItems& in_classes )
{
	in_classes.squeeze();
	if( !in_classes.empty() ) {
		const ClassParser::ClassItems::const_iterator end = in_classes.end();
		ClassParser::ClassItems::const_iterator it = in_classes.begin();
		while( it != end ) {
			in_classes[ it.key() ].squeeze();
			++it;
		}
	}
}
// end of squeeze_classes


void Classer::print( const ClassParser::ClassItems& in_src )
{
	qDebug() << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++";
	
	ClassParser::ClassItems::const_iterator it = in_src.begin();
	while( it != in_src.end() ) {
		const QString class_name = it.key();
		//qDebug() << ">> Class name: " << class_name;
		const QVector< ParsedItem >& items = in_src[ class_name ];
		QVector< ParsedItem >::const_iterator items_it = items.begin();
		while( items_it != items.end() ) {
			(*items_it).print();
			++items_it;
		}
		++it;
	}
}


//*******************************************************************
// get_thread                                                PRIVATE
//*******************************************************************
qint32 Classer::get_thread() const 
{
	qint32 retval = -1;

	for( qint32 i = 0; i < THREAD_MAX_COUNT; ++i ) {
		if( 0 == d_used[ i ] ) {
			retval = i;
			break;
		}
	}

	return retval;
}
// end of get_thread

//*******************************************************************
// all_finished                                              PRIVATE
//*******************************************************************
bool Classer::all_finished() const
{
	bool retval = true;

	for( qint32 i = 0; i < THREAD_MAX_COUNT; ++i ) {
		if( d_used[ i ] != 0 ) {
			retval = false;
			break;
		}
	}
	
	return retval;
}
// end of all_finished
