/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : DbgDataTable.cpp
 * DATE   : 16.02.2007
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "DbgDataTable.h"
#include "EventsController.h"
#include <QHeaderView>
#include <QtDebug>

/*------- local constants:
-------------------------------------------------------------------*/
const QRgb   DbgDataTable::GRAY               = qRgb( 100, 100, 100 );
const QRgb   DbgDataTable::NORMAL_TEXT_COLOR  = qRgb( 0, 0, 0 ); 
const QRgb   DbgDataTable::CHANGED_TEXT_COLOR = qRgb( 0, 0, 255 );
const QRgb   DbgDataTable::ERROR_TEXT_COLOR   = qRgb( 255, 0, 0 );


//*******************************************************************
// DbgDataTable                                   PUBLIC constructor
//*******************************************************************
DbgDataTable::DbgDataTable( QWidget* const in_parent, const qint32 in_cols,
							const bool in_set_min_width,
							const qint32 in_n_char_width  )
: QTableWidget    ( in_parent )
, d_X_width       ( fontMetrics().width( 'X' ) )
, d_n_char_width  ( in_n_char_width )
, d_row_height    ( fontMetrics().height() + 2 )
, d_idx           ( -1 )
, d_disabled_slot ( false )
, d_set_min_width ( in_set_min_width )
{
	d_prv_data.clear();

	QPalette plt = palette();
	plt.setColor( QPalette::Highlight, GRAY );
	setPalette( plt );
	
	setColumnCount( in_cols );
	verticalHeader()->hide();
	horizontalHeader()->hide();
	setAlternatingRowColors( true );
	setSelectionBehavior( SelectRows );
	setSelectionMode( SingleSelection );
}
// end of DbgDataTable

//*******************************************************************
// showEvent                                       PRIVATE inherited
//*******************************************************************
void DbgDataTable::showEvent( QShowEvent* in_event )
{
	QTableWidget::showEvent( in_event );
	resize_columns();
}
// end of showEvent

//*******************************************************************
// resizeEvent                                     PRIVATE inherited
//*******************************************************************
void DbgDataTable::resizeEvent( QResizeEvent* in_event )
{
	QTableWidget::resizeEvent( in_event );
	resize_columns();
}
// end of resizeEvent

//*******************************************************************
// finished                                                  PRIVATE
//*******************************************************************
void DbgDataTable::finished()
{
	remove_rows( 0 );
	d_prv_data.clear();
}
// end of finished

//*******************************************************************
// add_row                                                   PRIVATE
//-------------------------------------------------------------------
// Dodanie nowego wiersza i UTWORZENIE KOMOREK W OBU KOLUMNACH.
// Numer wiersza wskazany jest przez przyslany parametr.
// Jesli przyslano -1 nowy wiersz nalezy dodac na koncu tablicy.
//*******************************************************************
qint32 DbgDataTable::add_row( const qint32 in_row )
{
	qint32 retval = -1;

	if( -1 == d_idx ) {
		d_disabled_slot = true;

		const qint32 row = ( ( in_row < 0 ) || ( in_row > rowCount() ) ? rowCount() : in_row );
		insertRow( row );

		QTableWidgetItem* itm = 0;
		const qint32      n   = columnCount();
		
		for( qint32 i = 0; i < n; ++i ) {
			itm = new QTableWidgetItem();
			itm->setTextAlignment( Qt::AlignBottom );
			setItem( row, i, itm );
		}

		resize_rows();
		d_disabled_slot = false;
		retval = row;
	}
	return retval;
}
// end of add_row

//*******************************************************************
// remove_rows                                                PUBLIC
//*******************************************************************
void DbgDataTable::remove_rows( const qint32 in_new_row_count )
{
	while( in_new_row_count < rowCount() ) {
		removeRow( rowCount() - 1 );
	}
}
// end of remove_rows

//*******************************************************************
// colors                                                     PUBLIC
//*******************************************************************
void DbgDataTable::colors( const qint32 in_row, const bool in_error )
{
	QBrush my_brush( NORMAL_TEXT_COLOR );

	if( in_error ) {
		my_brush.setColor( ERROR_TEXT_COLOR );
	}
	else {
		QString name, value;

		get_item( in_row, 0, name );
		get_item( in_row, 1, value );

		if( d_prv_data.contains( name ) ) {
			if( d_prv_data[ name ] != value ) {
				my_brush.setColor( CHANGED_TEXT_COLOR );
			}
		}
	}

	QTableWidgetItem* itm = 0;
	const qint32      n   = columnCount();

	for( qint32 i = 0; i < n; ++i ) {
		itm = item( in_row, i );
		if( itm ) itm->setForeground( my_brush );
	}
}
// end of colors

//*******************************************************************
// set_item                                                   PUBLIC
//*******************************************************************
void DbgDataTable::set_item( const qint32 in_row, const qint32 in_col, const QString& in_text )
{
	QTableWidgetItem* const itm = item( in_row, in_col );
	if( itm ) itm->setText( in_text.trimmed() );
}
// end of set_item

//*******************************************************************
// get_item                                                   PUBLIC
//*******************************************************************
bool DbgDataTable::get_item( const qint32 in_row, const qint32 in_col, QString& out_text )
{
	out_text.clear();
	const QTableWidgetItem* const itm = item( in_row, in_col );
	if( itm ) out_text = itm->text();

	return ( out_text.isEmpty() == false );
}
// end of get_item

//*******************************************************************
// set_tool_tip                                               PUBLIC
//*******************************************************************
void DbgDataTable::set_tool_tip( const qint32 in_row, const qint32 in_col, const QString& in_text )
{
	QTableWidgetItem* const itm = item( in_row, in_col );
	if( itm ) itm->setToolTip( in_text.trimmed() );
}
// end of set_tool_tip

//*******************************************************************
// get_tool_tip                                               PUBLIC
//*******************************************************************
bool DbgDataTable::get_tool_tip( const qint32 in_row, const qint32 in_col, QString& out_text )
{
	out_text.clear();
	const QTableWidgetItem* const itm = item( in_row, in_col );
	if( itm ) out_text = itm->toolTip();

	return ( out_text.isEmpty() == false );
}
// end of get_tool_tip

//*******************************************************************
// resize_columns                                            PRIVATE
//*******************************************************************
void DbgDataTable::resize_columns()
{
	const qint32 n          = columnCount();
	const qint32 min_width  = d_n_char_width * d_X_width;
	const qint32 view_width = viewport()->width();

	for( qint32 i = 0; i < n; ++i ) {
		resizeColumnToContents( i );

		if( i == ( n - 1 ) ) {
			const qint32 total_width = total_columns_width();
			if( total_width < view_width ) {
				setColumnWidth( i, view_width - ( total_width - columnWidth( i ) ) );
			}
		}
		else {
			if( d_set_min_width ) {
				if( columnWidth( i ) < min_width ) {
					setColumnWidth( i, min_width );
				}
			}
		}
	}
}
// end of resize_columns

//*******************************************************************
// total_columns_width                                       PRIVATE
//*******************************************************************
qint32 DbgDataTable::total_columns_width()
{
	qint32 retval = 0;
	const qint32 n = columnCount();

	for( qint32 i = 0; i < n; ++i ) {
		retval += columnWidth( i );
	}
	return retval;
}
// end of total_columns_width

//*******************************************************************
// resize_rows                                               PRIVATE
//*******************************************************************
void DbgDataTable::resize_rows()
{
	for( qint32 i = 0; i < rowCount(); ++i ) {
		if( d_row_height != rowHeight( i ) ) {
			setRowHeight( i, d_row_height );
		}
	}
}
// end of resize_rows

//*******************************************************************
// get_prv_data                                              PRIVATE
//*******************************************************************
void DbgDataTable::get_prv_data()
{
	d_prv_data.clear();
	QString name, value;
		
	for( qint32 i = 0; i < rowCount(); ++i ) {
		get_item( i, 0, name );
		if( !name.isEmpty() ) {
			get_item( i, 1, value );
			d_prv_data[ name ] = value;
		}
	}
}
// end of get_prv_data
