/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : DbgDisasm.cpp
 * DATE   : 19.02.2007
 *******************************************************************/

/*------- include files:
--------------------------------------------------------------------*/
#include "DbgDisasm.h"
#include "EventsController.h"
#include <QApplication>
#include <QDesktopWidget>
#include <QHBoxLayout>
#include <QtDebug>

/*------- local constants:
-------------------------------------------------------------------*/
const char* const DbgDisasm::CAPTION          = QT_TR_NOOP( "Machine code" );
const QString     DbgDisasm::ASM_START_MARKER = "Dump of assembler code for ";
const QString     DbgDisasm::ASM_END_MARKER   = "End of assembler dump";
const QString     DbgDisasm::HEX_MARKER       = "0x";
const QString     DbgDisasm::ERROR            = "error";

//*******************************************************************
// DbgDisasm                                             CONSTRUCTOR
//*******************************************************************
DbgDisasm::DbgDisasm( QWidget* const in_parent )
: DbgToolDialog ( in_parent )
, d_table       ( new DbgDataTable( this, 4, false ) )
{
	setWindowTitle( tr( CAPTION ) );

	QHBoxLayout* const mlay = new QHBoxLayout( this );
	mlay->setMargin( 0 );
	mlay->setSpacing( 0 );
	mlay->addWidget( d_table );

	EventsController::instance()->append( this, Event::DBG_DISASM_DATA );
}
// end of DbgDisasm

//*******************************************************************
// DbgDisasm                                              DESTRUCTOR
//*******************************************************************
DbgDisasm::~DbgDisasm()
{
	EventsController::instance()->remove( this );
}
// end of ~DbgDiasasm

//*******************************************************************
// showEvent                                       PRIVATE inherited
//*******************************************************************
void DbgDisasm::showEvent( QShowEvent* const in_event )
{
	DbgToolDialog::showEvent( in_event );
	move( qApp->desktop()->screenGeometry().width() - size().width() - 4 - 90, 150 );

	EventsController::instance()->send_event( Event::DBG_DISASM_START );
	EventsController::instance()->send_event( Event::DBG_DISASM_READ );
}
// end of showEvent

//*******************************************************************
// closeEvent                                      PRIVATE inherited
//*******************************************************************
void DbgDisasm::closeEvent( QCloseEvent* const in_event )
{
	EventsController::instance()->send_event( Event::DBG_DISASM_STOP );
	DbgToolDialog::closeEvent( in_event );
}
// end of closeEvent

//*******************************************************************
// customEvent                                     PRIVATE inherited
//*******************************************************************
void DbgDisasm::customEvent( QEvent* const in_event )
{
	Event* const event = dynamic_cast< Event* >( in_event );
	const qint32 type = static_cast<qint32>( event->type() );
         
	switch( type ) {
		case Event::DBG_DISASM_DATA:
			parse_data( event );
			break;
	}
}
// end of customEvent

//*******************************************************************
// parse_data                                                PRIVATE
//*******************************************************************
void DbgDisasm::parse_data( Event* const in_event )
{
	const QStringList data = in_event->d_data[ 0 ].toStringList();
	d_table->get_prv_data();
	
	qint32 row = 0;
	QString text = "";
	bool ok = false;

	QStringList::const_iterator it = data.begin();
	const QStringList::const_iterator end = data.end();
	while( it != end ) {
		if( !ok ) {
			if( (*it).startsWith( ASM_START_MARKER ) ) {
				ok = true;
				++it;
				continue;
			}
		}
		else {
			if( (*it).contains( ERROR ) ) {
				break;
			}
			if( (*it).startsWith( ASM_END_MARKER ) ) {
				break;
			}
			if( (*it).startsWith( HEX_MARKER )  ) {
				if( !text.isEmpty() ) {
					set_item( row++, text.trimmed() );
					text = (*it);
				}
				else {
					text += (*it);
				}
			}
			else {
				text += (*it);
			}
		}
		++it;
	}
	if( ok ) set_item( row++, text.trimmed() );
	d_table->remove_rows( row );
}
// end of parse_data

//*******************************************************************
// set_item                                                  PRIVATE
//*******************************************************************
void DbgDisasm::set_item( const qint32 in_idx, const QString& in_text )
{
	if( !in_text.isEmpty() ) {
		const QString addr = in_text.section( ':', 0, 0 ).trimmed();
		const QString code = in_text.section( ':', 1 ).trimmed();
		const QString hex  = addr.section( ' ', 0, 0 ).trimmed();
		const QString off  = addr.section( ' ', 1 ).trimmed();
		const qint32 row   = ( in_idx < d_table->rowCount() ) ? in_idx : d_table->add_row( in_idx );
	
		d_table->set_item( row, 0, hex );
		d_table->set_item( row, 1, code.section( ' ', 0, 0 ).trimmed() );
		d_table->set_item( row, 2, code.section( ' ', 1 ).trimmed() );
		d_table->set_item( row, 3, off );
		d_table->resize_columns();
	}
}
// end of set_item
