/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : Debugger.cpp
 * DATE   : 04.02.2007
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "Debugger.h"
#include "EventsController.h"
#include "QBtParser.h"
#include "Breakpoints.h"
#include <signal.h>
#include <sys/types.h>
#include <QRegExp>
#include <QApplication>
#include <QtDebug>

/*------- local constants:
-------------------------------------------------------------------*/
const QChar   Debugger::NEW_LINE           = '\n';
const QChar   Debugger::SPACE              = ' ';
const QChar   Debugger::CTRL_Z             = 0x1A;
const QChar   Debugger::COLON              = ':';
const QChar   Debugger::COMMAS             = '\"';
const QString Debugger::PTR                = "*";
const QString Debugger::GDB_CALL           = "gdb -annotate 2 -d %1 %2";
const QString Debugger::GDB_RUN            = "run";
const QString Debugger::GDB_QUIT           = "quit";
const QString Debugger::STEP               = "step";
const QString Debugger::NEXT               = "next";
const QString Debugger::FINISH             = "finish";
const QString Debugger::SIG_INT            = "signal SIGINT";
const QString Debugger::PRINT              = "print %1";
const QString Debugger::CONTINUE           = "continue";
const QString Debugger::REGISTERS          = "info registers";
const QString Debugger::BACKTRACE          = "backtrace";
const QString Debugger::DISASSEMBLE        = "disassemble";
const QString Debugger::EXAMINE            = "x/%1ub %2";
const QString Debugger::SET_BREAKPOINT     = "break %1:%2";
const QString Debugger::CLR_BREAKPOINT     = "clear %1:%2";
const QString Debugger::PROMPT_MARKER      = "prompt";
const QString Debugger::SYMBOL_MARKER      = "Symbol ";
const QString Debugger::INFO_SCOPE         = "info scope %1";
const QString Debugger::SOURCE_MARKER      = "source ";
const QString Debugger::END_OF_PROGRAM     = "exited ";
const QString Debugger::ERROR_MARKER       = "error";
const QString Debugger::FUNCTION_MARKER    = "frame-function-name";
const QString Debugger::VALUE_MARKER       = "value-history-begin";
const QString Debugger::SCOPE_MARKER       = "Scope for ";
const QString Debugger::THREAD_MARKER      = "[New Thread";
const QString Debugger::PROG_START_MARKER  = "Starting program:";
const QString Debugger::HEX_MARKER         = "0x";
const QString Debugger::SIZE_MARKER        = "size";
const QString Debugger::DATA_MARKER        = "data";
const QString Debugger::STD_STRING_MARKER  = "_M_p";
const QString Debugger::VALUE_START_MARKER = "value-history-value";
const QString Debugger::VALUE_END_MARKER   = "value-history-end";
const QString Debugger::PROG_EXIT_MARKER   = "Program exited ";
const QString Debugger::TYPE_MARKER        = "type =";;
const QString Debugger::WHAT_IS            = "whatis %1";
const QString Debugger::PRINT_STRING       =  "x/%1s %2";
const QString Debugger::PRINT_PTR          = "print *%1";
const QString Debugger::PRINT_ARROW        = "print %1->d[0]";

//*******************************************************************
// Debugger                                              CONSTRUCTOR
//*******************************************************************
Debugger::Debugger()
: QProcess         ()
, d_cmd            ( None )
, d_program        ( "" )
, d_pid            ( 0 )
, d_started        ( false )
, d_current_file   ( "" )
, d_current_line   ( -1 )
, d_varname        ( "" )
, d_vartype        ( "" )
, d_registers      ( false )
, d_callstack      ( false )
, d_disasm         ( false )
, d_memory         ( false )
, d_next_stl_string( false )
, d_next_qt_string ( false )
, d_memory_addr    ( 0 )
, d_memory_size    ( 0 )
, d_busy_mutex     ( QMutex::Recursive )
{
   d_buffer.clear();
   
   EventsController::instance()->append( this, Event::DBG_STEP_INTO_REQUEST );
   EventsController::instance()->append( this, Event::DBG_STEP_OVER_REQUEST );
   EventsController::instance()->append( this, Event::DBG_STEP_OUT_REQUEST );
   EventsController::instance()->append( this, Event::DBG_GET_ITEM_TYPE );
   EventsController::instance()->append( this, Event::DBG_GET_ITEM_VALUE );
   EventsController::instance()->append( this, Event::DBG_PAUSE_REQUEST );
   EventsController::instance()->append( this, Event::DBG_CONTINUE_REQUEST );
   EventsController::instance()->append( this, Event::DBG_AFTER_DATA );
   EventsController::instance()->append( this, Event::DBG_REGISTERS_START );
   EventsController::instance()->append( this, Event::DBG_REGISTERS_STOP );
   EventsController::instance()->append( this, Event::DBG_REGISTERS_READ );
   EventsController::instance()->append( this, Event::DBG_CALLSTACK_START );
   EventsController::instance()->append( this, Event::DBG_CALLSTACK_STOP );
   EventsController::instance()->append( this, Event::DBG_CALLSTACK_READ );
   EventsController::instance()->append( this, Event::DBG_DISASM_START );
   EventsController::instance()->append( this, Event::DBG_DISASM_STOP );
   EventsController::instance()->append( this, Event::DBG_DISASM_READ );
   EventsController::instance()->append( this, Event::DBG_MEMORY_START );
   EventsController::instance()->append( this, Event::DBG_MEMORY_STOP );
   EventsController::instance()->append( this, Event::DBG_MEMORY_READ );
   EventsController::instance()->append( this, Event::BREAKPOINT_SET );
   EventsController::instance()->append( this, Event::BREAKPOINT_UNSET );
   
   connect( this, SIGNAL( readyReadStandardOutput() ), this, SLOT( read_stdout()         ));
   connect( this, SIGNAL( readyReadStandardError()  ), this, SLOT( read_stderr()         ));
   connect( this, SIGNAL( started()                 ), this, SLOT( process_started()     ));
   connect( this, SIGNAL( finished(int)             ), this, SLOT( process_finished(int) ));
}
// end of Debugger

//*******************************************************************
// ~Debugger                                              DESTRUCTOR
//*******************************************************************
Debugger::~Debugger()
{
   EventsController::instance()->remove( this );
}
// end of ~Debugger

//*******************************************************************
// customEvent                                     PRIVATE inherited
//*******************************************************************
void Debugger::customEvent( QEvent* const in_event )
{
   const Event* const event = dynamic_cast< Event* >( in_event );
   const qint32 type = static_cast< qint32 >( event->type() );
   
   d_varname.clear();
   switch( type ) {
      case Event::DBG_STEP_INTO_REQUEST:
         request( STEP );
         break;
      case Event::DBG_STEP_OVER_REQUEST:
         request( NEXT );
         break;
      case Event::DBG_STEP_OUT_REQUEST:
         execute_command( FINISH );
         break;
      case Event::DBG_PAUSE_REQUEST:
         execute_command( SIG_INT );
         break;
      case Event::DBG_CONTINUE_REQUEST:
         execute_command( CONTINUE );
         break;
      case Event::DBG_AFTER_DATA:
         after_data();
         break;
      case Event::DBG_GET_ITEM_TYPE:
         get_item_type( event );
         break;
      case Event::DBG_GET_ITEM_VALUE:
         get_item_value( event );
         break;
      case Event::DBG_CALLSTACK_START:
         d_callstack = true;
         break;
      case Event::DBG_CALLSTACK_STOP:
         d_callstack = false;
         break;
      case Event::DBG_CALLSTACK_READ:
         if( d_callstack ) callstack_request();
         break;
      case Event::DBG_REGISTERS_START:
         d_registers = true;
         break;
      case Event::DBG_REGISTERS_STOP:
         d_registers = false;
         break;
      case Event::DBG_REGISTERS_READ:
         if( d_registers ) registers_request();
         break;
      case Event::DBG_DISASM_START:
         d_disasm = true;
         break;
      case Event::DBG_DISASM_STOP:
         d_disasm = false;
         break;
      case Event::DBG_DISASM_READ:
         if( d_disasm ) disasm_request();
         break;
      case Event::DBG_MEMORY_START:
         d_memory = true;
         break;
      case Event::DBG_MEMORY_STOP:
         d_memory = false;
         break;
      case Event::DBG_MEMORY_READ:
         if( d_memory ) {
            d_memory_addr = event->d_data[ 0 ].toUInt();
            d_memory_size = event->d_data[ 1 ].toUInt();
            memory_request();
         }
         break;
      case Event::BREAKPOINT_SET:
         set_breakpoint( event );
         break;
      case Event::BREAKPOINT_UNSET:
         unset_breakpoint( event );
         break;
   }
}
// end of customEvent

//*******************************************************************
// request                                                   PRIVATE
//*******************************************************************
void Debugger::request( const QString& in_cmd )
{
   if( d_busy_mutex.tryLock() ) {
      EventsController::instance()->send_event( Event::DBG_DISABLE_ALL_REQUEST );
      QApplication::setOverrideCursor( Qt::BusyCursor );
      execute_command( in_cmd );
   }
}
// end of request

//*******************************************************************
// callstack_request                                         PRIVATE
//*******************************************************************
void Debugger::callstack_request()
{
   d_cmd = CallStack;
   execute_command( BACKTRACE );
}
// end of callstack_request

//*******************************************************************
// registers_request                                         PRIVATE
//*******************************************************************
void Debugger::registers_request()
{
   d_cmd = Registers;
   execute_command( REGISTERS );
}
// end of registers_request

//*******************************************************************
// disasm_request                                            PRIVATE
//*******************************************************************
void Debugger::disasm_request()
{
   d_cmd = DisAsm;
   execute_command( DISASSEMBLE );
}
// end of disasm_request

//*******************************************************************
// memory_request                                             PRIVATE
//*******************************************************************
void Debugger::memory_request()
{
   d_cmd = Memory;
   execute_command( EXAMINE.arg( d_memory_addr ).arg( d_memory_size ) );
}
// end of memory_request

//*******************************************************************
// after_data                                                PRIVATE
//*******************************************************************
void Debugger::after_data()
{
        if( d_callstack ) callstack_request();
   else if( d_registers ) registers_request();
   else if( d_disasm    ) disasm_request();
   else if( d_memory    ) memory_request();
   else                   set_active();
}
// end of after_data

//*******************************************************************
// parse_callstack                                           PRIVATE
//*******************************************************************
void Debugger::parse_callstack( const QStringList& in_data )
{
   EventsController::instance()->send_event( Event::DBG_CALLSTACK_DATA, in_data );
   d_cmd = None;

        if( d_registers ) registers_request();
   else if( d_disasm    ) disasm_request();
   else if( d_memory    ) memory_request();
   else                   set_active();
}
// end of parse_callstack

//*******************************************************************
// parse_registers                                           PRIVATE
//*******************************************************************
void Debugger::parse_registers( const QStringList& in_data )
{
   EventsController::instance()->send_event( Event::DBG_REGISTERS_DATA, in_data );
   d_cmd = None;

        if( d_disasm ) disasm_request();
   else if( d_memory ) memory_request();
   else                set_active();
}
// end of parse_registers

//*******************************************************************
// parse_disasm                                              PRIVATE
//*******************************************************************
void Debugger::parse_disasm( const QStringList& in_data )
{
   EventsController::instance()->send_event( Event::DBG_DISASM_DATA, in_data );
   d_cmd = None;

   if( d_memory ) memory_request();
   else           set_active();
}
// end of parse_disasm

//*******************************************************************
// parse_memory                                              PRIVATE
//*******************************************************************
void Debugger::parse_memory( const QStringList& in_data )
{
   EventsController::instance()->send_event( Event::DBG_MEMORY_DATA, in_data );
   d_cmd = None;
   set_active();
}
// end of parse_memory

//*******************************************************************
// set_active                                                PRIVATE
//-------------------------------------------------------------------
// Przywrocenie mozliwosci dalszej pracy z debuggerem.
// Byla blokada na czas odczytu wszystkich potrzebnych informacji.
//*******************************************************************
void Debugger::set_active()
{
   d_busy_mutex.unlock();
   EventsController::instance()->send_event( Event::DBG_DISABLE_RUN_REQUEST );
   QApplication::restoreOverrideCursor();
}
// end of set_active

//*******************************************************************
// start                                                      PUBLIC
//-------------------------------------------------------------------
// Uruchomienie sesji debuggera.
//*******************************************************************
void Debugger::start( const QString& in_program, const QString& in_wrkdir )
{
   if( !d_started ) {
      d_program = in_program;
      setWorkingDirectory( in_wrkdir );
      QProcess::start( GDB_CALL.arg( in_wrkdir ).arg( in_program ) );
   }
}
// end of start

//*******************************************************************
// stop                                                       PUBLIC
//-------------------------------------------------------------------
// Zakonczenie sesji debuggera.
//*******************************************************************
void Debugger::stop()
{
   if( d_started ) {
      if( d_pid ) {
         ::kill( d_pid, SIGKILL );
         d_pid = 0;
      }
      execute_command( GDB_QUIT );
   }
}
// end of stop

//*******************************************************************
// execute_command                                           PRIVATE
//-------------------------------------------------------------------
// Uzupelnienie komendy o znak konca linii i wyslanie jej do gdb.
//*******************************************************************
void Debugger::execute_command( const QString& in_command )
{
   write( ( in_command + NEW_LINE ).toLatin1() );
}
// end of execute_command

//*******************************************************************
// get_pid                                                   PRIVATE
//-------------------------------------------------------------------
// Odczyt numeru pid uruchomiego programu.
// Potrzebne nam to do pozniejszego jego zabicia.
//*******************************************************************
qint32 Debugger::get_pid( const QString& in_text )
{
   qint32 retval = 0;

   const qint32 sidx = in_text.indexOf( '(' );
   if( sidx != -1 ) {
      const qint32 eidx = in_text.indexOf( ')', sidx + 1 );
      if( eidx != -1 ) {
         const QStringList tokens = in_text.mid( sidx + 1, eidx - sidx - 1 ).split( SPACE );
         retval = tokens.last().toInt();
      }
   }
   return retval;
}
// end of get_pid

//*******************************************************************
// process_started                                      PRIVATE slot
//-------------------------------------------------------------------
// Slot informujacy o uruchomieniu debugger'a.
// Na poczatku wykonujemy zawsze trzy czynnosci:
// 1. okreslamy sposob formatowania danych zwracanych przez debugger,
// 2. przekazujemy do debugger'a informacje o breakpoint'ach,
// 3. startujemy wykonanie programu debuggowanego.
//*******************************************************************
void Debugger::process_started()
{
   EventsController::instance()->send_event( Event::DBG_STARTED );

   execute_command( "set width 0" );
   execute_command( "set height 0" );
   execute_command( "set print address on" );
   execute_command( "set print symbol-filename on" );
   execute_command( "set print array on" );

   execute_command( "handle SIGINT stop print" );
   apply_breakpoints();
   execute_command( GDB_RUN );
}
// end of process_started

//*******************************************************************
// apply_breakponts                                          PRIVATE
//-------------------------------------------------------------------
// Zbiorcze ustawienie wszystkich breakpointow ustawionych we
// wszystkich plikach zrodlowych aktualnego projektu.
//*******************************************************************
void Debugger::apply_breakpoints()
{
   const QStringList data = Breakpoints::instance()->data().keys();
   QStringList::const_iterator it_data = data.begin();
   const QStringList::const_iterator end_data = data.end();
   while( it_data != end_data ) {
      QList< qint32 > pts = Breakpoints::instance()->data()[ *it_data ].keys();
      QList< qint32 >::const_iterator it_pts = pts.begin();
      const QList< qint32 >::const_iterator end_pts = pts.end();
      while( it_pts != end_pts ) {
         execute_command( SET_BREAKPOINT.arg( *it_data ).arg( *it_pts + 1 ) );
         ++it_pts;
      }
      ++it_data;
   }
}
// end of apply_breakpoints

//*******************************************************************
// set_breakpoint                                            PRIVATE
//-------------------------------------------------------------------
// Ustawienie nowego breakpointu.
// Breakpoint okreslony jest przez nazwe pliku i numer linii.
// Obie te informacje sa przyslane w evencie.
//*******************************************************************
void Debugger::set_breakpoint( const Event* const in_event )
{
   execute_command( SET_BREAKPOINT.arg( in_event->d_data[ 0 ].toString() ).arg( in_event->d_data[ 1 ].toInt() + 1 ) );
}
// end of set_breakpoint

//*******************************************************************
// unset_breakpoint                                          PRIVATE
//-------------------------------------------------------------------
// Likwidacja istniejacego breakpointu.
// Breakpoint okreslony jest przez nazwe pliku i numer linii.
// Obie te informacje sa przyslane w evencie.
//*******************************************************************
void Debugger::unset_breakpoint( const Event* const in_event )
{
   execute_command( CLR_BREAKPOINT.arg( in_event->d_data[ 0 ].toString() ).arg( in_event->d_data[ 1 ].toInt() + 1 ) );
}
// end of unset_breakpoint

//*******************************************************************
// get_item_type                                             PRIVATE
//-------------------------------------------------------------------
// Wyslanie zapytania o tym zmiennej okreslonej przez przyslana
// nazwe.
//*******************************************************************
void Debugger::get_item_type( const Event* const in_event )
{
   execute_command( WHAT_IS.arg( in_event->d_data.at( 0 ).toString() ) );
}
// end of get_item_type

//*******************************************************************
// get_item_value                                            PRIVATE
//-------------------------------------------------------------------
// Wyslanie zapytania o wartosc zmiennej okreslonej przez przyslana
// nazwe. Przyslany typ zmiennej przechowujemy na potrzeby analizy
// zmiennych strukturalnych (QString i std::string).
//*******************************************************************
void Debugger::get_item_value( const Event* const in_event )
{
   d_varname = in_event->d_data.at( 0 ).toString();
   d_vartype = in_event->d_data.at( 1 ).toString();
   execute_command( PRINT.arg( d_varname ) );
}
// end of get_item_value

//*******************************************************************
// process_finished                                     PRIVATE slot
//-------------------------------------------------------------------
// Slot informujacy o zakoczeniu pracy debugger'a.
//*******************************************************************
void Debugger::process_finished( int )
{
   d_cmd = None;
   d_program = "";
   d_pid = 0;
   d_started = false;
   d_current_file = "";
   d_current_line = -1;
   d_registers = false;
   d_callstack = false;
   d_disasm = false;

   EventsController::instance()->send_event( Event::DBG_FINISHED );
}
// end of process_finished 

//*******************************************************************
// read_stdout                                          PRIVATE slot
//*******************************************************************
void Debugger::read_stdout()
{
   const QString str = QString::fromLocal8Bit( readAllStandardOutput() ).remove( CTRL_Z );
   d_output += str.split( NEW_LINE, QString::SkipEmptyParts );
   
   QStringList tmp;
   QString text;
   bool is_scope = false;
   
   d_buffer.clear();
   tmp.clear();

   QStringList::const_iterator it = d_output.begin();
   const QStringList::const_iterator end = d_output.end();
   while( it != end ) {
      text = (*it).simplified().trimmed();
      if( !is_scope ) {
         if( text == "post-prompt" ) {
            is_scope = true;
            if( is_scope ) tmp << text;
         }
      }
      else {
         if( text == "pre-prompt" ) {
            if( !tmp.empty() ) {
               tmp << text;
               tmp.removeFirst();
               tmp.removeLast();
               d_buffer << tmp;
               tmp.clear();
               is_scope = false;
            }
         }
         else {
            tmp << text;
         }
      }
      ++it;
   }
   d_output.clear();
   if( is_scope ) d_output << tmp;
   parse_stdout();
}
// end of read_stdout

//*******************************************************************
// parse_scope                                               PRIVATE
//-------------------------------------------------------------------
// Informacje o nazwach zmiennych zawartych w aktualnej funkcji.
//*******************************************************************
void Debugger::parse_scope( const QStringList& in_data )
{
   QStringList vars;
   QStringList::const_iterator it = in_data.begin();
   const QStringList::const_iterator end = in_data.end();
   while( it != end ) {
      if( (*it).startsWith( SYMBOL_MARKER ) ) {
         vars << (*it).section( SPACE, 1, 1, QString::SectionSkipEmpty );
      }
      ++it;
   }

   EventsController::instance()->send_event( Event::DBG_SCOPE, vars );
}
// end of parse_scope

//*******************************************************************
// parse_value                                               PRIVATE
//*******************************************************************
void Debugger::parse_value( const QStringList& in_data )
{
   QStringList lines;

   if( in_data.indexOf( VALUE_START_MARKER ) != -1 ) {
      // Wyodrebienie linii z wartosciami (bez naglowkow).
      bool add = false;
      QStringList::const_iterator it = in_data.begin();
      const QStringList::const_iterator end = in_data.end();
      while( it != end ) {
         if( add ) {
            if( VALUE_END_MARKER == (*it) ) {
               break;
            }
            else {
               lines << (*it);
            }
         }
         else {
            add = ( VALUE_START_MARKER == (*it) );
         }
         ++it;
      }
   }
   else {
      lines = in_data;
   }

   //print_data( lines );

   if( d_next_qt_string ) {
      QString text;
      const qint32 retval = get_qt_string_value( lines, text );
      if( 1 == retval ) {
         d_next_qt_string = false;
         lines.clear();
         lines << ( COMMAS + text + COMMAS );
         EventsController::instance()->send_event( Event::DBG_ITEM_VALUE, lines );
      }
      else if( -1 == retval ) {
         d_next_qt_string = false;
         EventsController::instance()->send_event( Event::DBG_ERROR );
      }
   }
   else {
      QString text   = lines.first();
      QString buffer = "";
      bool    send   = true;

      // ******* CZY TO string (STL) ? *******
      QRegExp regexp( "\\bstring\\b\\s*([\\&|\\*]*)" );
      if( d_vartype.indexOf( regexp ) != -1 ) {
         const QString& cap = regexp.cap( 1 );
         if( PTR == cap ) {
            send = false;
            d_vartype = d_vartype.remove( PTR ).trimmed();
            execute_command( PRINT_PTR.arg( d_varname ) );
         }
         else {
            if( get_std_string_value( lines, buffer ) ) {
               lines.clear();
               lines << buffer;
            }
            send = true;
         }
      }
      else {
         // ******* CZY TO QString ? *******
         regexp.setPattern( "\\bQString\\b\\s*([\\&|\\*]*)" );
         if( d_vartype.indexOf( regexp ) != -1 ) {
            const QString& cap = regexp.cap( 1 );
            if( PTR == cap ) {
               send = false;
               d_vartype = d_vartype.remove( PTR ).trimmed();
               execute_command( PRINT_PTR.arg( d_varname ) );
            }
            else {
               send = false;
               d_next_qt_string = true;
               execute_command( PRINT_ARROW.arg( d_varname ) );
            }
         }
      }

      if( send ) {
         EventsController::instance()->send_event( Event::DBG_ITEM_VALUE, lines );
      }
   }
}
// end of parse_value

//*******************************************************************
// parse_source_line                                         PRIVATE
//-------------------------------------------------------------------
// Przyslano jedna linie z informacje o pozycji w pliku zrodlowym.
//*******************************************************************
void Debugger::parse_source_line( const QString& in_text )
{
   const qint32 idx = in_text.indexOf( SPACE );
   if( idx != -1 ) {
      const QString info = in_text.mid( idx ).trimmed();
      d_current_file = info.section( COLON, 0, 0 );
      d_current_line = info.section( COLON, 1, 1 ).toInt();
      EventsController::instance()->send_event( Event::SELECTED_FILE_AND_ROW, d_current_file, d_current_line );
      execute_command( INFO_SCOPE.arg( d_current_line ) );
   }
}
// end of parse_source_line

//*******************************************************************
// parse_stdout                                              PRIVATE
//*******************************************************************
void Debugger::parse_stdout()
{
   if( !d_buffer.empty() ) {
      QVector< QStringList >::const_iterator vit = d_buffer.begin();
      const QVector< QStringList >::const_iterator vend = d_buffer.end();
      while( vit != vend ) {
//         print_data( *vit );
         QStringList::const_iterator it = (*vit).begin();
         const QStringList::const_iterator end = (*vit).end();
         while( it != end ) {
            if( d_started ) {
               if( d_next_qt_string ) {
                  parse_value( *vit );
                  break;
               }
               else if( Registers == d_cmd ) {
                  parse_registers( *vit );
                  break;
               }
               else if( CallStack == d_cmd ) {
                  parse_callstack( *vit );
                  break;
               }
               else if( DisAsm == d_cmd ) {
                  parse_disasm( *vit );
                  break;
               }
               else if( Memory == d_cmd ) {
                  parse_memory( *vit );
                  break;
               }
               else {
                  if( (*it).startsWith( TYPE_MARKER ) ) {
                     const QString text = (*it).section( '=', 1 ).trimmed();
                     EventsController::instance()->send_event( Event::DBG_ITEM_TYPE, text );
                  }
                  else if( (*it).startsWith( SOURCE_MARKER ) ) {
                     parse_source_line( *it );
                  }
                  else if( (*it ).startsWith( END_OF_PROGRAM ) ) {
                     EventsController::instance()->send_event( Event::DBG_STOP_REQUEST );
                     break;
                  }
                  else if( ERROR_MARKER == (*it) ) {
                     EventsController::instance()->send_event( Event::DBG_ERROR );
                  }
                  else if( FUNCTION_MARKER == (*it) ) {
                     ++it;
                     d_current_function = (*it);
                  }
                  else if( (*it).startsWith( VALUE_MARKER ) ) {
                     parse_value( *vit );
                     break;
                  }
                  else if( (*it).startsWith( SCOPE_MARKER ) ) {
                     parse_scope( *vit );
                     break;
                  }
                  else if( (*it).startsWith( PROG_EXIT_MARKER ) ) {
                     EventsController::instance()->send_event( Event::DBG_STOP_REQUEST );
                     break;
                  }
                  else {
                     set_active();
                  }
               
                  if( 0 == d_pid ) {
                     if( (*it).startsWith( THREAD_MARKER ) ) {
                        d_pid = get_pid( *it );
                     }
                  }
                  /*
                  else if( (*it).startsWith( "Breakpoint " ) ) {
                  }
                  else if( (*it).startsWith( "Deleted breakpoint" ) ) {
                  }
                  */
               }
            }
            else {
               d_started = (*it).startsWith( PROG_START_MARKER );
            }
            ++it;
         }
         ++vit;
      }
   }
}
// end of parse_stdout

//*******************************************************************
// get_std_string_value                                      PRIVATE
//*******************************************************************
bool Debugger::get_std_string_value( const QStringList& in_data, QString& in_text )
{
   in_text.clear();

   const qint32 nr = in_data.indexOf( STD_STRING_MARKER );
   if( nr != -1 ) {
      const QString& buffer = in_data[ nr + 4 ];

      const qint32 idx1 = buffer.indexOf( COMMAS );
      if( idx1 != -1 ) {
         const qint32 idx2 = buffer.indexOf( COMMAS, idx1 + 1 );
         if( idx2 != -1 ) {
            in_text = buffer.mid( idx1, idx2 - idx1 + 1 );
         }
      }
   }
   return ( in_text.isEmpty() == false );
}
// end of get_std_string_value

//*******************************************************************
// get_qt_string_value                                       PRIVATE
//*******************************************************************
qint32 Debugger::get_qt_string_value( const QStringList& in_data, QString& in_text )
{
   qint32 retval = -1;

   in_text.clear();

   if( in_data.first().startsWith( HEX_MARKER ) ) {
      retval = 1;
   
      QStringList::const_iterator it = in_data.begin();
      const QStringList::const_iterator end = in_data.end();
      while( it != end ) {
         const qint32 idx1 = (*it).indexOf( COMMAS );
         const qint32 idx2 = (*it).indexOf( COMMAS, idx1 + 1 );
         in_text += (*it).mid( idx1 + 1, idx2 - idx1 - 1 );
         ++it;
      }
   }
   else {
      QString size = "";
      QString addr = "";
      qint32  nr1  = -1;
      qint32  nr2  = -1;

      if( ( nr1 = in_data.indexOf( SIZE_MARKER ) ) != -1 ) {
         size = in_data[ nr1 + 4 ];
         if( ( nr2 = in_data.indexOf( DATA_MARKER, nr1 + 1 ) ) != -1 ) {
            addr = in_data[ nr2 + 4 ];
         }
      }
   
      if( ( nr1 != -1 ) && ( nr2 != -1 ) ) {
         retval = 0;
         execute_command( PRINT_STRING.arg( size ).arg( addr ) );
      }
   }
   return retval;
}
// end of get_std_string_value

//*******************************************************************
// read_stderr                                          PRIVATE slot
//*******************************************************************
void Debugger::read_stderr()
{
   const QString text = QString::fromLocal8Bit( readAllStandardError() );
   //qDebug() << text;
}
// end of read_stderr

/*
void Debugger::print_data( const QStringList& in_data )
{
      qint32 i = 0;
   
      qDebug() << "+++++++++++++++++++++++++++++++++++++++++++++++++++";
      QStringList::const_iterator it = in_data.begin();
      const QStringList::const_iterator end = in_data.end();
      while( it != end ) {
         qDebug() << ++i << ") " << *it;
         ++it;
      }
}
void Debugger::analyze_value( const QStringList& in_data )
{
   QStringList::const_iterator it = in_data.begin();
   qDebug() << "{";
   print_fields( 1, in_data, it );
   qDebug() << "}";
   qDebug() << " ";
}
// end of analyze_value

void Debugger::print_fields ( const qint32 in_level, const QStringList& in_data, QStringList::const_iterator& in_it )
{
   enum {
      None = 0,
      Begin,
      NameEnd,
      Value,
      End,
      Array,
   };

   QString text = "";
   qint32 stat = None;

   while( in_it != in_data.end() ) {
      if( *in_it == "{" ) {
         ++in_it;
         qDebug() << text;

         text = "";
         text = text.fill( '\t', in_level );
         text += "{";
         qDebug() << text;

         text.clear();
         print_fields( in_level + 1, in_data, in_it );


         text = text.fill( '\t', in_level );
         text += "}";
         qDebug() << text;
         stat = None;
         text.clear();
      }
      else if( *in_it == "}" ) {
         qDebug() << text;
         ++in_it;
         break;
      }
      if( (*in_it).startsWith( "field-begin" ) ) {
         stat = Begin;
      }
      else if( *in_it == "field-name-end" ) {
         stat = NameEnd;
      }
      else if( *in_it == "field-value" ) {
         stat = Value;
      }
      else if( *in_it == "field-end" ) {
         stat = End;
      }
      else if( (*in_it).startsWith( "array-section-begin" ) ) {
         stat = Array;
      }
      else if( (*in_it).startsWith( "array-section-end" ) ) {
         stat = None;
      }
      else if( Begin == stat ) {
         text = "";
         text = text.fill( '\t', in_level );
         text += *in_it;
      }
      else if( NameEnd == stat ) {
         text += " =";
      }
      else if( Value == stat ) {
         text += ( " " + (*in_it) ); 
      }
      else if( End == stat ) {
         stat = None;
         qDebug() << text;
      }
      else if( Array == stat ) {
         text = "";
         text = text.fill( '\t', in_level );
         text += *in_it;
      }
      ++in_it;
   }
}
*/
