/*******************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : Editor.cpp
 * DATE   : 3.09.2006
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "Editor.h"
#include "Config.h"
#include "QBtWorkspace.h"
#include "Margin.h"
#include "Folder.h"
#include "EventsController.h"
#include "CppSyntax.h"
#include "Classer.h"
#include "QBtParser.h"
#include "Renumer.h"
#include "Settings.h"
#include "GotoLine.h"
#include "QBtFind.h"
#include "Bookmarks.h"
#include "Breakpoints.h"
#include "QBTParserFunDef.h"
#include "QBTFunctionsBox.h"
#include "QBtPrinter.h"
#include <QDir>
#include <QFile>
#include <QFont>
#include <QFileInfo>
#include <QClipboard>
#include <QTextStream>
#include <QMessageBox>
#include <QCloseEvent>
#include <QFileDialog>
#include <QTextDocument>
#include <QApplication>
#include <QFontMetrics>
#include <QTextBlock>
#include <QTextCursor>
#include <QRegExp>
#include <QEvent>
#include <QPrinter>
#include <QPrintDialog>
#include <QTextLayout>
#include <QScrollBar>
#include <QPainter>
#include <QTimer>
#include <QMenu>
#include <QAction>
#include <QAbstractTextDocumentLayout>
#include <QStack>
#include <QtDebug>
#include <memory>
using namespace std;

/*------- local constants:
-------------------------------------------------------------------*/
const char* const Editor::NO_NAME         = QT_TR_NOOP( "(%1). noname.txt" );
const char* const Editor::CHANGE_MARK     = "*";
const char* const Editor::SAVING          = QT_TR_NOOP( "Saving the file" );
const char* const Editor::OPENING         = QT_TR_NOOP( "Open a file" );
const char* const Editor::FIND            = QT_TR_NOOP( "Find" );
const char* const Editor::NOT_FOUND       = QT_TR_NOOP( "Text not found" );
const char* const Editor::DOC_CHANGED_MSG = QT_TR_NOOP( "This document has been modified.\nDo you want to save your changes?" );
const char* const Editor::FILE_FILTERS    = QT_TR_NOOP( "Source files (*.cpp);;Header files (*.h);;Project files (*.pro);;Resources (*.qrc);;All files (*)" );
const char* const Editor::SAVE_ERROR_MSG  = QT_TR_NOOP( "Can't save the file: %1\n%2" );
const char* const Editor::READ_ERROR_MSG  = QT_TR_NOOP( "Can't read the file: %1\n%2" );
const char* const Editor::CHKDT_CAPTION   = QT_TR_NOOP( "File changed" );
const char* const Editor::CHKDT_MESSAGE   = QT_TR_NOOP( "This file has been changed by other application.\nWould you like to reload the file?" );
const char* const Editor::DEFAULT_FONT    = "Monospace,10,-1,5,50,0,0,0,0,0";


//*******************************************************************
// Editor                                                CONSTRUCTOR
//*******************************************************************
Editor::Editor( QWidget* const in_parent )
: TextEdit        ( in_parent )
, d_workspace     ( dynamic_cast< QBtWorkspace* >( in_parent ))
, margin_         ( new Margin( this ) )
, folder_         ( 0 )
, d_is_noname     ( true )
, d_file_path     ( "" )
, d_send_events   ( true )
, d_can_parse     ( false )
, d_syntax        ( 0 )
, d_find_flags    ( 0 )
, d_find_text     ( "" )
, d_timer         ( new QTimer( this ) )
, use_folder_     ( true )
{

   
   setAttribute( Qt::WA_DeleteOnClose );
   setAutoFillBackground( true );

   if( use_folder_ ) {
      folder_ = new Folder( this );
      connect( document()->documentLayout(), SIGNAL( update( const QRectF& ) ), folder_, SLOT( update() ) );
      connect( verticalScrollBar(), SIGNAL( valueChanged( int ) ), folder_, SLOT( update() ) );
   }
      
   set_properties();
   setLineWrapMode( QTextEdit::NoWrap );
   
   connect( d_timer   , SIGNAL( timeout() ), this, SLOT( check_datetime() ) );

   connect( document(), SIGNAL( contentsChanged()       ), this, SLOT( contents_changed()        ));
   connect( this      , SIGNAL( copyAvailable( bool )   ), this, SLOT( copy_available( bool )    ));
   connect( this      , SIGNAL( redoAvailable( bool )   ), this, SLOT( redo_available( bool )    ));
   connect( this      , SIGNAL( undoAvailable( bool )   ), this, SLOT( undo_available( bool )    ));
   connect( this      , SIGNAL( cursorPositionChanged() ), this, SLOT( cursor_position_changed() ));
   
   connect( document()->documentLayout(), SIGNAL( update( const QRectF& ) ), margin_, SLOT( update() ) );
   connect( verticalScrollBar()         , SIGNAL( valueChanged( int ) )    , margin_, SLOT( update() ) );
   connect( horizontalScrollBar()       , SIGNAL( valueChanged( int ) )    , this   , SLOT( update() ) );

   EventsController::instance()->append( this, Event::UNDO_REQUEST );
   EventsController::instance()->append( this, Event::REDO_REQUEST );
   EventsController::instance()->append( this, Event::COPY_REQUEST );
   EventsController::instance()->append( this, Event::CUT_REQUEST );
   EventsController::instance()->append( this, Event::PASTE_REQUEST );
   EventsController::instance()->append( this, Event::DELETE_REQUEST );
   EventsController::instance()->append( this, Event::SELECT_ALL_REQUEST );
   EventsController::instance()->append( this, Event::GOTO_LINE_REQUEST );
   EventsController::instance()->append( this, Event::FIND_REQUEST );
   EventsController::instance()->append( this, Event::FIND_NEXT_REQUEST );
   EventsController::instance()->append( this, Event::FIND_PREVIOUS_REQUEST );
   EventsController::instance()->append( this, Event::REPLACE_REQUEST );
   EventsController::instance()->append( this, Event::FONT_CHANGED );
   EventsController::instance()->append( this, Event::FILE_IN_PROJECT_RENAMED );
   EventsController::instance()->append( this, Event::BOOKMARK_NEXT_REQUEST );
   EventsController::instance()->append( this, Event::BOOKMARK_PREV_REQUEST );
   EventsController::instance()->append( this, Event::BOOKMARK_TOGGLE_REQUEST );
   EventsController::instance()->append( this, Event::BOOKMARK_ALL_REQUEST );
   EventsController::instance()->append( this, Event::BOOKMARK_CLEAR_ALL_REQUEST );
   EventsController::instance()->append( this, Event::DBG_BREAKPOINT_REQUEST );
   EventsController::instance()->append( this, Event::DBG_BREAKPOINT_CLEAR_ALL_REQUEST );

   d_timer->start( 1000 );
}
// end of Editor

//*******************************************************************
// ~Editor                                               CONSTRUCTOR
//*******************************************************************
Editor::~Editor()
{
   EventsController::instance()->remove( this );

   if( d_syntax ) {
      delete d_syntax;
      d_syntax = 0;
   }
   if( d_timer ) {
      d_timer->stop();
      delete d_timer;
   }
}

// end of ~Editor

//*******************************************************************
// set_properties                                            PRIVATE
//*******************************************************************
void Editor::set_properties()
{
   const QString fext = Shared::extension( d_file_path );

   QRgb qrgb = qRgb( 255, 255, 255 );
   if     ( Shared::is_src_ext( fext ) ) qrgb = Config::instance()->cpp_background();
   else if( Shared::is_hdr_ext( fext ) ) qrgb = Config::instance()->hpp_background();

   {  // kolor biurka, czyli widgetu jako takiego
      QPalette p;
      p.setColor( backgroundRole(), QColor::fromRgba( qrgb ) );
      //p.setColor( foregroundRole(), QColor::fromRgba( qrgb ) );
      setPalette( p );
   }
   
   QPalette palette;
   palette.setColor( viewport()->backgroundRole(), QColor::fromRgba( qrgb ) );
   viewport()->setPalette( palette );

   // Ustawienie nowego fontu.
   QFont fnt;
   fnt.fromString( Config::instance()->syntax()[ Config::Text ][ 0 ].toString() );
   setFont( fnt );

   // Ustawinie nowego rozmiaru tabulatora.
   setTabStopWidth( Config::instance()->tab_width() * fontMetrics().width( "X" ) );

   // Syntaktyka.
   if( ( Shared::is_src_ext( fext ) || Shared::is_hdr_ext( fext ) ) ) {
      if( d_syntax ) {
         d_syntax->rehighlight();
      }
      else {
         d_syntax = new CppSyntax( document() );
      }
      d_can_parse = true;
   }
   else {
      if( d_syntax ) {
         delete d_syntax;
         d_syntax = 0;
      }
      d_can_parse = false;
   }

   // Margines (rozmiar ikony musi byc dopasowany do wielkosci fontu).
   set_margin();
}
// end of set_properties

//*******************************************************************
// set_margin                                                 PUBLIC
//*******************************************************************
void Editor::set_margin()
{
   margin_->set_properties();
   if( use_folder_ ) folder_->set_properties();
   update_margins();
}
// end of set_margin

//*******************************************************************
// update_margins                                             PUBLIC
//*******************************************************************
void Editor::update_margins()
{
   margin_->update_geometry();
   int mw = margin_->width();
   if( use_folder_ ) {
      folder_->update_geometry( mw );
      folder_->parse();
      mw += folder_->width();
   }

   setViewportMargins( mw, 0, 0, 0 );
   margin_->update();
}
// end of update_margins

//*******************************************************************
// create_new_file                                            PUBLIC
//-------------------------------------------------------------------
// Utworzenie nowego, nienazwanego pliku.
//********************************************************************
void Editor::new_file()
{
   d_is_noname = TRUE;
   const QWidgetList windows = d_workspace->windowList();
   d_file_path = tr( NO_NAME ).arg( windows.count() + 1 );
   set_window_title( d_file_path );
   d_workspace->addWindow( this );
   Renumer::run( this );
   set_properties();
}
// end of create_new_file


//###################################################################
//#                                                                 #
//#                          P R I V A T E                          #
//#                                                                 #
//###################################################################


//*******************************************************************
// sizeHint                                        PRIVATE inherited
//*******************************************************************
QSize Editor::sizeHint() const
{
   qint32 const width  = static_cast<int>( parentWidget()->width() * 0.9 );
   qint32 const height = static_cast<int>( parentWidget()->height() * 0.9 );
   return QSize( width, height );
}
// end of sizeHint

//*******************************************************************
// focusInEvent                                    PRIVATE inherited
//*******************************************************************
void Editor::focusInEvent( QFocusEvent* const in_event )
{
   QTextEdit::focusInEvent( in_event );
   EventsController::instance()->send_event( Event::ACTIVE_DOCUMENT_CHANGED, d_file_path );
   cursor_position_changed();
}
// end of focusInEvent

//*******************************************************************
// resizeEvent                                     PRIVATE inherited
//*******************************************************************
void Editor::resizeEvent( QResizeEvent* const in_event )
{
   QTextEdit::resizeEvent( in_event );
   update_margins();
}
// end of resizeEvent

//*******************************************************************
// showEvent                                       PRIVATE inherited
//-------------------------------------------------------------------
// Zmiana tytulu okna: pelna sciezka do pliku.
//*******************************************************************
void Editor::showEvent( QShowEvent* const in_event )
{
   QTextEdit::showEvent( in_event );
   set_window_title( d_file_path );
   redo_available( document()->isRedoAvailable() );
   undo_available( document()->isUndoAvailable() );
}
// end of showEvent

//*******************************************************************
// hideEvent                                       PRIVATE inherited
//-------------------------------------------------------------------
// Zmiana tytulu okna: tylko nazwa pliku (ikona jest mala).
//*******************************************************************
void Editor::hideEvent( QHideEvent* const in_event )
{
   QTextEdit::hideEvent( in_event );
   set_window_title( d_file_path );
   redo_available( false );
   undo_available( false );

   // Wysylamy informacje do workspace, zeby przekazal fokus
   // do innego, otwartego okna.
   EventsController::instance()->send_event( Event::NEXT_WINDOW_FOCUS );
}
// end of hideEvent

//*******************************************************************
// set_window_title                                          PRIVATE
//*******************************************************************
void Editor::set_window_title( const QString& in_file_path )
{
   const QFileInfo fi( in_file_path );
   QString title = ( document()->isModified() ) ? CHANGE_MARK : "";
   title += ( windowState() & Qt::WindowMinimized ) ? fi.fileName() : fi.absoluteFilePath();
   setWindowTitle( title );
}
// end of set_window_title

//*******************************************************************
// closeEvent                                      PRIVATE inherited
//*******************************************************************
void Editor::closeEvent( QCloseEvent* const in_event )
{
   if( try_save() ) {
      if( d_send_events ) {
         EventsController::instance()->send_event( Event::FILE_SELF_CLOSED, d_file_path );
         EventsController::instance()->send_event( Event::CURSOR_POSITION_CHANGED, -1, -1 );
         redo_available( false );
         undo_available( false );
      }
      EventsController::instance()->remove( this );
      in_event->accept();
   }
   else {
      in_event->ignore();
   }
}
// end of closeEvent

//*******************************************************************
// customEvent                                     PRIVATE inherited
//*******************************************************************
void Editor::customEvent( QEvent* const in_event )
{
   Event* const event = dynamic_cast< Event* >( in_event );
   const qint32 type = static_cast<int>( event->type() );

   if( hasFocus() ) {
      switch( type ) {
         case Event::UNDO_REQUEST:
            do_undo();
            break;
         case Event::REDO_REQUEST:
            do_redo();
            break;
         case Event::COPY_REQUEST:
            do_copy();
            break;
         case Event::CUT_REQUEST:
            do_cut();
            break;
         case Event::PASTE_REQUEST:
            do_paste();
            break;
         case Event::DELETE_REQUEST:
            do_delete();
            break;
         case Event::SELECT_ALL_REQUEST:
            do_select_all();
            break;
         case Event::GOTO_LINE_REQUEST:
            goto_line();
            break;
         case Event::FIND_REQUEST:
            find_in_document();
            break;
         case Event::FIND_NEXT_REQUEST:
            find_next();
            break;
         case Event::FIND_PREVIOUS_REQUEST:
            find_previous();
            break;
         case Event::REPLACE_REQUEST:
            replace();
            break;
         case Event::BOOKMARK_NEXT_REQUEST:
            bookmark_next();
            break;
         case Event::BOOKMARK_PREV_REQUEST:
            bookmark_prev();
            break;
         case Event::BOOKMARK_TOGGLE_REQUEST:
            bookmarking();
            break;
         case Event::DBG_BREAKPOINT_REQUEST:
            breakpointing();
            break;
      }
   }
   else {
      switch( type ) {
         case Event::FONT_CHANGED:
            set_properties();
            break;
         case Event::FILE_IN_PROJECT_RENAMED:
            file_in_project_renamed( event->d_data.at( 0 ).toString(),
                                     event->d_data.at( 1 ).toString() );
            break;
         case Event::BOOKMARK_CLEAR_ALL_REQUEST:
         case Event::DBG_BREAKPOINT_CLEAR_ALL_REQUEST:
            update_margins();
            break;
      }
   }
}
// end of customEvent

//*******************************************************************
// contextMenuEvent                                PRIVATE inherited
//*******************************************************************
void Editor::contextMenuEvent( QContextMenuEvent* const in_event )
{
   QMenu* const menu = new QMenu( this );
   if( menu ) {
      QAction* const sall  = menu->addAction( "Select All" );
      menu->addSeparator();
      QAction* const undo  = menu->addAction( "&Undo" );
      QAction* const redo  = menu->addAction( "&Redo" );
      menu->addSeparator();
      QAction* const cut   = menu->addAction( "Cu&t" );
      QAction* const copy  = menu->addAction( "&Copy" );
      QAction* const paste = menu->addAction( "&Paste" );
      QAction* const del   = menu->addAction( "Delete" );
      menu->addSeparator();
      QAction* const run   = menu->addAction( "R&un" );
      QAction* const dbg   = menu->addAction( "&Debug" );
      menu->addSeparator();
      QAction* const func  = menu->addAction( "&Functions" );
      
      sall->setShortcut  ( tr( "Ctrl+A" )  );
      undo->setShortcut  ( tr( "Ctrl+Z" )  );
      redo->setShortcut  ( tr( "Ctrl+Y" )  );
      cut->setShortcut   ( tr( "Ctrl+X" )  );      
      copy->setShortcut  ( tr( "Ctrl+C" )  );
      paste->setShortcut ( tr( "Ctrl+V" )  );
      run->setShortcut   ( tr( "Ctrl+F5" ) );
      dbg->setShortcut   ( tr( "F5" )      );
      func->setShortcut  ( tr( "Ctrl+U" )  );
      
      undo->setIcon  ( QIcon( ":/img/undo.png"   ) );
      redo->setIcon  ( QIcon( ":/img/redo.png"   ) );
      cut->setIcon   ( QIcon( ":/img/cut.png"    ) );
      copy->setIcon  ( QIcon( ":/img/copy.png"   ) );
      paste->setIcon ( QIcon( ":/img/paste.png"  ) );
      run->setIcon   ( QIcon( ":/img/run.png"    ) );
      dbg->setIcon   ( QIcon( ":/img/dbgrun.png" ) );

      const bool is_selection = textCursor().hasSelection();
      
      undo->setEnabled( document()->isUndoAvailable() && document()->isModified() );
      redo->setEnabled( document()->isRedoAvailable() );
      paste->setEnabled( canPaste() );
      cut->setEnabled( is_selection );
      copy->setEnabled( is_selection );
      del->setEnabled( is_selection );

      connect( undo , SIGNAL( triggered() ), this, SLOT( do_undo      () ) );
      connect( redo , SIGNAL( triggered() ), this, SLOT( do_redo      () ) );
      connect( cut  , SIGNAL( triggered() ), this, SLOT( do_cut       () ) );
      connect( copy , SIGNAL( triggered() ), this, SLOT( do_copy      () ) );
      connect( paste, SIGNAL( triggered() ), this, SLOT( do_paste     () ) );
      connect( del  , SIGNAL( triggered() ), this, SLOT( do_delete    () ) );
      connect( sall , SIGNAL( triggered() ), this, SLOT( do_select_all() ) );
      connect( run  , SIGNAL( triggered() ), this, SLOT( do_run       () ) );
      connect( dbg  , SIGNAL( triggered() ), this, SLOT( do_dbg_run   () ) );
      connect( func , SIGNAL( triggered() ), this, SLOT( do_parse_func() ) );

      menu->exec( in_event->globalPos() );
      delete menu;
   }
}
// end of contextMenuEvent

//*******************************************************************
// do_undo                                              PRIVATE slot
//*******************************************************************
void Editor::do_undo()
{
   if( document()->isModified() ) {
      Renumer renumer( this );
      document()->undo();
   }
}
// end of do_undo

//*******************************************************************
// do_redo                                              PRIVATE slot
//*******************************************************************
void Editor::do_redo()
{
   if( document()->isRedoAvailable() ) {
      Renumer renumer( this );
      document()->redo();
   }
}
// end of do_redo

//*******************************************************************
// do_copy                                              PRIVATE slot
//*******************************************************************
void Editor::do_copy()
{
   if( textCursor().hasSelection() ) {
      Renumer renumer( this );
      copy();
   }
}
// end of do_copy

//*******************************************************************
// do_cut                                               PRIVATE slot
//*******************************************************************
void Editor::do_cut()
{
   if( textCursor().hasSelection() ) {
      Renumer renumer( this );
      cut();
   }
}
// end of do_cut

//*******************************************************************
// do_paste                                             PRIVATE slot
//*******************************************************************
void Editor::do_paste()
{
   if( canPaste() ) {
      Renumer renumer( this );
      paste();
      if( d_syntax ) d_syntax->rehighlight();
      ensureCursorVisible();
   }
}
// end of do_paste

//*******************************************************************
// do_delete                                            PRIVATE slot
//*******************************************************************
void Editor::do_delete()
{
   Renumer renumer( this );
   if( textCursor().hasSelection() ) textCursor().removeSelectedText();
   else                              textCursor().deleteChar();

}
// end of do_delete

//*******************************************************************
// do_select_all                                        PRIVATE slot
//*******************************************************************
void Editor::do_select_all()
{
   selectAll();
}
// end of do_select_all

//*******************************************************************
// do_run                                               PRIVATE slot
//*******************************************************************
void Editor::do_run()
{
   EventsController::instance()->send_event( Event::RUN_REQUEST_EXTERNAL );
}
// end of do_run

//*******************************************************************
// do_dbg_run                                           PRIVATE slot
//*******************************************************************
void Editor::do_dbg_run()
{
   EventsController::instance()->send_event( Event::DBG_RUN_REQUEST_EXTERNAL );
}
// end of do_dbg_run

//*******************************************************************
// do_parse_func                                        PRIVATE slot
//-------------------------------------------------------------------
// Otwierany jest dialog, w ktorym zostana wyswietlone wszystkie
// definicje funkcji w tym pliku. Jesli uzytkownik wybierze, ktoras
// z nich, to wyszukujemy ja w tekscie i kursor ustawia
// sie na poczatku linii z naglowkiem funkcji.
//*******************************************************************
void Editor::do_parse_func()
{
   qint32 pos = -1;

   {
      auto_ptr<QBTFunctionsBox> box( new QBTFunctionsBox( toPlainText() ) );
      if( box.get() ) {
         if( QDialog::Accepted == box->exec() ) {
            pos = box->position();
         }
      }
   }
   if( pos != -1 ) {
      QTextCursor cursor = textCursor();
      cursor.setPosition( pos );
      setTextCursor( cursor );
   }
}
// end of do_parse_func

//*******************************************************************
// goto_line                                                 PRIVATE
//*******************************************************************
void Editor::goto_line()
{
   const auto_ptr<GotoLine> dialog( new GotoLine( this ) );
   if( dialog.get() ) {
      if( GotoLine::Accepted == dialog->exec() ) {
         set_cursor_position( dialog->value(), 1 );
      }
   }
}
// end of goto_line

//*******************************************************************
// get_selected_text                                         PRIVATE
//*******************************************************************
QString Editor::get_selected_text() const
{
   QString retval = "";
   
   if( textCursor().hasSelection() ) {
      retval = textCursor().selectedText();
      textCursor().clearSelection();
   }
   return retval;
}
// end of get_selected_text

//*******************************************************************
// replace                                                   PRIVATE
//*******************************************************************
void Editor::replace()
{
   QTextDocument::FindFlags flags = 0;
   qint32 start_idx = -1;
   qint32 end_idx = -1;
   QString src_text = "";
   QString dst_text = "";
   bool    search = false;
   bool    whole_document = false;
   bool    selected_text = false;

   {  //===== Dialog w ktorym uzytkownik wpisuje o co chodzi =====
      QBtFind dialog( this, "", false );
      
      if( textCursor().hasSelection() ) {
         start_idx = textCursor().selectionStart();
         end_idx = textCursor().selectionEnd();
         dialog.in_selected_text_active();
      }

      if( QBtFind::Accepted == dialog.exec() ) {
         whole_document = dialog.is_whole_document();
         selected_text = dialog.is_selected_text();
         if( dialog.is_whole() ) flags |= QTextDocument::FindWholeWords;
         if( dialog.is_sensitive() ) flags |= QTextDocument::FindCaseSensitively;
         dialog.input_text( src_text );
         dialog.replace_text( dst_text );
         search = true;
      }
   }

   if( search ) {
      QTextCursor cursor = textCursor();
      const QTextBlock block = cursor.block();

      cursor.beginEditBlock(); //---------------------------------
           if( whole_document ) replace_in_whole_document( flags, src_text, dst_text );
      else if( selected_text  ) replace_in_selected_text ( flags, src_text, dst_text, start_idx, end_idx );
      else                      replace_next             ( flags, src_text, dst_text );
      cursor.endEditBlock(); //-----------------------------------
         
      if( whole_document ) {
         cursor.setPosition( block.position() );
         setTextCursor( cursor );
      }
   }
}
// end of replace

//*******************************************************************
// replace_in_whole_document                                 PRIVATE
//*******************************************************************
void Editor::replace_in_whole_document(   const QTextDocument::FindFlags in_flags,
                                          const QString& in_src_text,
                                          const QString& in_dst_text )
{
   QTextCursor cursor = textCursor();
   cursor.movePosition( QTextCursor::Start );
   setTextCursor( cursor );
   
   while( find( in_src_text, in_flags ) ) {
      textCursor().insertText( in_dst_text );
   }
}
// end of replace_in_whole_document

//*******************************************************************
// replace_in_selected_text                                  PRIVATE
//*******************************************************************
void Editor::replace_in_selected_text( const QTextDocument::FindFlags in_flags,
                                       const QString& in_src_text,
                                       const QString& in_dst_text,
                                       const qint32 in_start_idx,
                                       const qint32 in_end_idx )
{
   QTextCursor cursor = textCursor();
   cursor.setPosition( in_start_idx );
   setTextCursor( cursor );

   const qint32 end_idx = in_end_idx - 1;
   while( find( in_src_text, in_flags ) && ( textCursor().selectionEnd() < end_idx ) ) {
      textCursor().insertText( in_dst_text );
   }

   cursor.setPosition( in_start_idx );
   setTextCursor( cursor );
}
// end of replace_in_selected_text

//*******************************************************************
// replace_next                                              PRIVATE
//*******************************************************************
void Editor::replace_next( const QTextDocument::FindFlags in_flags,
                           const QString& in_src_text,
                           const QString& in_dst_text )
{
   if( find( in_src_text, in_flags ) ) {
      textCursor().insertText( in_dst_text );
   }
}
// end of replace_next

//*******************************************************************
// find_in_document                                          PRIVATE
//*******************************************************************
void Editor::find_in_document()
{
   d_find_flags = 0;
   const QString selected_text = get_selected_text();
   
   QBtFind dialog( this, selected_text );
   if( QBtFind::Accepted == dialog.exec() ) {
      if( dialog.is_backward()  ) d_find_flags |= QTextDocument::FindBackward;
      if( dialog.is_whole()     ) d_find_flags |= QTextDocument::FindWholeWords;
      if( dialog.is_sensitive() ) d_find_flags |= QTextDocument::FindCaseSensitively;
      dialog.input_text( d_find_text );
      if( !find( d_find_text, d_find_flags ) ) {
         QMessageBox::information( this, tr( FIND ), tr( NOT_FOUND ), QMessageBox::Ok );
      }
   }
}
// end of find_in_document

//*******************************************************************
// find_next                                                 PRIVATE
//*******************************************************************
void Editor::find_next()
{
   if( !d_find_text.isEmpty() ) {
      d_find_flags &= ~QTextDocument::FindBackward;
      if( !find( d_find_text, d_find_flags ) ) {
         QMessageBox::information( this, tr( FIND ), tr( NOT_FOUND ), QMessageBox::Ok );
      }
   }
}
// end of find_next

//*******************************************************************
// find_previous                                             PRIVATE
//*******************************************************************
void Editor::find_previous()
{
   if( !d_find_text.isEmpty() ) {
      if( textCursor().hasSelection() ) {
         QTextCursor cursor = textCursor();
         cursor.setPosition( cursor.selectionStart() );
         setTextCursor( cursor );
         ensureCursorVisible();
      }
      d_find_flags |= QTextDocument::FindBackward;
      if( !find( d_find_text, d_find_flags ) ) {
         QMessageBox::information( this, tr( FIND ), tr( NOT_FOUND ), QMessageBox::Ok );
      }
   }
}
// end of find_previous

//*******************************************************************
// set_file_name                                             PRIVATE
//*******************************************************************
void Editor::set_file_name( const QString& in_fpath )
{
   Q_ASSERT( !in_fpath.isEmpty() );

   d_is_noname = false;
   set_window_title( d_file_path = in_fpath ); 
   document()->setModified( false );
   set_properties();
}
// end of set_file_name

//*******************************************************************
// keyPressEvent                                   PRIVATE inherited
//*******************************************************************
void Editor::keyPressEvent( QKeyEvent* const in_event )
{
   Q_ASSERT( in_event );

   const qint32 key = in_event->key();

   if( !can_edit( key ) ) {
      in_event->accept();
      return;
   }

   if( in_event->modifiers() & Qt::ShiftModifier ) {
      switch( key ) {
         case Qt::Key_Delete:
            do_cut();
            in_event->accept();
            return;
         case Qt::Key_Insert:
            do_paste();
            in_event->accept();
            return;
      }
   }
   
   if( in_event->modifiers() & Qt::ControlModifier ) {
      switch( key ) {
         case Qt::Key_Down:
            bookmark_next();
            break;
         case Qt::Key_Up:
            bookmark_prev();
            break;
         case Qt::Key_Z:
            do_undo();
            break;
         case Qt::Key_Y:
            do_redo();
            break;
         case Qt::Key_X:
            do_cut();
            break;
         case Qt::Key_C:
            do_copy();
            break;
         case Qt::Key_V:
            do_paste();
            break;
         case Qt::Key_A:
            selectAll();
            break;
         case Qt::Key_U:
            do_parse_func();
            break;
         case Qt::Key_Delete:
            ctrl_delete_key( in_event );
            break;
         default:
            QTextEdit::keyPressEvent( in_event );
      }
   }
   else {
      switch( key ) {
         case Qt::Key_Tab:
            tab_key( in_event );
            break;
         case Qt::Key_Backspace:
            backspace_key( in_event );
            break;
         case Qt::Key_Delete:
            delete_key( in_event );
            break;
         case Qt::Key_Return:
         case Qt::Key_Enter:
            return_key( in_event );
            break;
         case Qt::Key_Home:
            home_key( in_event );
            break;
         case Qt::Key_BraceRight:
            brace_right_key( in_event );
            break;
         default:
            QTextEdit::keyPressEvent( in_event );
      }
   }

   if( use_folder_ ) {
      if( !is_navi_key( key ) ) {
         folder_->parse();
      }
   }
}
// end of keyPressEvent

//*******************************************************************
// can_edit                                                  PRIVATE
//*******************************************************************
bool Editor::can_edit( const qint32 in_key ) const
{
   bool retval = true;

   if( use_folder_ && folder_ ) {
      LineInfo* const info = dynamic_cast<LineInfo*>( textCursor().block().userData() );
      if( info ) {
         if( FOLDER_MARKER_BEGIN_CLOSED == info->folding_marker ) {
            retval = is_navi_key( in_key );
         }
      }
   }
   
   return retval;
}
// end of can edit

//*******************************************************************
// is_navi_key                                               PRIVATE
//*******************************************************************
bool Editor::is_navi_key( const qint32 in_key ) const
{
   return ( ( in_key <= Qt::Key_PageDown ) && ( in_key >= Qt::Key_Home ) );
}
// end of is_navi_key

//*******************************************************************
// bookmarking                                               PRIVATE
//-------------------------------------------------------------------
// Zmiana znacznika bookmark.
//*******************************************************************
void Editor::bookmarking()
{
   const QTextBlock block = textCursor().block();
   if( block.isValid() ) {
      LineInfo* const info = dynamic_cast< LineInfo* >( block.userData() );
      if( info ) {
         const qint32 idx = row();
         Bookmarks::instance()->change_status( d_file_path, idx, block.text().trimmed() );
         info->bookmark = ( Bookmarks::instance()->status( d_file_path, idx ) == Points::SET );
         margin_->update();
      }
   }
}
// end of bookmarking

//*******************************************************************
// breakpointing                                             PRIVATE
//-------------------------------------------------------------------
// Zmiana znacznika breakpoint.
//*******************************************************************
void Editor::breakpointing()
{
   const QTextBlock block = textCursor().block();
   if( block.isValid() ) {
      LineInfo* const info = dynamic_cast< LineInfo* >( block.userData() );
      if( info ) {
         const qint32 idx = row();
         Breakpoints::instance()->change_status( d_file_path, idx, block.text().trimmed() );
         info->breakpoint = ( Breakpoints::instance()->status( d_file_path, idx ) == Points::SET );
         margin_->update();
      }
   }
}
// end of bookmarking

//*******************************************************************
// bookmark_next                                             PRIVATE
//*******************************************************************
void Editor::bookmark_next()
{
   const int idx = Bookmarks::instance()->next( d_file_path, row() );
   if( idx != -1 ) {
      set_cursor_position( idx + 1, 1 );
   }
}
// end of bookmark_next

//*******************************************************************
// bookmark_prev                                             PRIVATE
//*******************************************************************
void Editor::bookmark_prev()
{
   const qint32 idx = Bookmarks::instance()->prev( d_file_path, row() );
   if( idx != -1 ) {
      set_cursor_position( idx + 1, 1 );
   }
}
// end of bookmark_prev

//*******************************************************************
// brace_right_key                                           PRIVATE
//*******************************************************************
void Editor::brace_right_key( QKeyEvent* const in_event )
{
   in_event->accept();

   QTextCursor cursor = textCursor();
   QTextBlock block = cursor.block();
   const qint32 pos = cursor.position();
   cursor.movePosition( QTextCursor::StartOfBlock, QTextCursor::KeepAnchor );

   if( QBtParser::is_white_string( cursor.selectedText() ) ) {
      // W aktualnej linii sa tylko biale spacje.
      // Szukamy pozycji nie domknietego nawiasu i
      // prawy zamykajacy nawias stawiamy w kolumnie,
      // takiej jak nawias otwierajacy.
      QString prefix = "";
      const qint32 idx = search_open_paranthesis( pos );
      if( idx != -1 ) {
         cursor.setPosition( idx );
         if( !cursor.atBlockStart() ) {
            cursor.movePosition( QTextCursor::StartOfBlock );
            cursor.movePosition( QTextCursor::NextWord, QTextCursor::KeepAnchor );
            const QString sel_chrs = cursor.selectedText();
            if( QBtParser::is_white_string( sel_chrs ) ) {
               prefix = sel_chrs;
            }
         }
      }
      cursor.setPosition( pos );
      cursor.movePosition( QTextCursor::StartOfBlock, QTextCursor::KeepAnchor );
      cursor.insertText( prefix + QString( '}' ) );
   }
   else {
      // Wczesniej w aktualnej linii juz cos jest.
      // Pozostanie nam tylko dopisac znak na aktualnej pozycji.
      cursor.setPosition( pos );
      cursor.insertText( QString( '}' ) );
   }
}
// end of brace_right_key
   
//*******************************************************************
// ctrl_delete_key                                           PRIVATE
//*******************************************************************
void Editor::ctrl_delete_key( QKeyEvent* const in_event )
{
   Renumer renumer(this );
   in_event->accept();

   QTextCursor cursor = textCursor();
   cursor.beginEditBlock();
   { 
      const QString text      = toPlainText();
      const qint32  start_idx = cursor.position();
      const qint32  last_idx  =  ( QBtParser::is_white_char( text[ start_idx ] ) )
                                 ? QBtParser::first_after_white_chr( text, start_idx )
                                 : QBtParser::first_white_chr( text, start_idx );
      if( start_idx < last_idx ) {
         cursor.setPosition( start_idx );
         cursor.setPosition( last_idx, QTextCursor::KeepAnchor );
         cursor.removeSelectedText();
      }
   }
   cursor.endEditBlock();
}
// end of ctrl_delete_key

//*******************************************************************
// delete_key                                                PRIVATE
//*******************************************************************
void Editor::delete_key( QKeyEvent* const in_event )
{
   QTextEdit::keyPressEvent( in_event );
   Renumer renumer( this );
}
// end of delete_key

//*******************************************************************
// return_key                                                PRIVATE
//*******************************************************************
void Editor::return_key( QKeyEvent* const in_event )
{
   Renumer renumer( this );
   in_event->accept();

   QTextCursor cursor = textCursor();
   cursor.beginEditBlock();
   {
      textCursor().insertText( "\n" );

      if( Config::instance()->auto_indent() ) {
         QTextCursor      cursor                   = textCursor();
         const QTextBlock current_block            = cursor.block();
         const qint32     pos                      = cursor.position();
         QString          prefix_text              = "";
         qint32           open_parantesis_line_idx = -1;
         qint32           no_empty_line_idx        = -1;

         // Szukamy linii z pierwszym nie domknietym nawiasem.
         open_parantesis_line_idx = search_open_paranthesis( pos );
         
         {  // Szukamy pierwszej niepustej linii (insteresuja nas linie po nawiasie)
            QTextBlock text_block = current_block.previous();
            while( ( text_block.isValid() ) && ( text_block.position() > open_parantesis_line_idx ) ) {
               if( !text_block.text().isEmpty() ) {
                  no_empty_line_idx = text_block.position();
                  break;
                }
                text_block = text_block.previous();
            }
         }

         {  // Odczytujemy prefix-text
            if( no_empty_line_idx > open_parantesis_line_idx ) {
               cursor.setPosition( no_empty_line_idx );
               prefix_text = get_prefix_text( cursor.block().text() );
            }
            else if( open_parantesis_line_idx != -1 ) {
               cursor.setPosition( open_parantesis_line_idx );
               prefix_text = get_prefix_text( cursor.block().text() );
               prefix_text += tab_chars( pos );
            }
         }

         {  // Jezeli poprzednia linia zawiera tylko wciecie
            // (biale znaki) to je usuwamy.
            QTextBlock text_block = current_block.previous();
            if( text_block.isValid() ) {
               const QString text = text_block.text();
               if( !text.isEmpty() ) {
                  if( QBtParser::is_white_string( text ) ) {
                     cursor.setPosition( text_block.position() );
                     cursor.movePosition( QTextCursor::EndOfBlock, QTextCursor::KeepAnchor );
                     cursor.removeSelectedText();
                  }
               }
            }
         }
         
         // Wstawiamy prefix-text
         cursor.setPosition( current_block.position() );
         if( !prefix_text.isEmpty() ) {
            cursor.insertText( prefix_text );
         }
      }
   }
   cursor.endEditBlock();
   setTextCursor( cursor );
}
// end of return_key

//*******************************************************************
// home_key                                                  PRIVATE
//*******************************************************************
void Editor::home_key( QKeyEvent* const in_event )
{
   in_event->accept();

   QTextCursor cursor = textCursor();
   cursor.beginEditBlock();
    
   // Ctrl + Home : kursor na poczatek dokumentu.
   if( Qt::ControlModifier & in_event->modifiers() ) {
      cursor.movePosition( QTextCursor::Start );
   }
   else {
      const qint32 col = cursor.position() - cursor.block().position();
      const bool   ist = cursor.atBlockStart();

      // Domyslnie stajemy na poczatku linii (jesli tam nie jestesmy).
      if( !ist ) {
         cursor.movePosition( QTextCursor::StartOfBlock );
      }
        
      const qint32 idx = QBtParser::first_after_white_chr( cursor.block().text() );
      if( idx > 0 ) {
         // Jezeli kursor byl gdzies w tekscie linii
         // lub byl juz na jej poczatku,to przeskakujemy
         // do pierwszego wyrazu.
         if( ( col > idx ) || ist ) {
            cursor.setPosition( cursor.position() + idx );
         }
      }
   }
   cursor.endEditBlock();
   setTextCursor( cursor );
}
// end of home_key

//*******************************************************************
// get_prefix_text                                           PRIVATE
//*******************************************************************
QString Editor::get_prefix_text( const QString& in_text ) const
{
   QString retval = "";
   const qint32 idx = QBtParser::first_after_white_chr( in_text );
   if( idx > 0 ) {
      retval = in_text.left( idx );
   }
   return retval;
}
// end of get_prefix_text

//*******************************************************************
// backspace_key                                             PRIVATE
//*******************************************************************
void Editor::backspace_key( QKeyEvent* const in_event )
{
   QTextCursor cursor = textCursor();
   if( cursor.hasSelection() ) {
      in_event->accept();

      const qint32  anchor_pos = cursor.anchor();
      const qint32  n          = Config::instance()->tab_width();
      const QString tab_muster = ( Config::instance()->tab2spc() ) ? QString( n, ' ' ) : QString( '\t' );

      const qint32 where = is_tab_at_begin( anchor_pos, tab_muster );
      if( where != NO_TAB ) {
         QString txt = cursor.selectedText();
         if( are_tabs_in_selblock( txt, tab_muster ) ) {
            const qint32 tab_muster_len = tab_muster.length();

            if( TAB_AFTER == where ) txt.remove( 0, tab_muster_len );
            txt.replace( QChar( 8233 ) + tab_muster, QChar( 8233 ) );
            cursor.insertText( txt );

            // Ponowne zaznaczenie bloku
            const qint32 current_pos = cursor.position();
            cursor.setPosition( anchor_pos );
            cursor.setPosition( current_pos, QTextCursor::KeepAnchor );

            // Ustawienie kursora z dokonanymi zmianami.
            setTextCursor( cursor );

            if( TAB_BEFORE == where ) {
               QTextCursor cursor = textCursor();
               cursor.setPosition( cursor.anchor() );
               cursor.setPosition( cursor.anchor() - tab_muster_len, QTextCursor::KeepAnchor );
               cursor.removeSelectedText();
            }
         }
         Renumer::run( this ); 
      }
   }
    else {
        QTextEdit::keyPressEvent( in_event );
        Renumer renumer( this ); 
    }
}
// end of backspace_key

//*******************************************************************
// tab_key                                                   PRIVATE
//-------------------------------------------------------------------
// 1) Jezeli kursor znajduje sie w zaznaczonym bloku, przesuwany
//    jest w prawo caly blok. Na POCZATKU KAZDEJ LINII bloku
//    wciskany jest 'znak tabulatora'.
// 2) W przeciwnym przypadku 'znak tabulatora' wciskany jest
//   na aktualnej pozycji kursora.
//
// 'Znak tabulatora' zalezy od konfiguracji programu. Moze to
// faktyczny znak tabulacji, lub lancuch zlozony ze kilku spacji.
// To ile spacji tworzy 'znak tabulatora' okreslone jest przez
// konfuguracje programu.
//*******************************************************************
void Editor::tab_key( QKeyEvent* const in_event )
{
   in_event->accept();

   QTextCursor cursor = textCursor();
   ( cursor.hasSelection() ) ? ins_tabs_in_selblock( cursor ) : ins_one_tab( cursor );
   setTextCursor( cursor );

   Renumer::run( this ); 
}
// end of tab_key

//*******************************************************************
// tab_chars                                                 PRIVATE
//-------------------------------------------------------------------
// Ustalenie z ilu i jakich znakow sklada sie 'znak-tabulacji'.
// Sprawa jest prosta jesli wstawiamy prawdziwe znaki tabulacji:
// zawsze wstawiamy jeden znal '\t'.
// Inaczej przedstawia sie sprawa jesli uzywamy spacji.
// Wowczas 'znak-tabulacji' sklada sie ze zmiennej liczby spacji.
// Liczba spacji musi byc tak dobrana, aby zawsze trafiac w pozycje
// tabulacji.
//*******************************************************************
QString Editor::tab_chars( const qint32 in_pos )
{
   QString result = "\t";

   // W przypadku spacji trzeba pokombinowac.
   if( Config::instance()->tab2spc() ) {
      QTextCursor cursor = textCursor();
      cursor.setPosition( in_pos );

      const qint32 col           = in_pos - cursor.block().position();
      const qint32 n             = Config::instance()->tab_width();
      const qint32 next_tab_mark = ( ( col / n ) * n ) + n;

      result = QString( next_tab_mark - col, ' ' );
   }
   return result;
}
// end of tab_chars

//*******************************************************************
// ins_one_tab                                               PRIVATE
//-------------------------------------------------------------------
// Wstawienie 'znaku tabulacji' na aktualnej pozycji kursora.
// Po operacji kursor znajduje sie ZA wstawionym 'znakiem tabulacji'.
//*******************************************************************
void Editor::ins_one_tab( QTextCursor& in_cursor )
{
   in_cursor.insertText( tab_chars( in_cursor.position() ) );
}
// end of ins_one_tab

//*******************************************************************
// ins_tabs_in_selblock                                      PRIVATE
//-------------------------------------------------------------------
// Wstawienie 'znaku tabulacji' na poczatku kazdej linii zaznaczonego
// bloku tekstu. UWAGA: czyli nie na faktycznej aktualnej
// pozycji kursora. Po operacji kursor znajduje sie na koncu
// bloku i blok nadal pozostaje zaznaczony.
//*******************************************************************
void Editor::ins_tabs_in_selblock( QTextCursor& in_cursor )
{
   QTextCursor tmp_cursor = in_cursor;
   QString     txt        = in_cursor.selectedText();

   if( !txt.isEmpty() ) {
      const qint32  sel_start_pos = in_cursor.selectionStart();
      const QString chars         = tab_chars( sel_start_pos );
      const qint32  n             = chars.length();
      qint32        idx           = 0;

      txt.insert( idx, chars );
      idx += n;
      do {
         if( 8233 == txt[ idx ].unicode() ) {
            txt.insert( ++idx, chars );
            idx += n;
            continue;
         }
         ++idx;
      } while( idx < txt.length() );

      in_cursor.insertText( txt );
      const qint32 current_pos = in_cursor.position();
      in_cursor.setPosition( sel_start_pos );
      in_cursor.setPosition( current_pos, QTextCursor::KeepAnchor );
   }
}
// end of ins_tabs_selblock

//*******************************************************************
// is_tab_at_begin                                           PRIVATE
//-------------------------------------------------------------------
// Sprawdza czy PRZED LUB PO wskazanej pozycja znajduje sie
// 'znak tabulacji'.
// UWAGA: zaznaczenie moze zaczynac sie na poczatku pierwszej linii
//        lub nie. Dlatego najpierw sprawdzamy znaki za wksazana
//        pozycja, a nastepnie przed wskazana pozycja.
//*******************************************************************
qint32 Editor::is_tab_at_begin( const qint32 in_idx, const QString& in_tab_muster ) const
{
   qint32        retval = NO_TAB;
   const QString txt    = toPlainText();

   // Szukamy za aktualna pozycja.
   if( QBtParser::starts_with( txt, in_idx, in_tab_muster ) ) {
      retval = TAB_AFTER;
   }

   // Jesli nie znaleziono odpowiednie liczby spacji
   // przed wskazana pozycja, to szukamy ich przed nia.
   if( NO_TAB == retval ) {
      if( QBtParser::starts_with( txt, in_idx - in_tab_muster.length(), in_tab_muster ) ) {
         retval = TAB_BEFORE;
      }
   }
   return retval;
}
// end of is_tab_at_begin

//*******************************************************************
// are_tabs_in_selblock                                      PRIVATE
//-------------------------------------------------------------------
// Sprawdza czy we wskazanym bloku znajduja sie okreslone
// 'znaki tabulacji'.
//*******************************************************************
bool Editor::are_tabs_in_selblock( const QString& in_txt, const QString& in_tab_muster) const
{
   bool retval = false;
   const qint32 n = in_tab_muster.length();

   qint32 idx = 0;
   while( idx < in_txt.length() ) {
      if( 8233 == in_txt[ idx ].unicode() ) {
         ++idx;
         if( QBtParser::starts_with( in_txt, idx, in_tab_muster ) ) {
            retval = true;
            idx += n;
            continue;
         }
         else {
            retval = false;
            break;
         }
         continue;
      }
      ++idx;
   }
   return retval;
}
// end of are_tabs_in_selblock

//*******************************************************************
// search_open_paranthesis                                   PRIVATE
//*******************************************************************
qint32 Editor::search_open_paranthesis ( const qint32 in_pos )
{
   const QString text = toPlainText();
   qint32        idx  =  0;

   QStack< qint32 > open_idxs;
   open_idxs.push( -1 );

   while( ( idx != -1 ) && ( idx < in_pos ) ) {
      idx = QBtParser::search_muster( text, "{", idx );
      if( ( idx != -1 ) && ( idx < in_pos ) ) {
         open_idxs.push( idx );
         const qint32 idx2 = QBtParser::search_marker_end( text, idx + 1, '{', '}' );
         if( ( idx2 != -1 ) && ( idx2 < in_pos ) ) {
            open_idxs.pop();
            idx = idx2 + 1;
            continue;
         }
         ++idx;
      }
   }
   return open_idxs.pop();
}
// end of search_open_paranthesis

/*
qint32 Editor::search_closed_marker( 
   const QString& in_text,       // aktualny tekst edytora
   const qint32   in_pos,        // aktualna pozycja kursora
   const QChar&   in_start_mark, // znak oznaczajacy poczatek bloku
   const QChar& in_end_mark )    // znak oznaczajacy koniec bloku
{
   qint32 retval = -1;
   qint32 idx    =  0;

   while( ( idx != -1 ) && ( idx < in_pos ) ) {
      idx = QBtParser::search_marker( in_text, in_start_mark, idx );
      if( ( idx != -1 ) && ( idx < in_pos ) ) {
         const qint32 idx2 = QBtParser::search_marker_end( in_text, idx + 1, in_start_mark, in_end_mark );
         if( idx2 != -1 ) {
            if( idx2 == in_pos ) {
               retval = idx;
               break;
            }
            if( idx2 < in_pos ) {
               idx = idx2 + 1;
               continue;
            }
         }
         ++idx;
      }
   }
   return retval;
}
*/

//###################################################################
//#                                                                 #
//#                      S A V E  F I L E                           #
//#                                                                 #
//###################################################################


//********************************************************************
// try_save                                                   PRIVATE
//********************************************************************
bool Editor::try_save()
{
   bool retval = true; // true oznacza - mozna zamknac okno

   if( document()->isModified() ) {
      const QString fname = QFileInfo( d_file_path ).fileName();
      const qint32 ret = QMessageBox::warning( this,
                        tr( SAVING ) + " - " + fname,
                        tr( DOC_CHANGED_MSG ),
                        QMessageBox::Yes|QMessageBox::Default,
                        QMessageBox::No,
                        QMessageBox::Cancel|QMessageBox::Escape);
      switch( ret ) {
         case QMessageBox::Yes:
            // zgoda na zamkniecie okna, ale po zapisie
            retval = save();
            break;
         case QMessageBox::No:
            // nie chce zapisac, ale chce zamknac okno
            break;
         case QMessageBox::Cancel:
            // nie chce zamykac okna
            retval = false;
            break;
      }
   }
   return retval;
}
// end of try_save

//********************************************************************
// save                                                       PRIVATE
//********************************************************************
bool Editor::save()
{
   return ( d_is_noname ? save_as() : save_file( d_file_path ) );
}
// end of save

//********************************************************************
// save_as                                                    PRIVATE
//********************************************************************
bool Editor::save_as()
{
   bool retval = FALSE;
   const QString last_dir = d_workspace->last_save_dir();
   const QString old_file_path = d_file_path;

   QString const fpath = QFileDialog::getSaveFileName( this, tr( SAVING ), last_dir, tr( FILE_FILTERS ) );
   if( fpath != QString::null ) {
      const QFileInfo fi( fpath );
      if( !fi.fileName().isEmpty() ) {
         d_workspace->last_save_dir( fi.absolutePath() );
         retval = save_file( fi.absoluteFilePath() );
         if( retval ) {
            EventsController::instance()->send_event( Event::FILE_RENAMED, old_file_path, d_file_path );
         }
      }
   }
   return retval;
}
// end of save_as

//*******************************************************************
// file_in_project_renamed                                   PRIVATE
//*******************************************************************
void Editor::file_in_project_renamed( const QString& in_old_fpath, const QString& in_new_fpath )
{
   if( d_file_path == in_old_fpath ) {
      set_file_name( in_new_fpath );
      EventsController::instance()->send_event( Event::FILE_RENAMED, in_old_fpath, in_new_fpath );
   }
}
// end of file_in_project_renamed

//********************************************************************
// save_file                                                  PRIVATE
//********************************************************************
bool Editor::save_file( const QString& in_fpath )
{
   bool retval = write_file( in_fpath );
   if( retval ) {
      const QFileInfo fi( in_fpath );
      d_last_modified = fi.lastModified();
      
      document()->setModified( false );
      set_file_name( in_fpath );
      
      // Jesli jest to plik c* lub h* i skonfigurowano
      // parsing on save to wywolujemy reparsowanie tego pliku.
      if( d_can_parse && Config::instance()->parse_class_onsave() ) {
         Classer::instance()->reparse( d_file_path, toPlainText() );     
      }
   }
   return retval;
}
// end of save_file

//*******************************************************************
// write_file                                                PRIVATE
//*******************************************************************
bool Editor::write_file( const QString& in_fpath )
{
   bool retval = false;
   QFile file( in_fpath );

   if( file.open( QIODevice::WriteOnly ) ) {
      QTextStream out( &file );
      QApplication::setOverrideCursor( Qt::WaitCursor );
      QString text = toPlainText();
      restore_folded_text( text );
      const QTextDocument doc( text );
      out << doc.toPlainText();
      QApplication::restoreOverrideCursor();
      file.close();
      retval = true;
   }
   else {
      QMessageBox::critical( this,
         tr( SAVING ),
         tr( SAVE_ERROR_MSG ).arg( file.fileName() ).arg( file.errorString() ));
   }
   return retval;
}
// end of write_file

//*******************************************************************
// restore_folded_text                                       PRIVATE
//*******************************************************************
void Editor::restore_folded_text( QString& inout_text ) const
{
   if( !folder_ ) return;
   
   QString dst_text = QString();
   QTextBlock block = document()->begin();
   
   while( block.isValid() ) {
      LineInfo* const info = dynamic_cast< LineInfo* >( block.userData() );
      if( info ) {
         if( FOLDER_MARKER_BEGIN_CLOSED == info->folding_marker ) {
            const QString src_text = block.text();
            if( folder_->get_pair_text( src_text, dst_text, false ) ) {
               inout_text.replace( src_text, dst_text );
            }
         }
      }
      block = block.next();
   }
}
// end of restore_folded_text

//###################################################################
//#                                                                 #
//#                      O P E N  F I L E                           #
//#                                                                 #
//###################################################################

//********************************************************************
// open_file_dialog                                            PUBLIC
//********************************************************************
bool Editor::open_file_dialog( const QString& in_dir )
{
   bool retval = FALSE;

   const QString fpath = QFileDialog::getOpenFileName( this, tr( OPENING ), in_dir, tr( FILE_FILTERS ) );
   if( fpath != QString::null ) {
      retval = open_file( fpath );
   }
   return retval;
}
// end of open_file_dialog

//*******************************************************************
// open_file                                                  PUBLIC
//*******************************************************************
bool Editor::open_file( const QString& in_file_path )
{
   const bool retval = read_file( in_file_path );
   if( retval ) {
      d_workspace->addWindow( this );
      d_workspace->last_open_dir( in_file_path );
      set_file_name( in_file_path );

      Renumer::run( this, false );
      apply_bookmarks();
      apply_breakpoints();

      QTextCursor cursor = textCursor();
      cursor.setPosition( 0 );
      setTextCursor( cursor );
      ensureCursorVisible();
   }
   return retval;
}
// end of open_file

//*******************************************************************
// apply_bookmarks                                           PRIVATE
//*******************************************************************
void Editor::apply_bookmarks()
{
   QHash<int,Points::PointInfo> data;
   Bookmarks::instance()->file_data( d_file_path, data );

   if( !data.empty() ) {
      QList<int> nrs = data.keys();
      qSort( nrs );
      
      QTextBlock block = document()->begin();
      for( int i = 0; i < nrs.size(); ++i ) {
         while( block.isValid() ) {
            LineInfo* const info = dynamic_cast<LineInfo*>( block.userData() );
            if( info ) {
               if( i == block.blockNumber() ) {
                  info->bookmark = true;
                  block = block.next();
                  break;
               }
            }
            block = block.next();
         }
      }
   }
}
// end of apply_bookmarks

//*******************************************************************
// apply_bookmarks                                           PRIVATE
//*******************************************************************
void Editor::apply_breakpoints()
{
   QHash<int,Points::PointInfo> data;
   Breakpoints::instance()->file_data( d_file_path, data );

   if( !data.empty() ) {
      QList<int> nrs = data.keys();
      qSort( nrs );
      
      QTextBlock block = document()->begin();
      for( int i = 0; i < nrs.size(); ++i ) {
         while( block.isValid() ) {
            LineInfo* const info = dynamic_cast<LineInfo*>( block.userData() );
            if( info ) {
               if( i == block.blockNumber() ) {
                  info->breakpoint = true;
                  block = block.next();
                  break;
               }
            }
            block = block.next();
         }
      }
   }
}
// end of apply_breakpoints

//*******************************************************************
// read_file                                                 PRIVATE
//*******************************************************************
bool Editor::read_file( const QString& in_fpath )
{
   bool retval = false;

   QFile file( in_fpath );
   if( file.open( QIODevice::ReadOnly ) ) {
      QTextStream in( &file );
      QApplication::setOverrideCursor( Qt::WaitCursor );
      QString text = in.readAll();

      if( Config::instance()->tab2spc() ) {
         const qint32 n = Config::instance()->tab_width();
         const QString tab_chars =   QString( n, ' ' );
         text.replace( '\t', tab_chars );
      }

      insertPlainText( text );
      QApplication::restoreOverrideCursor();
      file.close();

      const QFileInfo fi( in_fpath );
      d_last_modified = fi.lastModified();

      retval = true;
   }
   else {
      QMessageBox::critical( this,
         tr( OPENING ),
         tr( READ_ERROR_MSG ).arg( file.fileName() ).arg( file.errorString() ));
   }
   return retval;
}
// end of readFile


//###################################################################
//#                                                                 #
//#                 P R I V A T E   S L O T S                       #
//#                                                                 #
//###################################################################


//*******************************************************************
// contents_changed                                     PRIVATE slot
//*******************************************************************
void Editor::contents_changed()
{
   if( document()->isModified() ) {
      set_window_title( d_file_path );
      // Jesli plik jest c* lub h* oraz ustawiono
      // parsing online wywolujemy reparsing tego pliku.
      if( d_can_parse && Config::instance()->parse_class_online() ) {
         Classer::instance()->reparse( d_file_path, toPlainText() );     
      }
   }
}
// end of contents_changed

//*******************************************************************
// contents_changed                                     PRIVATE slot
//*******************************************************************
void Editor::copy_available( bool const in_status )
{
   EventsController::instance()->send_event( Event::COPY_AVAILABLE, in_status );
}
// end of copy_available

//*******************************************************************
// contents_changed                                     PRIVATE slot
//*******************************************************************
void Editor::redo_available( bool const in_status )
{
   EventsController::instance()->send_event( Event::REDO_AVAILABLE, in_status );
}
// end of redo_available

//*******************************************************************
// check_datetime                                       PRIVATE slot
//*******************************************************************
void Editor::check_datetime()
{
   if( !d_is_noname ) {
      const QFileInfo fi( d_file_path );
      if( fi.exists() ) {
         if( d_last_modified != fi.lastModified() ) {
            const qint32 ret = QMessageBox::question( this,
                           tr( CHKDT_CAPTION ), tr( CHKDT_MESSAGE ),
                           QMessageBox::No, QMessageBox::Yes );
            if( QMessageBox::Yes == ret ) {
               const qint32 curpos = cursor_position();
               clear();
               read_file( d_file_path );
               set_file_name( d_file_path );
               set_cursor_position( curpos );
            }
            d_last_modified = fi.lastModified();
         }
      }
   }
}
// end of check_datetime

//*******************************************************************
// contents_changed                                     PRIVATE slot
//*******************************************************************
void Editor::undo_available( bool const in_status )
{
   EventsController::instance()->send_event( Event::UNDO_AVAILABLE, in_status );
}
// end of undo_available

//*******************************************************************
// cursor_position                                            PUBLIC
//*******************************************************************
qint32 Editor::cursor_position() const
{
   return textCursor().position();
}
// end of cursor_position

//*******************************************************************
// cursor_position_changed                              PRIVATE slot
//*******************************************************************
void Editor::cursor_position_changed( const bool in_update_margins )
{
   const QTextCursor cursor = textCursor();
   const QTextBlock  block  = cursor.block();
   
   int col = -1;
   int row = -1;
   if( block.isValid() ) {
      col = cursor.position() - block.position() + 1;
      row = block.blockNumber() + 1;
   }
   EventsController::instance()->send_event( Event::CURSOR_POSITION_CHANGED, row, col );

   if( in_update_margins ) {
      margin_->update();
      if( use_folder_ ) folder_->update();
   }
   viewport()->update();
}
// end of cursor_position_changed

//*******************************************************************
// set_cursor_position                                       PRIVATE
//*******************************************************************
void Editor::set_cursor_position( const int in_row, const int in_col )
{
   const QTextBlock block = document()->findBlockByNumber( in_row - 1 );
   if( block.isValid() ) {
     set_cursor_position( block.position() + in_col - 1 );
   }
}
void Editor::set_cursor_position( const int in_position )
{
   const int pos = ( in_position < 0 ) ? 0 : in_position;
   QTextCursor cursor = textCursor();
   cursor.setPosition( pos );
   setTextCursor( cursor );
   ensureCursorVisible();
}
// end of set_cursor_position

//*******************************************************************
// search_class                                               PUBLIC
//*******************************************************************
void Editor::search_class( const QString& in_class_name )
{
   const QString muster = "\\bclass\\s+" + in_class_name;

   QTextCursor cursor;
   QTextBlock block = document()->begin();
   const QTextBlock end = document()->end();

   while( block != end ) {
      QString line = block.text();
      if( !line.isEmpty() ) {
         if( line.indexOf( QRegExp( muster ) ) != -1 )  {
            cursor = textCursor();
            cursor.setPosition( block.position() );
            setTextCursor( cursor );
            ensureCursorVisible();
            break;
         }
      }
      block = block.next();
   }
}
// end of search_expr

//*******************************************************************
// print                                                      PUBLIC
//-------------------------------------------------------------------
// Wydrukowanie dokumentu na drukarce.
//*******************************************************************
void Editor::print( const bool in_fast ) const
{
   QBtPrinter::instance()->print( file_path(), this, font(), in_fast );
}
// end of print
