#ifndef INCLUDED_Editor_h
#define INCLUDED_Editor_h

/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : Editor.h
 * DATE   : 3.09.2006
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "Bookmarks.h"
#include "Cobras.h"
#include "TextEdit.h"
#include "Shared.h"
#include "QBTCodeBlockInfo.h"
#include <QRegExp>
#include <QDateTime>
#include <QMutexLocker>
#include <QTextBlockUserData>
#include <QContextMenuEvent>

/*------- forward declarations:
-------------------------------------------------------------------*/
class QBtWorkspace;
class CppSyntax;
class Margin;
class Folder;
class QTimer;

/*------- class declaration:
-------------------------------------------------------------------*/
class LineInfo : public QTextBlockUserData
{
public:
   LineInfo() : QTextBlockUserData()
   , bookmark  ( false )
   , breakpoint( false )
   , folding_marker( FOLDER_NO_MARKER )
   , folding( QBTCodeBlockInfo() )
   {}
private:
   LineInfo( const LineInfo& );
   LineInfo& operator=( const LineInfo& );
public:
   bool bookmark;
   bool breakpoint;
   int  folding_marker;
   QBTCodeBlockInfo folding;
};

class Editor : public TextEdit
{
   Q_OBJECT

// ******* CONSTRUCTION/DESTRUCTION *******
public:
   explicit Editor( QWidget* );
   ~Editor();
private:
   Editor( const Editor& );
   Editor& operator=( const Editor& );

// ******* CONSTANTS *******
private:
   enum {
      NO_TAB = 0,
      TAB_BEFORE,
      TAB_AFTER,
   };
   static const char* const NO_NAME;
   static const char* const CHANGE_MARK;
   static const char* const SAVING;
   static const char* const OPENING;
   static const char* const FIND;
   static const char* const NOT_FOUND;
   static const char* const DOC_CHANGED_MSG;
   static const char* const FILE_FILTERS;
   static const char* const SAVE_ERROR_MSG;
   static const char* const READ_ERROR_MSG;
   static const char* const DEFAULT_FONT;
   static const char* const CHKDT_CAPTION;
   static const char* const CHKDT_MESSAGE;

// ******* MEMBERS *******
private:
   QBtWorkspace*      const d_workspace;
   Margin*            const margin_;
   Folder*                  folder_;
   bool                     d_is_noname;
   QString                  d_file_path;
   bool                     d_send_events;
   bool                     d_can_parse;
   CppSyntax*               d_syntax;
   QTextDocument::FindFlags d_find_flags;
   QString                  d_find_text;
   QTimer*            const d_timer;
   QDateTime                d_last_modified;
   bool                     use_folder_;

// ******* METHODS *******
public:
   void   set_margin              ();
   void   update_margins          ();
   const QString& file_path       () const;
   void   block_events            ( bool );
   void   new_file                ();
   qint32 cursor_position         () const;
   void   set_cursor_position     ( qint32 );
   void   set_cursor_position     ( int, int );
   void   search_class            ( const QString& in_class_name );
   void   print                   ( bool ) const;
   Margin* margin           () const;
   Folder* folder           () const;
private:
   bool    can_edit                ( qint32 ) const;
   void    apply_bookmarks         ();
   void    apply_breakpoints       ();
   int     row                     () const { return textCursor().block().blockNumber(); }
   QSize   sizeHint                () const;
   void    focusInEvent            ( QFocusEvent* );
   void    resizeEvent             ( QResizeEvent* );
   void    showEvent               ( QShowEvent* );
   void    hideEvent               ( QHideEvent* );
   void    set_window_title        ( const QString& );
   void    closeEvent              ( QCloseEvent* );
   void    customEvent             ( QEvent* );
   void    keyPressEvent           ( QKeyEvent* );
   bool    is_navi_key             ( const qint32 in_key ) const;
   void    set_properties          ();
   void    set_file_name           ( const QString& );
   void    bookmarking             ();
   void    breakpointing           ();
   void    bookmark_next           ();
   void    bookmark_prev           ();
   void    brace_right_key         ( QKeyEvent* );
   void    ctrl_delete_key         ( QKeyEvent* );
   void    backspace_key           ( QKeyEvent* );   
   void    delete_key              ( QKeyEvent* );
   void    return_key              ( QKeyEvent* );
   void    home_key                ( QKeyEvent* );
   void    tab_key                 ( QKeyEvent* );
   void    ins_one_tab             ( QTextCursor& );
   void    ins_tabs_in_selblock    ( QTextCursor& );
   qint32  is_tab_at_begin         ( qint32, const QString& ) const;
   bool    are_tabs_in_selblock    ( const QString&, const QString& ) const;
   void    goto_line               ();
   QString get_selected_text       () const;
   void    replace                 ();
   void    replace_in_whole_document( QTextDocument::FindFlags, const QString&, const QString& );
   void    replace_in_selected_text ( QTextDocument::FindFlags, const QString&, const QString&, qint32, qint32 );
   void    replace_next             ( QTextDocument::FindFlags, const QString&, const QString& );
   void    find_in_document        ();
   void    find_next               ();
   void    find_previous           ();
   void    file_in_project_renamed ( const QString&, const QString& );
   bool    try_save                ();
   void    contextMenuEvent        ( QContextMenuEvent* );
   qint32  search_open_paranthesis ( qint32 );
   qint32  search_closed_marker    ( const QString&, qint32, const QChar&, const QChar& );
   QString tab_chars               ( qint32 );
   QString get_prefix_text         ( const QString& ) const;
   void    restore_folded_text     ( QString& ) const;
public:
   bool    save                    ();
   bool    save_as                 ();
private:
   bool    save_file               ( const QString& );
   bool    write_file              ( const QString& );
public:
   bool    open_file_dialog        ( const QString& );
   bool    open_file               ( const QString& );
private:
   bool    read_file               ( const QString& );
private slots:
   void    do_cut                  ();
   void    do_copy                 ();
   void    do_paste                ();
   void    do_undo                 ();
   void    do_redo                 ();
   void    do_delete               ();
   void    do_select_all           ();
   void    do_run                  ();
   void    do_dbg_run              ();
   void    do_parse_func           ();
   void    contents_changed        ();
   void    copy_available          ( bool );
   void    redo_available          ( bool );
   void    undo_available          ( bool );
   void    check_datetime          ();
public slots:
   void   cursor_position_changed  ( bool = true );
};

inline Margin* Editor::margin() const {
   return margin_;
}
inline Folder* Editor::folder() const {
   return folder_;
}
inline const QString& Editor::file_path() const {
   return d_file_path;
}

inline void Editor::block_events( const bool in_blocked ) {
   d_send_events = ( false == in_blocked );
}

#endif // INCLUDED_Editor_h
