/********************************************************************
 * Copyright (C) 2006, 2007 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : Folder.cpp
 * DATE   : 31.07.2007
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "Folder.h"
#include "Cobras.h"
#include "Editor.h"
#include "Config.h"
#include "Renumer.h"
#include <QPainter>
#include <QScrollBar>
#include <QTextLayout>
#include <QtDebug>
using namespace std;

/*------- local constants:
-------------------------------------------------------------------*/
const QString Folder::FOLD_MARKER = "...";


//*******************************************************************
// Folder                                                CONSTRUCTOR
//*******************************************************************
Folder::Folder( QWidget* const in_parent ) : QWidget( in_parent )
, parser_     ( new QBTParserFunDef( this ) )
, width_      ( 7 )
, can_parse_  ( true )
, can_update_ ( true )
, plain_text_ ( "" )
{
   setAutoFillBackground( true );
   set_properties();
   connect( parser_, SIGNAL( finished() ), this, SLOT( parser_finished() ) );
}
// end of Folder

//*******************************************************************
// update_geometry                                            PUBLIC
//*******************************************************************
void Folder::update_geometry( const qint32 in_offset_x )
{
   Editor* const editor = dynamic_cast<Editor*>( parent() );
   if( editor ) {
      const QRect rect( in_offset_x, 0, width_, editor->viewport()->height() );
      setGeometry( rect );
   }
}
// update_geometry

//*******************************************************************
// set_properties                                            PRIVATE
//-------------------------------------------------------------------
// Folder ma takie kolory jak edytor.
//*******************************************************************
void Folder::set_properties()
{
   const Editor* const editor = dynamic_cast<Editor*>( parent() );
   if( editor ) {
      const QColor bkg_color = editor->viewport()->palette().color( editor->viewport()->backgroundRole() );
      const QColor frg_color = editor->viewport()->palette().color( editor->viewport()->foregroundRole() );

      QPalette p = palette();
      p.setColor( backgroundRole(), bkg_color );
      p.setColor( foregroundRole(), frg_color );
      setPalette( p );
   }
}
// end of set_properties

//*******************************************************************
// on_screen                                                 PRIVATE
//*******************************************************************
bool Folder::on_screen( const qint32 in_top_y, const qint32 in_bottom_y, const qint32 in_y ) const
{
   return ( ( in_y > in_top_y ) && ( in_y < in_bottom_y ) );
}
// end of on_screen

//*******************************************************************
// paintEvent                                      PRIVATE inherited
//*******************************************************************
void Folder::paintEvent( QPaintEvent* )
{
   const Editor* const edt = dynamic_cast<Editor*>( parent() );
   if( edt ) {
      plain_text_ = edt->toPlainText();
      
      const QTextBlock cblock          = edt->textCursor().block();
      const qint32     y_screen_top    = edt->verticalScrollBar()->value();
      const qint32     y_screen_bottom = y_screen_top + edt->viewport()->height() - 1;
      
      QPainter p( this );
      QTextBlock block = edt->document()->begin();
      while( block.isValid() ) {
         // Wspolrzedne aktualnego bloku.
         const qint32 block_yt     = qRound( block.layout()->position().y() );
         const qint32 block_height = block.layout()->boundingRect().toRect().height();
         const qint32 block_yb     = block_yt + block_height - 1;

         // Jezeli gorna czesc bloku jest ponizej ekranu to przerywamy prace.
         // (Szkoda czasu na robienie czegokolwiek dla niewidocznej czesci edytora.)
         if( block_yt > y_screen_bottom ) {
            break;
         }

         // Jesli chocby ulamek aktualnego bloku zawiera sie w widoku
         // to go musimy obsluguzyc.
         if( on_screen( y_screen_top, y_screen_bottom, block_yt ) ||
             on_screen( y_screen_top, y_screen_bottom, block_yb ) ) {

            LineInfo* const info = dynamic_cast<LineInfo*>( block.userData() );
            if( info ) {
               const qint32 marker_type = info->folding_marker;
               if( marker_type != FOLDER_NO_MARKER ) {
                  switch( marker_type ) {
                     case FOLDER_MARKER_BEGIN_OPENED:
                        draw_opened_begin( p, y_screen_top, block_yt, block_yb );
                        break;
                     case FOLDER_MARKER_BEGIN_CLOSED:
                        draw_closed_begin( p, y_screen_top, block_yt, block_yb );
                        break;
                     case FOLDER_MARKER_LINE:
                        draw_line( p, y_screen_top, block_yt, block_yb );
                        break;
                     case FOLDER_MARKER_END:
                        draw_end( p, y_screen_top, block_yt, block_yb );
                        break;
                     default:
                        // taka sytuacja nie moze sie zdarzyc
                        break;
                  }
               }
            }
         }
         block = block.next();
      }
      p.end();
   }
   plain_text_.clear();
}
// end of paintEvent

//*******************************************************************
// update_data                                               PRIVATE
//*******************************************************************
void Folder::update_data()
{
   const Editor* const edt = dynamic_cast<Editor*>( parent() );
   if( edt ) {
      if( data_.empty() ) {
         const QBTCodeBlockInfo empty_cbi;
         QTextBlock block = edt->document()->begin();
         while( block.isValid() ) {
            LineInfo* const info = dynamic_cast<LineInfo*>( block.userData() );
            if( info ) {
               info->folding = empty_cbi;
               info->folding_marker = FOLDER_NO_MARKER;
            }
           block = block.next();
         }
      }
      else {
         QTextBlock block = edt->document()->begin();
         while( block.isValid() ) {
            set_line_info( block );
            block = block.next();
         }
      }
   }
}
// end of update_data

//*******************************************************************
// set_line_info                                             PRIVATE
//*******************************************************************
void Folder::set_line_info( QTextBlock& in_block ) const
{
   LineInfo* const info = dynamic_cast<LineInfo*>( in_block.userData() );
   if( info ) {
      // Pozycja poczatkowa i koncowa tesktu w aktualnym bloku.
      const qint32 sidx = in_block.position();
      const qint32 eidx = sidx + in_block.text().length();

      QBTCodeBlockInfo cbi = QBTCodeBlockInfo();
      qint32 marker = FOLDER_NO_MARKER;
      
      vector<QBTCodeBlockInfo>::const_iterator it = data_.begin();
      const vector<QBTCodeBlockInfo>::const_iterator end = data_.end();
      while( it != end ) {
         const qint32 close_idx = it->close_idx();
         const qint32 pos_idx  = it->pos_idx();

         if( -1 == close_idx ) {
            ++it;
            continue;
         }
         if( close_idx < sidx ) {                  // Jesli koniec bloku jest
            ++it;                                  // przed poczatkiem aktualnego
            continue;                              // bloku to idziemy dalej.
         }
         if( pos_idx > eidx ) {                    // Jesli poczatek bloku znajduje za
            ++it;                                  // za koncem aktualnego bloku
            continue;                              // to idziemy dalej.
         }
         if( this_line( pos_idx, sidx, eidx ) ) {
            if( ( pos_idx >= sidx ) && ( close_idx <= eidx ) ) {
               marker = FOLDER_MARKER_BEGIN_CLOSED;
            }
            else {
               marker = FOLDER_MARKER_BEGIN_OPENED;
            }
            cbi = *it;
            break;
         }
         else if( this_line( close_idx, sidx, eidx ) ) {
            cbi = *it;                             // Koniec bloku zawiara sie 
            marker = FOLDER_MARKER_END;            // wewnatrz aktualnego bloku.
            break;
         }
         else if( ( sidx > pos_idx ) && ( eidx < close_idx ) ) {
            cbi = *it;                             // Aktualny blok jest gdzies w
            marker = FOLDER_MARKER_LINE;           // srodku tego bloku.
            break;
         }
         ++it;
      }
   
      info->folding_marker = marker;
      info->folding = cbi;
   }
}
// end of set_line_info

//*******************************************************************
// is_this_line                                              PRIVATE
//*******************************************************************
bool Folder::this_line( const qint32 in_pos,
                        const qint32 in_start_pos,
                        const qint32 in_end_pos ) const
{
   return ( ( in_pos >= in_start_pos ) && ( in_pos <= in_end_pos ) );
}
// end of is_this_line

//*******************************************************************
// draw_opened_begin                                         PRIVATE
//*******************************************************************
void Folder::draw_opened_begin(  QPainter& in_painter,
                                 const qint32 in_screen_top,
                                 const qint32 in_y_top,
                                 const qint32 in_y_bottom )
{
   const qint32 y = ( in_y_top + ( ( in_y_bottom - in_y_top + 1 ) >> 1 ) )- in_screen_top + 1;

   in_painter.drawLine( 3, y + 1, 3, y - 1 );   // pionowa kreska plusika
   in_painter.drawLine( 2, y, 4, y );           // pozioma kreska plusika
   in_painter.drawRect( 0, y - 3, 6, 6 );       // kwadracik
   in_painter.drawLine( 3, y + 4, 3, in_y_bottom - in_screen_top );
}
// end of draw_opned_begin

//*******************************************************************
// draw_closed_begin                                         PRIVATE
//*******************************************************************
void Folder::draw_closed_begin(  QPainter& in_painter,
                                 const qint32 in_screen_top,
                                 const qint32 in_y_top,
                                 const qint32 in_y_bottom )
{
   const qint32 y = ( in_y_top + ( ( in_y_bottom - in_y_top + 1 ) >> 1 ) )- in_screen_top + 1;
   in_painter.drawLine( 2, y, 4, y );     // pozioma kreska minusika
   in_painter.drawRect( 0, y - 3, 6, 6 ); // kwadracik
}
// end of draw_closed_begin

//*******************************************************************
// draw_line                                                 PRIVATE
//*******************************************************************
void Folder::draw_line( QPainter& in_painter,
                        const qint32 in_screen_top,
                        const qint32 in_y_top,
                        const qint32 in_y_bottom )
{
   in_painter.drawLine( 3, in_y_top - in_screen_top, 3, in_y_bottom - in_screen_top ); 
}
// end of draw_line

//*******************************************************************
// draw_end                                                  PRIVATE
//*******************************************************************
void Folder::draw_end(  QPainter& in_painter,
                        const qint32 in_screen_top,
                        const qint32 in_y_top,
                        const qint32 in_y_bottom )
{
   const qint32 y_center = in_y_top + ( ( in_y_bottom - in_y_top ) >> 1 ) + 1;
   const QPoint start_point ( 3, in_y_top - in_screen_top );
   const QPoint center_point( 3, y_center - in_screen_top );
   const QPoint end_point   ( 6, y_center - in_screen_top );
   in_painter.drawLine( start_point, center_point ); 
   in_painter.drawLine( center_point, end_point );
}
// end of draw_end

//*******************************************************************
// parse                                                      PUBLIC
//*******************************************************************
void Folder::parse( const bool in_can_update )
{
   if( can_parse_ ) {
      const Editor* const edt = dynamic_cast<Editor*>( parent() );
      if( edt && edt->isVisible() ) {
         can_parse_ = false;   
         can_update_ = in_can_update;
         parser_->parse_it( edt->toPlainText(), data_ );
      }
   }
}
// end of parse

//*******************************************************************
// parser_finished                                      PRIVATE slot
//*******************************************************************
void Folder::parser_finished()
{
   parser_->data( data_ );
   can_parse_ = true;
   if( can_update_ && !data_.empty() ) {
      update_data();
      update();
   }
}
// end of parser_finished

//*******************************************************************
// mousePressEvent                                PRIVATE inherited
//*******************************************************************
void Folder::mousePressEvent( QMouseEvent* const in_event )
{
   Editor* const edt = dynamic_cast<Editor*>( parent() );
   if( edt ) {
      QTextBlock block;
      const qint32 y_pos = in_event->y() + edt->verticalScrollBar()->value();
      if( search_text_block( y_pos, block ) ) {
         edt->set_cursor_position( block.position() );
      }
   }
}
// mousePressEvent

//*******************************************************************
// mouseDoubleClickEvent                          PRIVATE inherited
//*******************************************************************
void Folder::mouseDoubleClickEvent( QMouseEvent* const in_event )
{
   Editor* const edt = dynamic_cast<Editor*>( parent() );
   if( edt ) {
      if( Qt::LeftButton == in_event->button() ) {
         QTextBlock block;
         const qint32 y_pos = in_event->y() + edt->verticalScrollBar()->value();
         if( search_text_block( y_pos, block ) ) {
            LineInfo* const info = dynamic_cast<LineInfo*>( block.userData() );
            if( info ) {
               bool need_parser = false;
               if( FOLDER_MARKER_BEGIN_OPENED == info->folding_marker ) {
                  Renumer renumer( edt );
                  need_parser = close_opened_block( edt, block, info->folding );
               }
               else if( FOLDER_MARKER_BEGIN_CLOSED == info->folding_marker ) {
                  Renumer renumer( edt );
                  need_parser = open_closed_block( edt, block );
               }
               if( need_parser ) {
                  parse();
               }
            }
         }
      }
   }
}
// end of mouseDoubleClickEvent

//*******************************************************************
// close_opened_block                                        PRIVATE
//*******************************************************************
bool Folder::close_opened_block( Editor* const           in_edt,
                                 QTextBlock&             in_block,
                                 const QBTCodeBlockInfo& in_cbi )
{
   bool retval = true;
   const qint32 pos = in_block.position();

   // Zaznaczenie i pobranie tekstu bloku.
   QTextCursor cursor = in_edt->textCursor();
   cursor.setPosition( in_cbi.pos_idx() );
   cursor.movePosition( QTextCursor::StartOfBlock );
   cursor.setPosition( in_cbi.close_idx(), QTextCursor::KeepAnchor );
   cursor.movePosition( QTextCursor::EndOfBlock, QTextCursor::KeepAnchor );
   const QString text = cursor.selectedText();

   // Wygenerowanie nowego, zastepczego, sktoconego tekstu  bloku
   // i wpisanie go w miejsce oryginalu.
   QString header = text.left( in_cbi.open_idx() - in_cbi.pos_idx() ).simplified().trimmed();
   const qint32 idx = dwukropek( header );
   if( idx != -1 ) {
      header = header.left( idx );
   }
   header += " {...}";
   cursor.insertText( header );

   // Kursor na poczatek bloku.
   cursor.setPosition( pos );
   in_edt->setTextCursor( cursor );
   // Zapamietanie oby wersji tekstu bloku.
   text_pair_.insert( TextPair::value_type( header, text ) );

   return retval;
}
// end of close_opened_block

//*******************************************************************
// open_closed_block                                         PRIVATE
//*******************************************************************
bool Folder::open_closed_block( Editor* const in_edt, QTextBlock& in_block )
{
   bool retval = false;
   
   const TextPair::const_iterator it = text_pair_.find( in_block.text() );
   if( it != text_pair_.end() ) {
      QTextCursor cursor = in_edt->textCursor();
      const qint32 pos = in_block.position();
      cursor.movePosition( QTextCursor::StartOfBlock );
      cursor.movePosition( QTextCursor::EndOfBlock, QTextCursor::KeepAnchor );
      cursor.insertText( it->second );
      cursor.setPosition( pos );
      in_edt->setTextCursor( cursor );
      text_pair_.erase( it->first );
      retval = true;
   }
   return retval;
}
// end of open_closed_block


//*******************************************************************
// get_pair_text                                              PUBLIC
//*******************************************************************
bool Folder::get_pair_text(   const QString& in_src,
                              QString&       out_dst,
                              bool           in_delete )
{
   out_dst.clear();
   const TextPair::const_iterator it = text_pair_.find( in_src );
   if( it != text_pair_.end() ) {
      out_dst = it->second;
      if( in_delete ) {
         text_pair_.erase( it->first );
      }
   }
   return !out_dst.isEmpty();
}
// end of get_pair_text

//*******************************************************************
// dwukropek                                                 PRIVATE
//*******************************************************************
qint32 Folder::dwukropek( const QString& in_text ) const
{
   qint32 retval = -1;

   qint32 idx = 0;
   const qint32 last_idx = in_text.length() - 1;
   while( 1 ) {
      idx = in_text.indexOf( ':', idx );
      if( -1 == idx ) {
         break;
      }
      else {
         if( idx < last_idx ) {
            if( ':' == in_text[ idx + 1 ] ) {
               idx += 2;
               continue;
            }
            else {
               retval = idx;
               break;
            }
         }
         else {
            retval = idx;
            break;
         }
      }
    }

   return retval;
}
// end of dwukropek

//*******************************************************************
// search_text_block                                         PRIVATE
//-------------------------------------------------------------------
// Wyszukiwanie w tekscie bloku/linii, ktora zawiera w sobie
// przyslana wspolrzedna.
//*******************************************************************
bool Folder::search_text_block ( const qint32 in_y, QTextBlock& out_block ) const
{
   bool retval = false;

   const Editor* const edt = dynamic_cast<Editor*>( parent() );
   if( edt ) {
      // Wspolrzedne gory i dolu widoku zawierajacego tekst w edytorze.
      const qint32 yt_view = edt->verticalScrollBar()->value();
      const qint32 yb_view = yt_view + edt->viewport()->height() - 1;

      QTextBlock block = edt->document()->begin();
      while( block.isValid() ) {
         if( in_y > yt_view ) {
            // Jesli jestesmy juz za widoczna czescia widoku
            // przerywamy prace. I tak juz nic nie ma do pokazania.
            if ( in_y > yb_view ) {
               break;
            }
         
            // Wspolrzedne gory i dolu aktualnego bloku.
            const qint32 yt_block = qRound( block.layout()->position().y() );
            const qint32 yb_block = yt_block + qRound( block.layout()->boundingRect().height() ) - 1;
            // Jezeli przyslana zawiera sie pomiedzy gora i dolem
            // aktualnego bloku, to ten blok jest tym czego szukamy.
            if( ( in_y >= yt_block ) && ( in_y <= yb_block ) ) {
               out_block = block;
               retval = true;
               break;
            }
            block = block.next();
         }
      }
   }
   return retval;
}
// end of search_text_block
