/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : Process.cpp
 * DATE   : 24.09.2006
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "Process.h"
#include "EventsController.h"
#include <QMessageBox>
#include <QtDebug>

/*------- local constants:
-------------------------------------------------------------------*/
const char* const Process::PROCESSING = QT_TR_NOOP( "Processing (%1)" );
const QString     Process::MSG        = "[>] %1";

//*******************************************************************
// Process                                               CONSTRUCTOR
//*******************************************************************
Process::Process()
: QProcess   ()
, commands_  ( QStringList() )
, first_time_( true )
{
	connect( this, SIGNAL( readyReadStandardOutput()    ), this, SLOT( read_stdout()                         ));
	connect( this, SIGNAL( readyReadStandardError()     ), this, SLOT( read_stderr()                         ));
	connect( this, SIGNAL( started()                    ), this, SLOT( process_started()                     ));
	connect( this, SIGNAL( finished(int)                ), this, SLOT( process_finished(int)                 ));
	connect( this, SIGNAL( error(QProcess::ProcessError)), this, SLOT( process_error(QProcess::ProcessError) ));
}
// end of Process

//*******************************************************************
// run                                                        PUBLIC
//-------------------------------------------------------------------
// Funkcja wywolywana z zewnatrz inicjuje wykonanie sekwencji
// polecen. Najpierw wszystkie polecenia sa zapamietywane w commands_,
// a nastepnie uruchamiane jest pierwsze z nich.
//*******************************************************************
void Process::run( const QString& in_wrkdir, const QStringList& in_commands )
{
	commands_   = in_commands;
	first_time_ = true;

	if( !commands_.empty() ) {
		setWorkingDirectory( in_wrkdir );
		start( *( it_ = commands_.begin() ) );
	}
}
// end of run

//*******************************************************************
// read_stdout                                          PRIVATE slot
//-------------------------------------------------------------------
// Odbior informacji o przebiegu wykonania procesu.
// Informacja jest wysylana do zakladki 'Build'.
//*******************************************************************
void Process::read_stdout()
{
	const QString text = QString::fromLocal8Bit( readAllStandardOutput() ).trimmed();
	if( !text.isEmpty() ) {
		EventsController::instance()->send_event( Event::PROCESS_BUILD_DATA, MSG.arg( text ) );
	}
}
// end of read_stdout

//*******************************************************************
// read_stderr                                          PRIVATE slot
//-------------------------------------------------------------------
// Odbior informacji o bledach w wykonaniu procesu.
// Informacja jest wysylana do zakladki 'Messages'.
//*******************************************************************
void Process::read_stderr()
{
	const QString text = QString::fromLocal8Bit( readAllStandardError() ).trimmed();
	if( !text.isEmpty() ) {
		EventsController::instance()->send_event( Event::PROCESS_MESSAGES_DATA, MSG.arg( text ) );
	}
}
// end of read_stdout

//*******************************************************************
// process_started                                      PRIVATE slot
//-------------------------------------------------------------------
// Slot ten jest wywolywany ZA KAZDYM RAZEM gdy uruchomione jest
// kolejny proces. Zeby kolejne procesy nie kasowaly informacji o
// wykonaniu sie poprzedniego, musimy kontrolowac czy jest to
// pierwszy proces, czy kolejny.
//*******************************************************************
void Process::process_started()
{
	if( first_time_ ) {
		EventsController::instance()->send_event( Event::PROCESS_STARTED, workingDirectory() );
		first_time_ = false;
	}
	else {
		EventsController::instance()->send_event( Event::PROCESS_BUILD_DATA_NEWLINE );
	}
	EventsController::instance()->send_event( Event::PROCESS_BUILD_DATA, tr( PROCESSING ).arg( *it_ ) );
}
// end of process_started

//*******************************************************************
// process_finished                                     PRIVATE slot
//*******************************************************************
void Process::process_finished( int in_code )
{
	if( ( 0 == in_code ) && ( NormalExit == exitStatus() ) ) {
		++it_;
		if( !commands_.empty() && ( it_ != commands_.end() ) ) {
			start( *it_ );
		}
		else {
			commands_.clear();
			EventsController::instance()->send_event( Event::PROCESS_BROKEN );
		}
	}
	else {
		commands_.clear();
		EventsController::instance()->send_event( Event::PROCESS_BROKEN );
	}
}
// end of process_finished

//*******************************************************************
// process_error                                        PRIVATE slot
//*******************************************************************
void Process::process_error( QProcess::ProcessError /*in_error*/ )
{
/*
	QString error_info = "Process error: ";
	switch( in_error ) {
		case FailedToStart:
			error_info += "failed to start";
			break;
		case Crashed:
			error_info += "crashed";
			break;
		case Timedout:
			error_info += "timedout";
			break;
		case WriteError:
			error_info += "write error";
			break;
		case ReadError:
			error_info += " read error";
			break;
		default:
			error_info += "unknown error";
	}
	qDebug() << error_info;
*/
	EventsController::instance()->send_event( Event::PROCESS_BROKEN );
}
// end of process_error
