/********************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : QBtParser.cpp
 * DATE   : 20.09.2006
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "QBtParser.h"
#include "QBTCodeBlockInfo.h"
#include <vector>
#include <string>
#include <QtDebug>
using namespace std;

/*------- local constants:
-------------------------------------------------------------------*/
const QChar   QBtParser::UNDER_SCORE     = '_';
const QChar   QBtParser::BACK_SLASH      = '\\';
const QChar   QBtParser::SLASH           = '/';
const QChar   QBtParser::STERN           = '*';
const QChar   QBtParser::HASH            = '#';
const QChar   QBtParser::SINGLE_QUOTE    = '\'';
const QChar   QBtParser::DOUBLE_QUOTE    = '\"';
const QChar   QBtParser::COLON           = ':';
const QChar   QBtParser::SEMICOLON       = ';';
const QChar   QBtParser::BLOCK_BEGIN     = '{';
const QChar   QBtParser::BLOCK_END       = '}';
const QChar   QBtParser::NEW_LINE        = '\n';
const QString QBtParser::C_COMMENT_BEGIN = "/*";
const QString QBtParser::C_COMMENT_END   = "*/";
const QString QBtParser::CPP_COMMENT     = "//";
const QString QBtParser::WHITE_CHARS     = " \b\f\n\r\t\v";
const QString QBtParser::DEC_DIGITS      = "0123456789";
const QString QBtParser::HEX_DIGITS      = "0123456789abcdefABCDEF";


//*******************************************************************
// reverse
//*******************************************************************
void QBtParser::reverse( QString& in_text )
{
   qint32 i   = 0;
   qint32 j   = in_text.length() - 1;
   QChar  tmp = QChar();
   
   while( i < j ) {
      tmp = in_text[ i ];
      in_text[ i ] = in_text[ j ];
      in_text[ j ] = tmp;
      ++i;
      --j;
   }
}

//*******************************************************************
// is_white_char
//*******************************************************************
bool QBtParser::is_white_char( const QChar& in_char )
{
   return ( WHITE_CHARS.indexOf( in_char ) != -1 );
}
// end of is_white_char

//*******************************************************************
// is_white_string
//*******************************************************************
bool QBtParser::is_white_string( const QString& in_text )
{
   const qint32 n = in_text.length();
   for( qint32 i = 0; i < n; ++i ) {
      if( !is_white_char( in_text[ i ] ) ) {
         return false;
      }
   }
   return true;
}
// end of is_white_string

//*******************************************************************
// is_dec_digit
//*******************************************************************
bool QBtParser::is_dec_digit( const QChar& in_char )
{
    return ( DEC_DIGITS.indexOf( in_char ) != -1 );
}
// end of is_dec_digit

//*******************************************************************
// is_hex_digit                                              PRIVATE
//*******************************************************************
bool QBtParser::is_hex_digit( const QChar& in_char )
{
    return ( HEX_DIGITS.indexOf( in_char ) != -1 );
}
// end of is_hex_digits

//*******************************************************************
// is_hex_number                                              PUBLIC
//*******************************************************************
bool QBtParser::is_hex_number( const QString& in_word )
{
   bool retval = false;
   const int length = in_word.length();
   int idx = -1;

        if( ( length > 2 ) && in_word.startsWith( "0x" ) ) idx = 2;
   else if( ( length > 1 ) && in_word.startsWith( "x"  ) ) idx = 1;

   if( idx != -1 ) {
      retval = true;
      while( idx < length ) {
         if( !is_hex_digit( in_word[ idx ] ) ) {
            retval = false;
            break;
         }
         ++idx;
      }
   }
   return retval;
}
// end of is_hex_number

//*******************************************************************
// is_dec_number                                              PUBLIC
//*******************************************************************
bool QBtParser::is_dec_number( const QString& in_word )
{
   bool retval = true;

   for( int i = 0; i < in_word.length(); ++i ) {
      if( !is_dec_digit( in_word[i] ) ) {
         retval = false;
         break;
      }
   }
   return retval;
}
// end of is_dec_number

//*******************************************************************
// is_number                                                  PUBLIC
//*******************************************************************
bool QBtParser::is_number( const QString& in_word )
{
    return ( is_hex_number( in_word ) || is_dec_number( in_word ) );
}
// end of is_number

//*******************************************************************
// is_valid_char                                             PRIVATE
//*******************************************************************
bool QBtParser::is_valid_char( const QChar& in_char )
{
    return ( in_char.isLetterOrNumber() || ( UNDER_SCORE == in_char ));
}
// end of is_valid_char

//*******************************************************************
// is_acceptable                                              PUBLIC
//*******************************************************************
bool QBtParser::is_acceptable( const QString& in_text, const int in_idx )
{
   bool retval = true;
   
   if( ( in_idx > 0 ) && ( BACK_SLASH == in_text[ in_idx - 1 ] ) ) {
      retval = ( ( in_idx >  1 ) && ( BACK_SLASH == in_text[ in_idx - 2 ] ));
   }
   return retval;
}
// end of is_acceptable


//*******************************************************************
// first_after_white_chr                                      PUBLIC
//*******************************************************************
int QBtParser::first_after_white_chr( const QString& in_text, const int in_idx )
{
   int retval = -1;
    
   if( !in_text.isEmpty() ) {
      if( in_idx >= 0 ) {
         for( qint32 i = in_idx; i < in_text.length(); ++i ) {
            if( !is_white_char( in_text[ i ] ) ) {
               retval = i;
               break;
            }
         }
      }
   }
   return retval;
}
// end of first_after_white_chr

//*******************************************************************
// first_white_chr
//*******************************************************************
qint32 QBtParser::first_white_chr( const QString& in_text, const qint32 in_idx )
{
   qint32 retval = -1;
   
   if( !in_text.isEmpty() ) {
      for( qint32 i = in_idx; i < in_text.length(); ++i ) {
         if( is_white_char( in_text[ i ] ) ) {
            retval = i;
            break;
         }
      }
   }
   return retval;
}
// end of first_white_chr

//*******************************************************************
// remove_tokens                                             PRIVATE
//-------------------------------------------------------------------
// Usuwa z podanego lancucha odcinki zaczynajace sie i konczace
// wskazanymi znakami.
//*******************************************************************
void QBtParser::remove_tokens( QString& in_text, const QChar& in_begin_char,  const QChar& in_end_char )
{
   int start_pos = 0;
   
   while( ( start_pos = search_char( in_text, in_begin_char, start_pos ) ) != -1 ) {
      const int end_pos = search_char( in_text, in_end_char, start_pos + 1 );
      if( -1 == end_pos ) {
         break;
      }
      in_text.remove( start_pos, end_pos - start_pos + 1 );
      start_pos = end_pos + 1;
   }
}
//end of remove_tokens

//*******************************************************************
// search_char                                               PRIVATE
//*******************************************************************
qint32 QBtParser::search_char( const QString& in_text, const QChar& in_char, const qint32 in_idx )
{
   qint32       retval = -1;
   qint32       idx    = in_idx;
   const qint32 n      = in_text.length();
   
   while( idx < n ) {
      if( in_text[ idx ] == in_char ) {
         if( is_acceptable( in_text, idx ) ) {
            retval = idx;
            break;
         }
      }
      ++idx;
   }
   return retval;
}
// end of search_char

//*******************************************************************
// search_next_item
//*******************************************************************
qint32 QBtParser::search_next_item( const QString& in_text, const qint32 in_start_pos )
{
   qint32 retval = -1;
   qint32 idx    = in_start_pos;

   while( idx < in_text.length() ) {
      if( !is_white_char( in_text[ idx ] ) ) {
         retval = idx;
         break;
      }
      ++idx;
   }
   return retval;
}
// end of search_next_item

//*******************************************************************
// search_item
//-------------------------------------------------------------------
// Wyszukuje czegokolwiek ograniczonego bialymi znakami.
// 
// WYNIK:
//   inout_idx - na wejsciu przekazuje pozycje poczatkowa poszukiwan,
//               jako wynik zwraca pozycje znalezionej jednostki.
//               Jesli nic nie zostalo znalezione wartosc nie ulega zmianie.
//   out_len   - dlugosc znalezionej jednostki
//*******************************************************************
bool QBtParser::search_item( const QString& in_text, int& inout_idx, int& out_len )
{
   bool retval    = false;
   int  start_idx = -1;
   int  end_idx   = -1;
   int  idx       = inout_idx;

   while( idx < in_text.length() ) {
      if( !is_white_char( in_text[ idx ] ) ) {
         start_idx = idx;
         break;
      }
      ++idx;
   }

   if( start_idx != -1 ) {
      idx = start_idx + 1;
      while( idx < in_text.length() ) {
         if( is_white_char( in_text[ idx ] ) ) {
            end_idx = idx - 1;
            break;
         }
         ++idx;
      }
   }

   if( start_idx != -1 ) {
      inout_idx = start_idx;
      out_len = ( end_idx != -1 ) ? ( end_idx - start_idx + 1 ) : -1;
      retval = true;
   }

   return retval;
}
// end of search_item

//*******************************************************************
// search_word
//*******************************************************************
bool QBtParser::search_word( const QString& in_text, int& inout_idx, int& out_len, const bool in_point )
{
   int start_idx = -1;
   int len = -1;

   int idx = inout_idx;

   while( idx < in_text.length() ) {
      if( is_word_begin( in_text, idx ) ) {
         start_idx = idx;
         idx = search_word_end( in_text, start_idx + 1, in_point );
         if( idx != -1 ) len = idx - start_idx + 1;
         break;
      }
      ++idx;
   }

   if( start_idx != -1 ) inout_idx = start_idx;
   out_len = len;
   return ( start_idx != -1 );
}
// end of search_word

//*******************************************************************
// is_word_begin                                              PUBLIC
//*******************************************************************
bool QBtParser::is_word_begin( const QString& in_text, const int in_idx )
{
   bool retval = false;
   
   if( ( in_idx >= 0 ) && ( in_idx < in_text.length() ) ) {
      if( is_valid_char( in_text[ in_idx ] ) ) {
         if( 0 == in_idx ) {
            retval = true;
         }
         else {
            retval = ( false == is_valid_char( in_text[ in_idx - 1 ] ) );
         }
      }
   }
   
   return retval;
}
// end of is_word_begin

//*******************************************************************
// search_word_end                                            PUBLIC
//*******************************************************************
int QBtParser::search_word_end( const QString& in_text, const int in_idx, const bool is_point )
{
   int retval = -1;
   int idx    = in_idx;
   const int last = in_text.length() - 1;
   
   while( idx <= last ) {
      if( is_valid_char( in_text[ idx ] ) || ( is_point && ( '.' == in_text[ idx ] ) ) ) {
         if( idx == last ) {
            retval = idx;
            break;
         }
      }
      else {
         if( idx > 0 ) {
            retval = idx - 1;
            break;
         }
      }
      ++idx;
   }
   return retval;
}
// end of search_word_end

//*******************************************************************
// search_chars
//-------------------------------------------------------------------
// Funkcja wyszukuje we wskazanym tekscie ('in_text') pierwszego
// wystapienia ktoregos ze znakow zawartych w 'in_chars'.
//
// PARAMETRY:
//   in_text      - bufor z tekstem, w ktorym szukamy wskazanych
//                  znakow,
//   in_start_pos - pozycja startowa w boforze, od ktorej rozpoczynamy
//                  poszukiwania. Poszukiwania trwaja do momentu
//                  osiagniecia konca bufora.
//   in_chars     - zbior znakow, ktore szukamy.
//
// WYNIK:
//   Funkcja zwraca indeks pierwszego wystapienie ktoregos z
//   szukanych znakow. Jesli nie znaleziono w tekscie zadnego ze
//   znakow zostanie zwrocona wartosc -1.
//
// UWAGI:
//   - funkcja omija teksty i znaki,
//   - funkcja NIE rozpoznaje komentarzy
//*******************************************************************
qint32 QBtParser::search_chars( const QString& in_text, const qint32 in_start_pos, const QString& in_chars )
{
   Q_ASSERT( in_start_pos >= 0 );
   
   qint32 retval = -1;

   if( !in_text.isEmpty() && !in_chars.isEmpty() && ( in_start_pos >= 0 ) ) {
      const qint32 n           = in_text.length();
      qint32      idx          = in_start_pos;
      bool        single_quote = false;
      bool        double_quote = false;
      QChar       znak         = QChar();
   
      while( idx < n ) {
         znak = in_text[ idx ];
         
         // omija single quote (znaki)
         if( single_quote ) {
            if( ( SINGLE_QUOTE == znak ) && ( is_acceptable( in_text, idx ) ) ) {
               single_quote = false;
            }
            ++idx;
            continue;
         }
         // omija double quote (lancuchy tekstowe)
         if( double_quote ) {
            if( ( DOUBLE_QUOTE == znak ) && ( is_acceptable( in_text, idx ) ) ) {
               double_quote = false;
            }
            ++idx;
            continue;
         }
         // poczatek single quote ?
         if( ( SINGLE_QUOTE == znak ) && is_acceptable( in_text, idx )  ) {
            single_quote = true;
            ++idx;
            continue;
         }
         // poczatek double quote ?
         if( ( DOUBLE_QUOTE == znak ) && is_acceptable( in_text, idx ) ) {
            double_quote = true;
            ++idx;
            continue;
         }
         
         if( in_chars.indexOf( znak ) != -1 ) {
            retval = idx;
            break;
         }
         ++idx;
      }
   }
   return retval;
}
// end of search_chars

//*******************************************************************
// starts_with
//-------------------------------------------------------------------
// Sprawdza czy znaki w buforze poczawszy od wkazanej pozycji sa
// zgodne ze wskazanym wzorcem.
//*******************************************************************
bool QBtParser::starts_with( const QString& in_text, qint32 in_start_pos, const QString& in_muster )
{
   Q_ASSERT( in_start_pos >= 0 );
   
   bool retval = false;

   if( !in_text.isEmpty() && !in_muster.isEmpty() && ( in_start_pos >= 0 ) ) {
      const qint32 n = in_muster.length();
      retval = ( 0 == QString::compare( in_muster, in_text.mid( in_start_pos, n ), Qt::CaseSensitive ) );
   }
   return retval;
}
// end of starts_with

//*******************************************************************
// search_muster
//-------------------------------------------------------------------
// Wyszukiwanie wzorca.
//   
// PARAMETRY:
//   in_text       - bufor tekstowy, ktory bedziemy przeszukiwac,
//   in_muster     - poszukiwany wzorzec,
//   in_start_pos  - pozycja startowa poszukiwan (w przyslanym buforze),
//
// WYNIK:
//   Funkcja zwraca pozycje, na ktorej znaleziono wzorzec
//   lub -1 jesli go nie znaleziono.
//
// UWAGI:
//   - funkcja rozpoznaje i omija komentarze,
//   - funkcja rozpoznaje i omija lancuchy tekstowe ("xyz") i znaki ('x'),
//*******************************************************************
qint32 QBtParser::search_muster (
      const QString& in_text,       // gdzie szukamy
      const QString& in_muster,     // czego szukamy
      const qint32   in_start_pos ) // od jakiej pozycji w 'in_text'
{
   Q_ASSERT( in_start_pos >= 0 );
   
   qint32 retval = -1;

   if( !in_text.isEmpty() && !in_muster.isEmpty() && ( in_start_pos >= 0 ) ) {
      const qint32 n                 = in_text.length();   // liczba znakow w buforze
      const qint32 m                 = in_muster.length(); // liczba znakow wzorca
      const qint32 last              = n - m;              // ostatni znak bufora do sprawdzenia
      const bool   one_char          = ( 1 == m );
      const QChar  mznak             = ( one_char ) ? in_muster[ 0 ] : QChar();
      const bool   check_c_comment   = ( C_COMMENT_BEGIN != in_muster ) && ( C_COMMENT_END != in_muster );
      const bool   check_cpp_comment = ( CPP_COMMENT != in_muster );
      const bool   check_squote      = !( one_char && ( SINGLE_QUOTE == mznak ) );
      const bool   check_dquote      = !( one_char && ( DOUBLE_QUOTE == mznak ) );
      qint32       idx               = in_start_pos;
      bool         cpp_comment       = false;
      bool         c_comment         = false;
      bool         single_quote      = false;
      bool         double_quote      = false;
      QChar        znak              = QChar();
      QChar        nznak             = QChar();
   
      while( idx <= last ) {
         znak = in_text[ idx ];
   
         //###################################################
         //#                                                 #
         //#   OMIJAMY BLOKI SPECJALNE I SZUKAMY ICH KONCA   #
         //#                                                 #
         //###################################################
   
         // czekamy na koniec komentarza c ( /* ... */ )
         if( check_c_comment && c_comment ) {
            if( ( STERN == znak ) && ( idx < last ) && ( SLASH == in_text[ idx + 1 ] ) ) {
               c_comment = false;
               idx += 2;
               continue;
            }
            ++idx;
            continue;
         }
         // czekamy na koniec komentarza cpp ( // ... \n )
         if( check_cpp_comment && cpp_comment ) {
            if( NEW_LINE == znak ) {
               cpp_comment = false;
            }
            ++idx;
            continue;
         }
         // czekamy na koniec single quote
         if( check_squote && single_quote ) {
            if( ( SINGLE_QUOTE == znak ) && is_acceptable( in_text, idx ) ) {
               single_quote = false;
            }
            ++idx;
            continue;
         }
         // czekamy na koniec double quote (lancuch tekstowy)
         if( check_dquote && double_quote ) {
            if( ( DOUBLE_QUOTE == znak ) && is_acceptable( in_text, idx ) ) {
               double_quote = false;
            }
            ++idx;
            continue;
         }
   
         //###################################################
         //#                                                 #
         //#      KONTROLA POCZATKOW BLOKOW SPECJANLNYCH     #
         //#                                                 #
         //###################################################
   
         // poczatek single quote ?
         if( check_squote && ( SINGLE_QUOTE == znak ) && is_acceptable( in_text, idx )  ) {
            single_quote = true;
            ++idx;
            continue;
         }
         // poczatek double quote ?
         if( check_dquote && ( DOUBLE_QUOTE == znak ) && is_acceptable( in_text, idx ) ) {
            double_quote = true;
            ++idx;
            continue;
         }
         // poczatek jakiegos komentarza ?
         if( ( SLASH == znak ) && ( idx < last ) ) {
            nznak = in_text[ idx + 1];
            // czy komentarz c ?
            if( check_c_comment && ( STERN == nznak ) ) {
               c_comment = true;
               idx += 2;
               continue;
            }
            // czy komentarz cpp
            if( check_cpp_comment && ( SLASH == nznak ) ) {
               cpp_comment = true;
               idx += 2;
               continue;
            }
         }
   
         //###################################################
         //#                                                 #
         //#            I O TO TUTAJ CHODZI                  #
         //#                                                 #
         //###################################################
         if( ( one_char ) ? ( mznak == znak ) : ( in_muster == in_text.mid( idx, m ) ) ) {
            retval = idx;
            break;
         }
         ++idx;
      }
   }
   return retval;
}
// end of search_muster

//*******************************************************************
// search_marker_end
//-------------------------------------------------------------------
// Wyszukiwanie konca jakiegos logicznego bloku znakow.
//   
// PARAMETRY:
//   in_text       - bufor tekstowy, ktory bedziemy przeszukiwac,
//   in_start_idx  - pozycja startowa poszukiwan (w przyslanym buforze),
//   in_mark_begin - znacznik oznaczajacy poczatek bloku,
//   in_mark_end   - znacznik oznaczajacy koniec bloku.
//
// WYNIK:
//   Funkcja zwraca pozycje znaku bedacego koncem bloku,
//   lub -1 jesli konca bloku nie znaleziono.
//
// UWAGI:
//   - zaklada sie, ze znaczniki bloku sa JEDNOZNAKOWE
//     np. '{' i '}, '(' i ')',
//   - bloki moga sie zaglebiac,
//   - zaklada sie, ze pozycja startowa (in_start_idx) znajduje sie
//     JUZ WEWNATRZ bloku, ktorego konca wlasnie szukamy,
//   - funkcja rozpoznaje i omija lancuchy tekstowe ("xyz")
//     i znaki ('x'),
//*******************************************************************
qint32 QBtParser::search_marker_end (
   const QString& in_text,         // gdzie szukamy
   const qint32   in_start_pos,    // od jakiej pozycji zaczynamy poszukiwania
   const QChar&   in_mark_begin,   // znak poczatku 'czegos'
   const QChar&   in_mark_end,     // znak konca 'czegos'
   qint32*        out_alt_idx  )   
{
   Q_ASSERT( in_start_pos >= 0 );
   
   qint32 retval = -1;
   
   if( !in_text.isEmpty() && ( in_start_pos >= 0 ) ) {
      const qint32 n            = in_text.length();
      const qint32 last         = n - 1;
      qint32       block_level  = 1;   // juz jestesmy w wewnatrz bloku
      qint32       open_idx     = in_start_pos - 1;
      qint32       close_idx    = -1;
      qint32       prv_close_idx= -1;
      bool         single_quote = false;
      bool         double_quote = false;
      bool         cpp_comment  = false;
      bool         c_comment    = false;
      qint32       idx          = in_start_pos;
      QChar        znak         = QChar();
      QChar        nznak        = QChar();
   
      while( idx <= last ) {
         znak = in_text[ idx ];
   
         // omijamy komentarz c
         if( c_comment ) {
            if( ( STERN == znak ) && ( idx < last ) && ( SLASH == in_text[ idx + 1 ] ) ) {
               c_comment = false;
               idx += 2;
               continue;
            }
            ++idx;
            continue;
         }
         // omijamy komentarz cpp
         if( cpp_comment ) {
            if( NEW_LINE == znak ) {
               cpp_comment = false;
            }
            ++idx;
            continue;
         }
         // omijamy single quote (znaki)
         if( single_quote ) {
            if( ( SINGLE_QUOTE == znak ) && is_acceptable( in_text, idx ) ) {
               single_quote = false;
            }
            ++idx;
            continue;
         }
         // omijamy double quote (lancuchy tekstowe)
         if( double_quote ) {
            if( ( DOUBLE_QUOTE == znak ) && is_acceptable( in_text, idx ) ) {
               double_quote = false;
            }
            ++idx;
            continue;
         }
   
         // poczatek single quote ?
         if( ( SINGLE_QUOTE == znak ) && is_acceptable( in_text, idx )  ) {
            single_quote = true;
            ++idx;
            continue;
         }
         // poczatek double quote ?
         if( ( DOUBLE_QUOTE == znak ) && is_acceptable( in_text, idx ) ) {
            double_quote = true;
            ++idx;
            continue;
         }
         // poczatek jakiegos komentarza ?
         if( ( SLASH == znak ) && ( idx < last ) ) {
            nznak = in_text[ idx + 1 ];
            // czy komentarz c ?
            if( STERN == nznak ) {
               c_comment = true;
               idx += 2;
               continue;
            }
            // czy komentarz cpp
            if( SLASH == nznak ) {
               cpp_comment = true;
               idx += 2;
               continue;
            }
         }
   
         // znacznik konca bloku (niekoniecznie mojego)
         if( in_mark_end == znak ) {
            prv_close_idx = close_idx;
            close_idx = idx;
            --block_level;
            if( 0 == block_level ) { // a jednak mojego :)
               retval = idx;
               break;
            }
         }
         // znacznik poczatku bloku (zaglebionego)
         else if( in_mark_begin == znak ) {
            open_idx = idx;
            ++block_level;
         }
         ++idx;
      }
      
      if( out_alt_idx ) {
         if( 0 == block_level ) {
            *out_alt_idx = close_idx;
         }
         else {
            *out_alt_idx = ( prv_close_idx > open_idx ) ? prv_close_idx : open_idx;
         }
      }
   }

   return retval;
}
// end of search_marker_end

//*******************************************************************
// search_code_block
//*******************************************************************
bool QBtParser::search_code_block( const QString& in_text, const qint32 in_pos, QBTCodeBlockInfo& out_info )
{
   Q_ASSERT( in_pos >= 0 );

   bool retval = false;
   out_info.clear();
   
   if( ( in_pos >= 0 ) && ( !in_text.isEmpty() ) ) {
      QString      head        = "";
      const qint32 n           = in_text.length();   // liczba znakow w buforze
      const qint32 last        = n - 1;              // ostatni znak bufora do sprawdzenia
      qint32       idx         = in_pos;
      bool         cpp_comment = false;
      bool         c_comment   = false;
      bool         preprocesor = false;
      QChar        znak        = QChar();
      QChar        nznak       = QChar();
   
      while( idx <= last ) {
         znak = in_text[ idx ];
   
         // czekamy na koniec komentarza c ( /* ... */ )
         if( c_comment ) {
            if( ( STERN == znak ) && ( idx < last ) && ( SLASH == in_text[ idx + 1 ] ) ) {
               c_comment = false;
               head += "*/";
               idx += 2;
               continue;
            }
            head += znak;
            ++idx;
            continue;
         }
         // czekamy na koniec komentarza cpp ( // ... \n )
         if( cpp_comment ) {
            if( NEW_LINE == znak ) {
               cpp_comment = false;
            }
            head += znak;
            ++idx;
            continue;
         }
         // czekamy na koniec instrukcji preprocesora
         if( preprocesor ) {
            if( NEW_LINE == znak ) {
               preprocesor = false;
            }
            ++idx;
            continue;
         }

         //###################################################
         //#                                                 #
         //#      KONTROLA POCZATKOW BLOKOW SPECJANLNYCH     #
         //#                                                 #
         //###################################################

         // poczatek instrukcji preprocesora?
         if( '#' == znak ) {
            preprocesor = true;
            ++idx;
            continue;
         }
         
         // poczatek jakiegos komentarza ?
         if( ( SLASH == znak ) && ( idx < last ) ) {
            nznak = in_text[ idx + 1];
            // czy komentarz c ?
            if( ( STERN == nznak ) ) {
               c_comment = true;
               head += "/*";
               idx += 2;
               continue;
            }
            // czy komentarz cpp
            if( ( SLASH == nznak ) ) {
               cpp_comment = true;
               head += "//";
               idx += 2;
               continue;
            }
         }
         
         //###################################################
         //#                                                 #
         //#            I O TO TUTAJ CHODZI                  #
         //#                                                 #
         //###################################################
         
         if( ';' == znak ) {
            head.clear();
            ++idx;
            continue;
         }
         head += znak;
         
         if( '{' == znak ) {
            const qint32 n    = head.length();
            const qint32 hidx = validate_header( head );
            if( hidx != -1 ) {
               out_info.set_header( head );
               out_info.set_pos_idx( idx - n + 1 + hidx );
               out_info.set_open_idx( idx );
               out_info.set_close_idx( search_marker_end( in_text, idx + 1, '{', '}' ) );
               retval = true;
               break;
            }
            head.clear();
            idx = search_marker_end( in_text, idx + 1, '{', '}' );
            if( -1 == idx ) break;
         }
         
         ++idx;
      }
   }
   return retval;
}
// end of search_code_block

//*******************************************************************
// validate_header
//-------------------------------------------------------------------
// Szukamy pelnej nazwy funkcji.
// Konczy sie ona po napotkaniu ':' (konstruktory) lub 
// nawiasu otwierajacego '{' blok ciala funkcji (normalna sytuacja).
//*******************************************************************
qint32 QBtParser::validate_header( QString& inout_text )
{
   //qDebug() << inout_text;
   
   qint32 retval = -1;
   const qint32 n = inout_text.length();
   
   QString buffer         = "";
   qint32  idx            = 0;
   qint32  open_level     = 0;
   bool    is_c_comment   = false;
   bool    is_cpp_comment = false;
   
   while( idx < n ) {
      const QChar znak = inout_text[ idx ];
      const QChar nznak = ( idx < ( n - 1 ) ) ? inout_text[ idx + 1 ] : QChar();

      // zamykamy komentarz c
      if( is_c_comment ) {
         if( ( STERN == znak ) && ( SLASH == nznak ) ) {
            if( open_level ) {
               buffer += "*/";
            }
            is_c_comment = false;
            idx += 2;
            continue;
         }
         if( open_level ) {
            buffer += znak;
         }
         ++idx;
         continue;
      }
      // omijamy komentarz cpp
      if( is_cpp_comment ) {
         if( NEW_LINE == znak ) {
            is_cpp_comment = false;
         }
         if( open_level ) {
            buffer += znak;
         }
         ++idx;
         continue;
      }
      // poczatek jakiegos komentarza ?
      if( SLASH == znak )  {
         // czy komentarz c ?
         if( STERN == nznak ) {
            is_c_comment = true;
            if( open_level ) {
               buffer += "/*";
            }
            idx += 2;
            continue;
         }
         // czy komentarz cpp
         if( ( SLASH == nznak ) ) {
            is_cpp_comment = true;
            if( open_level ) {
               buffer += "//";
            }
            idx += 2;
            continue;
         }
         ++idx;
         buffer += znak;
         continue;
      }
      if( '(' == znak ) {
         buffer += znak;
         ++open_level;
         ++idx;
         continue;
      }
      if( ')' == znak ) {
          buffer += znak;
         --open_level;
         ++idx;
         continue;
      }
      if( open_level ) {
         buffer += znak;
         ++idx;
         continue;
      }
      if( ':' == znak ) {
         if( ':' == nznak ) {
            buffer += "::";
            idx += 2;
            continue;
         }
         break;
      }
      if( '{' == znak ) {
         break;
      }
      if( buffer.isEmpty() ) {
         if( !is_white_char( znak ) ) {
            buffer += znak;
         }
      }
      else {
         buffer += znak;
      }
      ++idx;
   }

   if( get_function( buffer ) ) {
      retval = inout_text.indexOf( buffer );
      inout_text = buffer;
   }

   return retval;
}
// end of validate_header

//*******************************************************************
// get_function
//*******************************************************************
bool QBtParser::get_function( QString& inout_text )
{
   if( inout_text.isEmpty() ) return false;

   static const QChar OPEN  = '(';
   static const QChar CLOSE = ')';
   
   QString buffer = "";
   qint32  idx    = inout_text.length() - 1;
   qint32  level  = 0;
   QChar   znak   = QChar();

   // Szukamy nawiasu zamykajacego
   while( idx >= 0 ) {
      if( CLOSE == inout_text[ idx ] ) {
         break;
      }
      --idx;
   }
   if( idx < 0 ) return false;
   
   // Teraz szukamy pasujacego do pary nawiasu otwierajacego.
   // Pasujacego, oznacza ze zakladamy mozliwosc zagniezdzonych
   // par nawiasow.
   while( idx >= 0 ) {
      znak = inout_text[ idx ];
      if( CLOSE == znak ) {
         ++level;
         --idx;
         buffer += znak;
         continue;
      }
      if( OPEN == znak ) {
         --level;
         --idx;
         buffer += znak;
         if( 0 == level ) {
            break;
         }
      }
      buffer += znak;
      --idx;
   }
   if( idx < 0 ) return false;
   
   // Jestesmy przed odpowiednim nawiasem otwierajacym.
   // Teraz omijamy ewentualne biale znaki, ktore moga
   // sie znajdowac pomiedzy nazwa funckji a nawiasem
   // otwierajacym.
   while( idx >= 0 ) {
      znak = inout_text[ idx ];
      if( !is_white_char( znak ) ) {
         break;
      }
      buffer += znak;
      --idx;
   }
   if( idx < 0 ) return false;
   
   // Zaczyna sie nazwa funkcji.
   // Zakladamy, ze ciagnie sie ona do nastepnego nawiasu zamykajacego.
   // Lub otwierajacego - co oczywiscie bedzie bledem.
   while( idx >= 0 ) {
      znak = inout_text[ idx ];
      if( ( OPEN == znak ) || ( CLOSE == znak ) ) {
         break;
      }
      if( '{' == znak ) return false;
      if( '}' == znak ) return false;
      buffer += znak;
      --idx;
   }
   
   // W buforze mamy ostatnia z funkcji.
   reverse( buffer );
   inout_text = buffer.trimmed();


   return true;
}
// end of get_function
