/*******************************************************************
 * Copyright (C) 2006 Piotr Pszczolkowski
 *-------------------------------------------------------------------
 * This file is part of Cobras.
 *
 * Cobras is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Cobras is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Cobras; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *-------------------------------------------------------------------
 * AUTHOR : Piotr Pszczolkowski (piotr@beesoft.org)
 * PROJECT: Cobras
 * FILE   : QBtPrinter.cpp
 * DATE   : 2007.12.14
 *******************************************************************/

/*------- include files:
-------------------------------------------------------------------*/
#include "QBtPrinter.h"
#include <QBrush>
#include <QColor>
#include <QTextEdit>
#include <QPainter>
#include <QPrintDialog>
#include <QApplication>
#include <QFontMetrics>
#include <QTextDocumentFragment>

/*------- constants:
-------------------------------------------------------------------*/
const qint32      QBtPrinter::DEFAULT_LEFT_MARGIN_OFFSET = 20;
const char* const QBtPrinter::FILE    = QT_TR_NOOP( "File: %1" );
const char* const QBtPrinter::PAGE    = QT_TR_NOOP( "Page %1/%2" );
const char* const QBtPrinter::CREATOR = QT_TR_NOOP( "Printed by Cobras" );


//*******************************************************************
// QBtPrinter                                    CONSTRUCTOR private
//*******************************************************************
QBtPrinter::QBtPrinter() : QObject()
, font_              ( QFont() )
, file_path_         ( QString() )
, page_x_            ( qint32() )
, page_y_            ( qint32() )
, page_dx_           ( qint32() )
, page_dy_           ( qint32() )
, x_                 ( qint32() )
, y_                 ( qint32() )
, left_margin_offset_( DEFAULT_LEFT_MARGIN_OFFSET )
, org_margin_        ( qint32() )
{
   setParent( qApp );
}
// end of QBtPrinter

//*******************************************************************
// ~QBtPrinter                                    DESTRUCTOR private
//*******************************************************************
QBtPrinter::~QBtPrinter()
{}
// end of ~QBtPrinter

//*******************************************************************
// instance                                            PUBLIC static
//*******************************************************************
QBtPrinter* QBtPrinter::instance()
{
   static QBtPrinter* ptr = 0;
   if( !ptr ) ptr = new QBtPrinter;
   return ptr;
}
// end of instance

//*******************************************************************
// print                                                      PUBLIC
//*******************************************************************
void QBtPrinter::print( const QString&         in_file_path,
                        const QTextEdit* const in_editor,
                        const QFont&           in_font,
                        const bool             in_fast )
{
   static bool first_time = true;
   bool use_dialog = false;

   //===== Print dialog =====
   if( first_time ) {
      use_dialog = true;
      first_time = false;
   }
   else {
      use_dialog = !in_fast;
   }
   if( use_dialog ) {
      QPrintDialog dialog( &printer_, 0 );
      dialog.setEnabledOptions( QPrintDialog::PrintToFile
                              | QPrintDialog::PrintSelection
                              | QPrintDialog::PrintPageRange
                              | QPrintDialog::PrintCollateCopies );
      dialog.setMinMax( 1, 1000 );
      if( in_editor->textCursor().hasSelection() ) {
         dialog.setPrintRange( QPrintDialog::Selection );
      }
      else {
         dialog.setPrintRange( QPrintDialog::AllPages );
      }
      if( !dialog.exec() ) {
          return;
      }
   }

   //===== Inicjacja zmiennych =====
   init();
   file_path_ = in_file_path;
   font_ = in_font;
   font_.setPointSize( font_.pointSize() - 1 );
   QFontMetrics metrics( font_ );

   //======= Parametry wydruku dla wydruku wszystkich stron =====
   QStringList lines = in_editor->toPlainText().split( '\n' );
   paginate( lines, metrics, true );
   qint32 total_lines = lines.size();

   qint32 print_area_dy = page_dy_ - ( header_height() << 1 );
   qint32 lines_on_page = print_area_dy / metrics.lineSpacing();
   qint32 pages_number = total_lines / lines_on_page;
   if( total_lines % lines_on_page ) ++pages_number;
   qint32 first_page_nr = 0;
   qint32 last_page_nr = pages_number - 1;

   //===== Modyfikacja parametrow wydruku ze wzgledu na wybrany zakres =====
   if( QPrinter::PageRange == printer_.printRange() ) {
      const qint32 min = qMin( printer_.fromPage(), printer_.toPage() ) - 1;
      const qint32 max = qMax( printer_.fromPage(), printer_.toPage() ) - 1;
      if( min > last_page_nr ) return;
      if( max > last_page_nr ) return;
      first_page_nr = min;
      last_page_nr  = max;
   }
   //===== Modyfikacja parametrow wydruku ze wzgledu na zaznaczony tekst =====
   else if( QPrinter::Selection == printer_.printRange() ) {
      lines = in_editor->textCursor().selection().toPlainText().split( '\n' );
      paginate( lines, metrics, true );
      total_lines = lines.size();
      print_area_dy = page_dy_ - ( header_height() << 1 );
      lines_on_page = print_area_dy / metrics.lineSpacing();
      pages_number = total_lines / lines_on_page;
      if( total_lines % lines_on_page ) ++pages_number;
      first_page_nr = 0;
      last_page_nr = pages_number - 1;
   }
   
   const qint32 copies_number = printer_.numCopies();

   //===== Drukowanie =====
   painter_.begin( &printer_ );
   painter_.setFont( font_ );

   for( qint32 copie_nr = 0; copie_nr < copies_number; ++copie_nr ) {
      if( QPrinter::FirstPageFirst == printer_.pageOrder() ) {
         for( qint32 i = first_page_nr; i <= last_page_nr; ++i ) {
            x_ = ( ( i + 1 ) % 2 )  ? ( org_margin_ + left_margin_offset_ )
                                    :   org_margin_;
            print_header();
      
            qint32 nr = i * lines_on_page;
            qint32 n = qint32();

            while( ( nr < total_lines ) && ( n < lines_on_page ) ) {
               y_ += print_text( lines.at( nr ), metrics );
               ++nr;
               ++n;         
            }
      
            print_footer( i + 1, pages_number );
            if( i < last_page_nr ) {
               printer_.newPage();
            }
         }
      }
      else if( QPrinter::LastPageFirst == printer_.pageOrder() ) {
         for( qint32 i = last_page_nr; i >= first_page_nr; --i ) {
            print_header();
      
            qint32 nr = i * lines_on_page;
            qint32 n = qint32();
            while( ( nr < total_lines ) && ( n < lines_on_page ) ) {
               y_ += print_text( lines.at( nr ), metrics );
               ++nr;
               ++n;         
            }
      
            print_footer( i + 1, pages_number );
            if( i > 0 ) {
               printer_.newPage();
            }
         }
      }
   }
   
   painter_.end();
}
// end of print

//*******************************************************************
// paginate                                                  PRIVATE
//*******************************************************************
void QBtPrinter::paginate( QStringList&        inout_lines,
                           const QFontMetrics& in_metrics,
                           const bool          in_wrap )
{
   QStringList result = QStringList();
   
   QStringList::const_iterator it = inout_lines.begin();
   const QStringList::const_iterator end = inout_lines.end();
   while( it != end ) {
      QString text = *it;
      trimm_end( text );
      if( in_wrap ) {
         result << clip_line( text, in_metrics );
      }
      else {
         result << text;
      }
      ++it;
   }
   
   inout_lines = result;
}
// end of paginate

//*******************************************************************
// clip_line                                                 PRIVATE
//*******************************************************************
QStringList QBtPrinter::clip_line(  const QString&      in_text,
                                    const QFontMetrics& in_metrics ) const
{
   QStringList result = QStringList();

   QString text = in_text;
   trimm_end( text );
   
   while( in_metrics.width( text ) >= page_dx_ ) {
      const qint32 n = text.length() - 1;
      qint32 idx = -1;
      for( qint32 i = n; i >= 0; --i ) {
         if( in_metrics.width( text, i ) < page_dx_ ) {
            idx = i;
            break;
         }
      }
      if( idx >= 0 ) {
         for( qint32 i = idx; i >= 0 ; --i ) {
            if( text.at( i ).isSpace() ) {
               idx = i;
               break;
            }
         }
         
         QString tmp = text.left( idx );
         trimm_end( tmp );
         result << tmp;
         
         text = text.mid( idx ).trimmed();
         continue;
      }
      break;
   }
   result << text;

   return result;
}
// end of clip_line

//*******************************************************************
// init                                                      PRIVATE
//*******************************************************************
void QBtPrinter::init()
{
   const QRect r = printer_.pageRect();
   //x_ = page_x_  = ( r.x() + left_margin_offset_ );
   org_margin_  = r.x();
   y_ = page_y_ = r.y();
   page_dx_ = r.width () - ( ( org_margin_ << 1 ) + left_margin_offset_ );
   page_dy_ = r.height() - ( y_ << 1 );
}
// end of init

//*******************************************************************
// header_height                                             PRIVATE
//*******************************************************************
qint32 QBtPrinter::header_height()
{
   
   QFont header_font = font_;
   header_font.setPointSize( font_.pointSize() - 1 );
   const QFontMetrics metrics( header_font );
   return ( metrics.lineSpacing() << 1 );
}
// end of header_height

//*******************************************************************
// print_header                                              PRIVATE
//*******************************************************************
void QBtPrinter::print_header()
{
   painter_.save();

   QFont header_font = font_;
   header_font.setPointSize( font_.pointSize() - 1 );
   painter_.setFont( header_font );

   const QFontMetrics metrics( header_font );
   const QString header_text = metrics.elidedText( tr( FILE ).arg( file_path_ ),
                                                   Qt::ElideMiddle,
                                                   page_dx_ );

   y_ = page_y_;
   {
      QPen pen( QColor( 110, 110, 110 ) );
      painter_.setPen( pen );
      y_ += print_text( header_text, metrics );
   }
   {
      QPen pen( QColor( 128, 128, 128 ) );
      painter_.setPen( pen );
      y_ += print_line( metrics );
   }
   
   painter_.restore();
}
// end of print_header

//*******************************************************************
// print_footer                                              PRIVATE
//*******************************************************************
void QBtPrinter::print_footer( const qint32 in_counter, const qint32 in_number )
{
   painter_.save();

   QFont header_font = font_;
   header_font.setPointSize( font_.pointSize() - 1 );
   painter_.setFont( header_font );
   const QFontMetrics metrics( header_font );

   y_ = page_y_ + page_dy_ - metrics.lineSpacing();
   {
      QPen pen( QColor( 110, 110, 110 ) );
      painter_.setPen( pen );
      qint32 creator_align = Qt::AlignLeft;
      qint32 page_align = Qt::AlignRight;
      if( printer_.doubleSidedPrinting() ) {
         if( 0 == ( in_counter % 2 ) ) {
            creator_align = Qt::AlignRight;
            page_align = Qt::AlignLeft;
         }
      }
      print_text( tr( CREATOR ), metrics, creator_align );
      print_text( tr( PAGE ).arg( in_counter ).arg( in_number ), metrics, page_align );
   }
   {
      QPen pen( QColor( 128, 128, 128 ) );
      painter_.setPen( pen );
      print_line( metrics );
      print_line( metrics );
   }
   
   painter_.restore();
}
// end of print_footer

//*******************************************************************
// print_text                                                PRIVATE
//*******************************************************************
qint32 QBtPrinter::print_text(   const QString&      in_text,
                                 const QFontMetrics& in_metrics,
                                 const qint32        in_align )
{
   const qint32 flags = Qt::TextSingleLine | Qt::AlignTop | in_align;
   const qint32 text_height = in_metrics.height();

   const QRect r( x_, y_, page_dx_, text_height );
   painter_.drawText( r, flags, in_text );

   return in_metrics.lineSpacing();
}
// end of print_text

//*******************************************************************
// print_line                                                PRIVATE
//*******************************************************************
qint32 QBtPrinter::print_line( const QFontMetrics& in_metrics )
{
   painter_.drawLine( x_, y_, x_ + page_dx_ - 1, y_ );
   return in_metrics.lineSpacing();
}
// end of print_line

//*******************************************************************
// trimm_end                                                 PRIVATE
//*******************************************************************
void QBtPrinter::trimm_end( QString& inout_text ) const
{
   if( inout_text.isEmpty() ) return;
   
   const qint32 n = inout_text.length() - 1;
   qint32 counter = 0;
   
   for( qint32 i = n; i >= 0; --i ) {
      if( inout_text.at( i ).isSpace() ) {
         ++counter;
         continue;
      }
      break;
   }
   if( counter > 0 ) {
      inout_text.chop( counter );
   }
}
// end of trimm_end

//*******************************************************************
// print_config                                              PRIVATE
//*******************************************************************
/*
void QBtPrinter::print_config() const
{
   static const QString COLOR_MODE  [] = { "GrayScale", "Color" };
   static const QString ORIENTATION [] = { "Portrait", "Landscape" };
   static const QString FORMAT      [] = { "Native format", "PDF format", "PostScriptFormat" };
   static const QString ORDER       [] = { "FirstPageFirst", "LastPageFirst" };
   static const QString RANGE       [] = { "AllPages", "Selection", "PageRange" };
   static const QString STATE       [] = { "Idle", "Active", "Aborted", "Error" };
   static const QString SIZE        [] = {
      "A4",          // 0
      "B5",          // 1
      "Letter",      // 2
      "Legal",       // 3
      "Executive",   // 4
      "A0",          // 5
      "A1",          // 6
      "A2",          // 7
      "A3",          // 8
      "A5",          // 9
      "A6",          // 10
      "A7",          // 11
      "A8",          // 12
      "A9",          // 13
      "B0",          // 14
      "B1",          // 15
      "B10",         // 16
      "B2",          // 17
      "B3",          // 18
      "B4",          // 19
      "B6",          // 20
      "B7",          // 21
      "B8",          // 22
      "B9",          // 23
      "C5E",         // 24
      "Comm10E",     // 25
      "DLE",         // 26
      "Folio",       // 27
      "Ledgar",      // 28
      "Tabloid",     // 29
      "Custom",      // 30
   };

   qDebug() << " ";
   qDebug() << "===== PRINTER CONFIG =====";
   qDebug() << "Collate copies       : " << printer_.collateCopies();
   qDebug() << "Color mode           : " << COLOR_MODE[ printer_.colorMode() ];
   qDebug() << "Application name     : " << printer_.creator();
   qDebug() << "Document name        : " << printer_.docName();
   qDebug() << "Double sided printing: " << printer_.doubleSidedPrinting();
   qDebug() << "Embedded fonts       : " << printer_.fontEmbeddingEnabled();
   qDebug() << "From page            : " << printer_.fromPage();
   qDebug() << "Full page            : " << printer_.fullPage();
   qDebug() << "Number of copies     : " << printer_.numCopies();
   qDebug() << "Orientation          : " << ORIENTATION[ printer_.orientation() ];
   qDebug() << "Output file name     : " << printer_.outputFileName();
   qDebug() << "Output format        : " << FORMAT[ printer_.outputFormat() ];
   qDebug() << "Page order           : " << ORDER[ printer_.pageOrder() ];
   qDebug() << "Page rect            : " << printer_.pageRect();
   qDebug() << "Page size            : " << SIZE[ printer_.pageSize() ];
   qDebug() << "Paper rect           : " << printer_.paperRect();
   qDebug() << "Print program        : " << printer_.printProgram();
   qDebug() << "Print range          : " << RANGE[ printer_.printRange() ];
   qDebug() << "Printer name         : " << printer_.printerName();
   qDebug() << "Printer state        : " << STATE[ printer_.printerState() ];
   qDebug() << "Resolution           : " << printer_.resolution();
   qDebug() << " ";
}
*/
// end of print_config
